# Distribution Checker
# Common Functionality Module (Common.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Common;
use strict;
use Fcntl qw/:flock :seek/;
use FindBin;

use Error;

# Export symbols
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	DEBUG
	$DC_BRANCH $DC_VERSION
	$MANIFEST_DIR $CACHE_DIR $RESULTS_DIR
	$PROBLEM_DB_FILE  $PROBLEM_DB_LINK
	$STATUS_FILE  $DEFAULT_STANDARD_VER $DEFAULT_TEMP_DIR
	$DEFAULT_MANIFEST_URL
	$OPT_DIR $VAR_OPT_DIR
	&append_string_to_file &cache_dest &shq &trim
	&create_empty_directory &does_shell_know &extract_dir &extract_filename
	&format_duration &format_time &guess_architecture &glob_hash
	&in_array &sort_cat &copy &merge_hash &read_file &write_string_as_file
	&read_config
	&is_process_running &file_info
	&detect_architecture &backward_compatible_platforms
	&detect_OS &detect_standard_version
	@all_archs
	$last_cmd_output &last_line
	&complain &debug_inform &fail &inform &warning &capture &cmd
	&sleep_ms &unpack_file
	&vercmp &version_filter
);

push @EXPORT, @Error::EXPORT;

#----------------------------------------------------------------------

use constant DEBUG => 0; # DBG: Whether debug mode is enabled

$Error::Debug = 1 if DEBUG;
our $last_cmd_output = "";

our @all_archs = qw( x86  x86-64  IA64  PPC32  PPC64  S390  S390X );

our $STATUS_FILE = 'test_status';

our $DEFAULT_TEMP_DIR = '/tmp/dist-checker';

# ---------------------------------------------------------------------

our $DC_VERSION;
our $DC_BRANCH; # "LSB" or "Moblin"
{
	# Look for $DC_VERSION and $DC_BRANCH in the utils/VERSION file
	(my $file_path = __FILE__) =~ s/\/?[^\/]*$//; # get path to this file
	if ( open my $FILE, $file_path."/VERSION" ) {
		if ( defined( local $_ = <$FILE> ) ) { # read first line
			chomp;
			($DC_VERSION, $DC_BRANCH) = split /:/;
		}
		close $FILE;
	}
}
$DC_BRANCH = "LSB" if !$DC_BRANCH;
$DC_VERSION = "" if ( !$DC_VERSION or $DC_VERSION !~ /^\d/ ); # Version should start with digit

our $DEFAULT_STANDARD_VER = ($DC_BRANCH eq "LSB" ? "LSB 4.1" : "Moblin 2.0");

our $OPT_DIR = "/opt/".(lc $DC_BRANCH)."/test/manager";
our $VAR_OPT_DIR = "/var".$OPT_DIR;

our $CACHE_DIR       = $VAR_OPT_DIR."/packages";
our $RESULTS_DIR     = $VAR_OPT_DIR."/results"; # Cache directory for downloaded packages and test data files
our $MANIFEST_DIR    = $VAR_OPT_DIR."/manifest";
our $PROBLEM_DB_FILE = $VAR_OPT_DIR."/problem_db/problem_db2";

# Link to the problem_db file on the server of the LF.
our $PROBLEM_DB_LINK = 'http://ftp.linuxfoundation.org/pub/lsb/updates/dist-checker-data/problem_db2';

our $DEFAULT_MANIFEST_URL;
if ( $DC_BRANCH eq 'Moblin' ) {
	$DEFAULT_MANIFEST_URL = 'http://compliance.moblin.org/testkit/dist-checker-data/Manifest';
} else {
	$DEFAULT_MANIFEST_URL = 'http://ftp.linuxfoundation.org/pub/lsb/updates/dist-checker-data/Manifest';
}

#----------------------------------------------------------------------

# Print error message.
# If $err is an error object, then it is printed.
# If $err is a string, then an error($err, @params) is created, then printed.
# Returns the error object.
our $complain_sub = sub {
	my ($err, @params) = @_;
	
	( defined $err ) or $err = "";
	
	if ( !is_err_obj($err) ) {
		local $Error::Depth = $Error::Depth + 1;
		$err = error($err, @params);
	}
	
	print STDERR "\nError:\n".$err->tostring()."\n";
	
	return $err;
};
sub complain { goto &$complain_sub }

# Like complain(), but prints the 'Warning' word instead of 'Error'.
our $warning_sub = sub {
	my ($err, @params) = @_;
	
	( defined $err ) or $err = "";
	
	if ( !is_err_obj($err) ) {
		local $Error::Depth = $Error::Depth + 1;
		$err = error($err, @params);
	}
	
	print STDERR "\nWarning:\n".$err->tostring()."\n";
	
	return $err;
};
sub warning { goto &$warning_sub; }

# Prints error message like complain(), then exits.
our $fail_sub = sub { 
	my (@params) = @_;
	
	local $Error::Depth = $Error::Depth + 1;
	complain(@params) if @params;
	
	inform("Failed.");
	
	exit 1;
};
sub fail { goto &$fail_sub }
#----------------------------------------------------------------------

# Print message and a linefeed.
our $inform_sub = sub {
	print "$_[0]\n";
};
sub inform { goto &$inform_sub; }

# Print message only if DEBUG mode is ON.
our $debug_inform_sub = sub {
	print "$_[0]\n" if DEBUG;
};
sub debug_inform { goto &$debug_inform_sub; }
#----------------------------------------------------------------------

# Returns the output of the given command.
our $capture_sub = sub {
	my ($command) = @_;
	
	$command = "( ".$command." ) 2>&1";
	
	return `$command`;
};
sub capture { goto &$capture_sub; }

# Return the exit code of the given command.
# Save the command output to $last_cmd_output.
our $cmd_sub = sub {
	my ($command) = @_;
	
	$last_cmd_output = "%> ".$command;
	
	my $resp = capture ( $command );
	if ( is_err $resp ) { complain $Error::Last; return 256; }
	
	$last_cmd_output .= "\n".$resp if $resp ne "";
	
	return $?;
};
sub cmd { goto &$cmd_sub; }
#----------------------------------------------------------------------

# Checks whether $needle is present in array.
# Returns position of $needle in array, starting from 1.
# 0 is returned if there is no such element in the array.
sub in_array {
	my $needle = shift;
	
	( defined $needle )
		or return error("in_array(undef, ...)");
	
	for ( my $i = 0; $i < @_; $i++ ) {
		next if !defined $_[$i];
		return ($i+1) if $_[$i] eq $needle;
	}
	return 0;
}
#----------------------------------------------------------------------

# Compare function for extended sorting.
# Usage: sort {sort_cat($a, $b, \@arr);} @to_sort
# Places elements present in @arr at the beginning (in the same order as in @arr), then - all the rest (alphabetically).
# Example: sort {sort_cat($a, $b, ['y', 'w', 'x']);} ('a', 'x', 'y', 'z') == ('y', 'x', 'a', 'z')
sub sort_cat($$$) {
	my ($a, $b, $arr) = @_;
	my $ai = in_array($a, @$arr);
	my $bi = in_array($b, @$arr);
	return $ai <=> $bi if ($ai && $bi);
	return $a cmp $b   if (!$ai && !$bi);
	return 1           if !$ai;
	return -1          if !$bi;
}
#----------------------------------------------------------------------

sub shq {
	my ($s) = @_;
	$s =~ s/'/'\''/g; # escape quotes. foo'bar -> 'foo'\''bar'
	return "'$s'";
}

sub trim {
	my ($s) = @_;
	(defined $s) or return error("trim(undefined)");
	$s =~ s/^\s+//; # Remove spaces in the beginning of the line
	$s =~ s/\s+$//; # Remove trailing spaces
	return $s;
}

# Read and return the file contents
sub read_file {
	my ($filename) = @_;
	
	my $to_lock = 0;
	if ( $filename =~ s/^!// ) {
		$to_lock = 1;
	}
	
	my $text = undef;
	my $file;
	
	open $file, $filename
		or return error("Can't open file for read '$filename': $!");
		
	$to_lock && flock($file, LOCK_SH);
	{
		local $/ = undef;
		$text = <$file>;
	}
	$to_lock && flock($file, LOCK_UN);
	close $file;
	
	( defined $text ) or return error("Can't read file '$filename': $!");
	
	return $text;
}
#---------------------------------------------------------------------

# Extracts filename from a full name
sub extract_filename {
	my ($fullname) = @_;
	
	( defined $fullname ) or return error("undefined argument");
	
	if ( $fullname =~ m/([^\/]+)$/ ) {
		return $1;
	}
	return $fullname;
}
#---------------------------------------------------------------------

# Extracts directory name (with a trailing slash)
sub extract_dir {
	my ($fullname) = @_;
	
	( defined $fullname ) or return error("undefined argument");
	
	my $dir = $fullname;
	
	$dir =~ s{/+[^/]*$}{}; # Remove everything after the last /

	return $dir;
}
#---------------------------------------------------------------------

# Returns size and mtime of the $file.
# Returns (undef,undef) if error.
sub file_info {
	my ($file) = @_;
	
	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime) = stat($file);
	
	return ($size, $mtime);
}
#---------------------------------------------------------------------

# Write the specified string to the specified file, removing the original
# contents of the file (if any).
sub write_string_as_file {
	my ($file_name, $content) = @_;
	
	( defined $file_name && defined $content )
		or return error("undefined argument");

	# "!" at the beginning of the filename means that flock is required.
	my $to_lock = 0;
	if ( $file_name =~ s/^!// ) {
		$to_lock = 1;
	}

	if ( !$to_lock ) {
		my $file;
		open $file, ">".$file_name
			or return error("Failed to open file '$file_name' for writing: $!");
		print {$file} $content;
		close $file;
	}
	else {
		my $file;
		open $file, "+>>".$file_name
			or return error("Failed to open file '$file_name' for writing: $!");
		flock($file, LOCK_EX);
		seek($file, 0, SEEK_SET);
		truncate($file, 0);

		# Print the string and close the file.
		print {$file} $content;

		flock($file, LOCK_UN);
		close $file;
	}

	return 1; # OK
}
#---------------------------------------------------------------------

# Appends the string to the file.
sub append_string_to_file {
	my ($file_name, $content) = @_;

	( defined $file_name && defined $content )
		or return error("undefined argument");
	
	my $to_lock = 0;
	if ( $file_name =~ s/^!// ) {
		$to_lock = 1;
	}
	
	my $file;
	
	# Open the file for appending.
	open $file, ">>".$file_name
		or return error("Failed to open file '$file_name' for appending: $!");
	
	if ( $to_lock ) {
		flock($file, LOCK_EX);
		seek($file, 0, SEEK_END);
	}
	
	# Print the string and close the file.
	print {$file} $content;
	
	if ( $to_lock ) {
		flock($file, LOCK_UN);
	}
	close $file;
	
	return 1; # OK
}
#---------------------------------------------------------------------

# Creates the specified directory, or removes and re-creates it
# if the directory already exists.
sub create_empty_directory {
	my ($dir_name) = @_;

	( defined $dir_name )
		or return error("undefined argument");

	# If the directory already exists then delete it with all contents
	# and re-create to ensure it is empty and with standard access rights.
	if ( -d $dir_name ) {
		system("rm -rf ".shq($dir_name)." >/dev/null 2>&1") == 0 
			or return error("Failed to delete '$dir_name'.");
	}

	# Create the directory.
	my $res = system( "mkdir -p ".shq($dir_name) );
	if ( $res != 0 || !-d $dir_name ) {
		return error("Failed to create '$dir_name'.");
	}
	return 1; # OK
}
#---------------------------------------------------------------------

# Checks whether the specified tool is known to the shell.
sub does_shell_know ($) {
	my ($tool_name) = @_;
	
	( defined $tool_name ) or return error("undefined argument");
	
	return ( system ("which ".shq($tool_name)." >/dev/null 2>&1") == 0 );
}
#---------------------------------------------------------------------

# Returns time in format "HH:MM:SS dd-mm-yyyy"
sub format_time {
	my ($timestamp) = @_;
	
	( defined $timestamp ) or return error("undefined argument");
	
	( $timestamp eq int($timestamp) )
		or return error("parameter isn't numeric: '$timestamp'");
	
	my ($sec, $min, $hour, $mday, $mon, $year) = localtime ($timestamp);

	return sprintf ('%02d:%02d:%02d %02d-%02d-%04d', $hour, $min, $sec,
			$mday, ($mon + 1), ($year + 1900) );
}
#----------------------------------------------------------------------

# Format time interval in form "HH hours, MM minutes, SS seconds"
sub format_duration {
	my ($time) = @_;
	
	( defined $time ) or return error("undefined argument");
	
	$time = int($time);
	
	my $seconds = $time % 60;
	my $tmp = int($time / 60);
	my $minutes = $tmp % 60;
	my $hours = int($tmp / 60);
	
	my $res = "";
	
	$res .= sprintf '%d hours, ', $hours if $time >= 3600;
	$res .= sprintf '%02d minutes, ', $minutes if $time >= 60;
	$res .= sprintf '%02d seconds', $seconds;
	
	return $res;
}
#----------------------------------------------------------------------

# Returns element of a hash or a value by key '*'.
sub glob_hash {
	my ($hashref, @keys) = @_;

	return $hashref if !@keys;
	return undef if !defined $hashref;
	
	( ref($hashref) eq 'HASH' )
		or return error("HASHREF is expected as the first argument");

	my $key = shift @keys;
	if ( defined $hashref->{$key} ) {
		return glob_hash( $hashref->{$key}, @keys );
	}
	elsif ( defined $hashref->{'*'} ) {
		return glob_hash( $hashref->{'*'} );
	}
	# else
	return undef;
}
#----------------------------------------------------------------------

sub copy {
	my ($a, %opts) = @_;
	
	my $ref = ref($a);
	
	if ( $ref eq '' ) {
		return $a;
	}
	elsif ( $ref eq 'ARRAY' ) {
		return [ map {copy($_, %opts)} @$ref ]; # recursion!
	}
	elsif ( $ref eq 'HASH' ) {
		my $h = {};
		return merge_hash($h, $a, %opts); # indirect recursion!
	}
	else {
		warning "Don't know what to do with '$ref' ('$_')";
	}
}

# Merges the $h2 hash INTO the $h1
sub merge_hash {
	my ($h1, $h2, %opts) = @_;
	
	( ref $h1 eq 'HASH' ) or return error("argument1 is not a hash ref");
	( ref $h2 eq 'HASH' ) or return error("argument2 is not a hash ref");
	
	local $_;
	
	foreach ( keys %$h2 ) {
		if ( !defined $h1->{$_} ) {
			$h1->{$_} = $h2->{$_};
			next;
		}
		
		( ref($h1->{$_}) eq ref($h2->{$_}) )
			or return error("incongruent hashes: '$_'");
		
		my $ref = ref($h1->{$_});
		
		if ( $ref eq "HASH" ) {
			is_ok merge_hash( $h1->{$_}, $h2->{$_}, %opts ) # recursion
				or return $Error::Last;
		}
		elsif ( $ref eq "ARRAY" ) {
			if ( $opts{'overwrite_array'} ) {
				$h1->{$_} = [ map {copy($_, %opts)} @{$h2->{$_}} ];
			} else {
				# default: join arrays
				push @{$h1->{$_}}, @{$h2->{$_}};
			}
		}
		elsif ( $ref eq "" ) { # SCALAR
			next if $h1->{$_} eq $h2->{$_};
			if ( $opts{'overwrite_scalar'} ) {
				$h1->{$_} = $h2->{$_};
			} else {
				return error("Scalar conflict: '$_': '$h1->{$_}' != '$h2->{$_}'");
			}
			
		}
		else {
			warning "Don't know what to do with '$ref' ('$_')";
		}
	}
	return $h1;
}
#----------------------------------------------------------------------

sub cache_dest {
	my ($subdir) = @_;
	
	return $CACHE_DIR.(defined $subdir ? "/".$subdir : "");
}
#------------------------------------------------------------------------

# Helps to determine an architecture name.
sub guess_architecture {
	my ($name, $noarch_label) = @_;
	
	return "x86"    if ( $name =~ /^(x86|ia32|i[3-6]86)$/ );
	return "x86-64" if ( $name =~ /^(amd64|x86_64|x86-64)$/);
	return "IA64"   if ( $name =~ /^(ia64)$/ );
	return "PPC32"  if ( $name =~ /^(ppc|powerpc|ppc32)$/ );
	return "PPC64"  if ( $name =~ /^(ppc64)$/ );
	return "S390"   if ( $name =~ /^(s390)$/ );
	return "S390X"  if ( $name =~ /^(s390x)$/ );
	
	if ( $noarch_label ) {
		return $noarch_label if ( $name =~ /^(noarch|all)$/ );
	}
	
	# else
	return undef;
}
#------------------------------------------------------------------------

# Detect the architecture name
sub detect_architecture {
	# If uname utility supports -i options then use it
	# Check if `uname' has -i option
	if ( system ("uname --help 2>&1 | grep -- ' -i\\b' >/dev/null") == 0 ) { 
		my $tmp = trim(`uname -i`);
		my $res = guess_architecture( $tmp );
		if ( $res ) { return $res; }
	}
	# else:
	
	# Try to find out the host architecture using the machine type.
	my $tmp = trim(`uname -m`);
	my $res = guess_architecture( $tmp );
	if ( $res ) { return $res; }
	
	return undef;
}

#------------------------------------------------------------------------

# Returns list of compatible architectures.
sub backward_compatible_platforms {
	my $arch = shift;

	if ( $arch eq "x86-64" ) {
		return ("x86", "x86-64");
	}
	elsif ( $arch eq "IA64" ) {
		return ("x86", "IA64");
	}
	elsif ( $arch eq "PPC64" ) {
		return ("PPC32", "PPC64");
	}
	elsif ( $arch eq "S390X" ) {
		return ("S390", "S390X");
	}
	# else: return itself
	return ($arch);
}

#------------------------------------------------------------------------

sub detect_OS {
	if ( does_shell_know('lsb_release') ) {
		# The OS version is determined via the lsb_release tool
		# (this should be a canonical, distro independent way).
		my $res = trim(`lsb_release -s -d`);
		$res =~ s/^\"(.*)\"$/$1/;  # remove quotes if any
		return $res;
	}
	else {
		my $res = read_file("/etc/issue");
		if ( !is_ok($res) ) { $res = ""; }
		$res =~ s/\n.*//s; # leave only the first line
		return $res;
	}
	return "";
}
#------------------------------------------------------------------------

sub detect_standard_version {
	my ($OS) = @_;
	
	# Try to detect supported LSB version
	if ( does_shell_know('lsb_release') ) {
		my $lsb_release_v = trim(`lsb_release -s -v`);
		
		if    ( $lsb_release_v =~ /4\.1/ ) { return "LSB 4.1"; }
		elsif ( $lsb_release_v =~ /4\.0/ ) { return "LSB 4.0"; }
		elsif ( $lsb_release_v =~ /3\.2/ ) { return "LSB 3.2"; }
		elsif ( $lsb_release_v =~ /3\.1/ ) { return "LSB 3.1"; }
		elsif ( $lsb_release_v =~ /3\.0/ ) { return "LSB 3.0"; }
		elsif ( $lsb_release_v =~ /2\.0/ ) { return "LSB 2.0"; }
	}
	
	if ( defined $OS && $OS =~ /\bMoblin\b/i ) {
		return "Moblin 2.0"; # TODO: detect Moblin version in future
	}
	
	return $DEFAULT_STANDARD_VER;
}
#------------------------------------------------------------------------

sub read_config {
	my ($filename) = @_;
	
	my $res = {};
	
	# TODO: use read_file to be able to lock the file
	open FILE, $filename
		or return error("Failed to open testinfo file '$filename': $!");
	
	my $section = undef;
	while ( my $line = <FILE> ) {
		$line = trim($line); # Trim spaces

		next if !$line;  # Skip empty lines
		next if $line =~ /^#/;  # Skip comments
		
		if ( $line =~ /^\[([^\]]+)\]$/  ) {
			$section = $1;
			next;
		}
		elsif ( $line =~ /([^:=\s]+)\s*[:=\s]\s*(.*)/ ) {
			my $name = $1;
			my $value = $2;
			
			$value =~ s/\x81/\n/g;
			
			if ( !defined $section ) {	
				$res->{$name} = $value;
			}
			else {
				$res->{SECTIONS}{$section}->{$name} = $value;
			}
		}
		else {
			warning "Cant parse line $. in file '$filename': '$line'";
		}
	}
	
	close FILE;
	
	return $res;
}
#----------------------------------------------------------------------

sub is_process_running($;$) {
	my ($PID, $procname) = @_;
	if (!-e "/proc/$PID") {
		return 0;
	}
	open(FILE, "/proc/$PID/stat") or return 0;
	my $info = <FILE>;
	close(FILE);
	if ($info =~ m/^\d+\s+\((.*)\)\s+(\S)\s+[^\(\)]+$/) {
		my $running = 1;
		if ($procname) {
			$running &&= ($1 eq $procname);
		}
		return ($running and ($2 ne 'Z'));
	}
	else {
		return 0;
	}
}
#----------------------------------------------------------------------

sub last_line {
	my ($text) = @_;
	
	$text =~ s/[\n\r]+$//;
	
	$text =~ s/.*[\n\r]//sg;
	return $text;
}

#----------------------------------------------------------------------

# Sleep $delay milliseconds
sub sleep_ms {
	my ($delay) = @_;
	
	select undef, undef, undef, $delay;
}
#----------------------------------------------------------------------

sub unpack_file {
	my ($file, $dest_dir) = @_;

	( -d $dest_dir ) or return error "Missing target directory '$dest_dir'";
	my $tar_opts;
	if ( $file =~ /\.(tgz|tar\.gz)$/ ) {
		$tar_opts = '-xzf';
	}
	elsif ( $file =~ /\.(tbz|tar\.bz)$/ ) {
		$tar_opts = '-xjf';
	}
	elsif ( $file =~ /\.(tar)$/ ) {
		$tar_opts = '-xf';
	} else {
		return error "Unknown archive type: ".extract_filename($file);
	}
	cmd('tar -C '.shq($dest_dir).' --no-same-owner '.$tar_opts.' '.shq($file)) == 0
		or return error "tar failed: $?\n".$last_cmd_output;

	return 1; # ok
}

#----------------------------------------------------------------------

# Smart package versions comparison
sub vercmp {
	my ($ver1, $ver2) = @_;
	
	( defined $ver1 && defined $ver2 )
		or return error "undefined parameter: '$ver1' '$ver2'";
	
	my @spl1 = split /[:\-_]/, $ver1;
	my @spl2 = split /[:\-_]/, $ver2;
	
	for (my $i=0; ; $i++ ) {
		if ( !defined($spl1[$i]) ) {
			if ( !defined($spl2[$i]) ) {
				return 0;
			}
			return -1;
		}
		if ( !defined $spl2[$i] ) {
			return 1;
		}
		
		my $res = vercmp_sub($spl1[$i], $spl2[$i]);
		return $res if $res != 0;
		# continue working if $res == 0;
	}
	return 0; # equal
}

sub vercmp_sub {
	my ($ver1, $ver2) = @_;
	
	my @spl1 = split /[\.]/, $ver1;
	my @spl2 = split /[\.]/, $ver2;

	for ( my $i = 0; ; $i++ ) {
		my $a = $spl1[$i];
		my $b = $spl2[$i];

		if ( !defined($a) ) {
			if ( !defined($b) ) {
				return 0;
			}
			$a = 0;
		}
		if ( !defined $b ) {
			$b = 0;
		}

		next if $a eq $b;

		while ( 1 ) {
			#print "< ".substr($a,pos$a)." ".substr($b,pos$b).$/ if pos$a && pos$b;
			if ( $a =~ /\G(\d+)/gc ) {
				my $A = $1;
				if ( $b =~ /\G(\d+)/gc ) {
					# both numeric
					return 1  if $A > $1;
					return -1 if $A < $1;
					# continue processing if $a == $b
				}
				#elsif ( $b =~ /\G([^\d]+)/gc ) {
					#return 1;
				#} # commented out because the result is same as in 'else'
				else { # end of $b
					return 1;
				}
			}
			elsif ( $a =~ /\G([^\d]+)/gc ) {
				my $A = $1;
				if ( $b =~ /\G([^\d]+)/gc ) {
					return 1 if $a gt $b;
					return -1 if $a lt $b;
				}
				#elsif ( $b =~ /\G(\d+)/gc ) {
					#return -1;
				#} # commented out because the result is same as in 'else'
				else {
					return -1; # 8.5a < 8.5
				}
			}
			else { # end of $a
				if ( $b =~ /\G(\d+)/gc ) {
					return -1;
				}
				elsif ( $b =~ /\G([^\d]+)/gc ) {
					return 1; # 8.5 > 8.5b
				}
				else {
					last;
				}
			}
		}
	}
	
	return 0; # equal
}
#----------------------------------------------------------------------

# Match $ver against $filter.
# $ver e.g. 'LSB 3.2'
# $filter e.g. 'LSB >=3.0 & <=3.2 | MOBLIN 2.0'
sub version_filter {
	my ( $filter, $ver ) = @_;
	
	( defined $ver ) or return error "undefined parameter";
	
	return 1 if (!defined $filter || $filter eq '*'); # Return "ok" if no filter.
	
	my $w = "";
	if ( $ver =~ s/^([a-z][\w_\-]*)\s*//i ) { # The version is prepended by a word: e.g. "LSB 4.0"
		$w = $1;
	}
	
	for my $s ( split /\|\s*/, $filter ) { # or

		if ( $s =~ /^([a-z][\w_\-]*)\s*/igc ) {
			next if $w ne $1;
			return 1 if $ver eq ""; # version_filter('LSB 4.0', 'LSB');
		}

		while ( $s !~ /\G$/ ) {
			my $op = "=";
			if ( $s =~ m/\G(<=|<|=|>=|>)\s*/gc ) {
				$op = $1;
			}

			if ( $s =~ m/\G(\d[a-z\d\.\-]+)[&\s]*/gc ) { # version (skip '&' and spaces then)
				if    ( $op eq "<"  ) {  return 0 unless vercmp($ver, $1) <  0;  }
				elsif ( $op eq "<=" ) {  return 0 unless vercmp($ver, $1) <= 0;  }
				elsif ( $op eq "="  ) {  return 0 unless vercmp($ver, $1) == 0;  }
				elsif ( $op eq ">=" ) {  return 0 unless vercmp($ver, $1) >= 0;  }
				elsif ( $op eq ">"  ) {  return 0 unless vercmp($ver, $1) >  0;  }
				elsif ( $op eq "!=" ) {  return 0 unless vercmp($ver, $1) != 0;  }
			}
			else {
				warning "Wrong filter near '$s' in '$filter'";
				return 0;
			}
		}
		
		return 1;
	}
	
	return 0;
}
#----------------------------------------------------------------------

1; #return value
