# Distribution Checker
# Download Routines Module (Download.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Download;

use strict;
use Time::Local;

use Common;

# Export symbols
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	&download_file &spider
	&download_setup
);
#----------------------------------------------------------------------

our $enabled = 0;

our $tool = "curl"; # "wget" is an alternative.

our $options = {
		'timeout' => 180, # seconds
		'retries' => 2,   # How many times to retry. 0 is for infinite retrying.
		'retr_wait' => 3, # seconds, wait between retries.
	};
# Other options:
#   http_proxy_host, http_proxy_port,
#   http_proxy_user, http_proxy_password,
#   http_proxy_auth, http_proxy_notunnel.
#  The same options are for FTP: ftp_proxy_*

#----------------------------------------------------------------------

# %params:
#  'to' => $file     - save to file
#  'continue' => 1   - continue downloading partially downloaded file
#  'spider' => 1     - request only header
#  'raw' => $params  - additional parameters that are appended as is
sub curl_cmd {
	my ($link, %params) = @_;
	
	# Curl options:
	# -R --remote-time  Keep timestamp
	# -L --location     Follow redirect
	# --compressed      Ask server to compress the requested data
	# Don't use '-#' (Linear progress meter) -- produces too much noise in logs.
	my $cmd = "curl -R -L --compressed";

	# Dump HTTP/FTP headers to output
	$cmd .= " -D -" unless $params{'spider'};
	# How many times to retry:
	$cmd .= " --retry ".$options->{'retries'} if $options->{'retries'};
	# Wait between retries:
	$cmd .= " --retry-delay ".$options->{'retr_wait'} if $options->{'retr_wait'};
	# Timeout:
	$cmd .= " --connect-timeout ".$options->{'timeout'} if $options->{'timeout'};
	# Continue downloading partially downloaded file?
	$cmd .= " -C -" if $params{'continue'};
	
	my $protocol;
	if ( $link =~ /^(http|ftp):/ ) {
		$protocol = $1;
	}
	# Proxy options
	if ( $protocol && $options->{$protocol.'_proxy_host'} ) {
		my $p = $protocol;
		
		my $proxy = $options->{$p.'_proxy_host'};
		$proxy .= ":".$options->{$p.'_proxy_port'} if $options->{$p.'_proxy_port'};
		
		# HTTP proxy
		$cmd .= " -x ".shq($proxy);
		
		if ( $options->{$p.'_proxy_user'} ) {
			my $proxy_user = $options->{$p.'_proxy_user'};
			$proxy_user .= ":".$options->{$p.'_proxy_password'} if $options->{$p.'_proxy_password'};
			
			# -U --proxy-user   Specify user and password to use for proxy authentication
			$cmd .= " -U ".shq($proxy_user);
		}
		
		if ( $options->{$p.'_proxy_auth'} ) {
			if ( $options->{$p.'_proxy_auth'} =~ /^(basic|digest|ntml|negotiate|anyauth)$/ ) {
				# Proxy authentication method
				$cmd .= " --proxy-".$options->{$p.'_proxy_auth'};
			} else {
				warning "Wrong $p\-proxy auth method: '".$options->{$p.'_proxy_auth'}."'";
			}
		}
		
		unless ( $options->{$p.'_proxy_notunnel'} ) {
			# -p --proxytunnel  Ftp protocol tunnel through the http proxy
			$cmd .= " -p";
		}
	}
	
	if ( $params{'spider'} ) {
		$cmd .= " --head";
	}
	
	$cmd .= " ".$params{'raw'} if $params{'raw'};
	
	# URL:
	$cmd .= " ".shq($link);
	
	# Output to file:
	$cmd .= " -o ".shq($params{'to'}) if $params{'to'};
	
	return $cmd;
}

sub wget_cmd {
	my ($link, %params) = @_;
	
	# Wget options:
	# -S                    Show headers
	# --progress=dot:mega   Print less dots (to avoid flooding to the log file)
	# --retr-symlinks       Get the linked-to file instead of the symlink itself
	my $cmd = "wget -S --progress=dot:mega --retr-symlinks";
	# How many times to retry:
	$cmd .= " -t ".$options->{'retries'} if $options->{'retries'};
	# Wait between retries:
	$cmd .= " -w ".$options->{'retr_wait'} if $options->{'retr_wait'};
	# Timeout:
	$cmd .= " --timeout=".$options->{'timeout'} if $options->{'timeout'};
	# Continue downloading partially downloaded file?
	$cmd .= " -c" if $params{'continue'};
	
	my $protocol;
	if ( $link =~ /^(http|ftp):/ ) {
		$protocol = $1;
	}
	if ( $protocol && $options->{$protocol.'_proxy_host'} ) {
		my $p = $protocol;
		
		my $proxy = $options->{$p.'_proxy_host'};
		$proxy .= ":".$options->{$p.'_proxy_port'} if $options->{$p.'_proxy_port'};
		
		# HTTP proxy
		$cmd = $p."_proxy=".shq($proxy)." ".$cmd;
		
		if ( $options->{$p.'_proxy_user'} ) {
			# --proxy-user   Specify the user name to use for proxy authentication
			$cmd .= " --proxy-user ".shq($options->{$p.'_proxy_user'});
		}
		
		if ( $options->{$p.'_proxy_password'} ) {
			# --proxy-user   Specify the user name to use for proxy authentication
			$cmd .= " --proxy-password ".shq($options->{$p.'_proxy_password'});
		}
		
		if ( $options->{$p.'_proxy_auth'} ) {
			unless ( $options->{$p.'_proxy_auth'} =~ /^(basic|anyauth)$/ ) {
				warning "Wget doesn't support authentication method ".$options->{$p.'_proxy_auth'};
			}
			# ignore
		}
		
		# ignore:  $options->{$p.'_proxy_notunnel'}
	}
	
	if ( $params{'spider'} ) {
		$cmd .= " --spider";
	}
	
	$cmd .= " ".$params{'raw'} if $params{'raw'};
	
	# URL:
	$cmd .= " ".shq($link);
	
	# Output to file:
	$cmd .= " -O ".shq($params{'to'}) if $params{'to'};

	return $cmd;
}

#----------------------------------------------------------------------

# Download $link to file $dest (file name with path).
#
# %options:
#  $options{RESUME} - continue downloading partially-downloaded files.
#  $options{OVERWRITE} - download and overwrite the existing file, even if it looks similar.
#  $options{SPIDER}->{} - Use this data instead of quering information about the file from server.
#  $options{QUIET} - Suppress some output.
#  $options{MD5SUM} - Do a checksum test.
#
# This function will create the destination directory if it doesn't exist.
#
# Returns an error if failed to download the file.
sub download_file {
	my ($link, $dest, %options) = @_;
	
	( $enabled ) or return error "Downloading isn't enabled.";
	
	# Create the directory if it doesn't exist.
	my $dest_dir = extract_dir( $dest )
		or return error "Wrong directory of file '$dest'";
	if ( !-d $dest_dir ) {
		cmd( "mkdir -p ".shq($dest_dir) );
	}
	( -d $dest_dir ) # check again
		or return error "Failed to create directory '$dest_dir'";
	
	# Do a spider query to receive information about the file from the server
	my $sp;
	if ( $options{SPIDER} ) {
		$sp = $options{SPIDER};
	} else {
		$sp = spider($link);
	}
	if ( $sp->{NOTFOUND} ) {
		return error "Can't download '$link':\n The file was removed or replaced.";
	}
	
	# Check if the file is already downloaded
	if ( -f $dest ) {
		my ($size, $mtime) = file_info($dest);
		if ( defined $sp->{SIZE} && defined $size && $sp->{SIZE} == $size
			 && defined $sp->{TIMESTAMP} && $sp->{TIMESTAMP} == $mtime )
		{
			# The file seems to be the same as the one on server
			debug_inform "File '$link' is already downloaded.";
			return 1 unless $options{OVERWRITE};
		}
	}
	
	# The $link will be downloaded to a temporary file,
	# then will be moved to $dest if everything is OK.
	my $temp_file = $dest.".#part";
	
	# Check if the file was partially downloaded
	my $continue = 0;
	if ( -f $temp_file ) {
		my ($tmp_size, $tmp_mtime) = file_info($temp_file);
		# Ensure the file part isn't out-of-date
		if ( $sp->{TIMESTAMP} && $tmp_mtime > $sp->{TIMESTAMP} ) {
			$continue = 1 if $options{RESUME};
		}
		if ( !$continue ) {
			# remove the temporary file if don't want to resume downloading
			cmd("rm -f ".shq($temp_file) );
		}
	}
	
	inform "Downloading '$link'".( $sp->{SIZE} ? " ".$sp->{SIZE}." bytes" : "")  unless $options{QUIET};
	
	my $cmd;
	if ( $tool eq 'curl' ) { # CURL
		$cmd = curl_cmd($link, 'to' => extract_filename($temp_file), 'continue' => $continue);
	}
	elsif ( $tool eq 'wget' ) { # WGET
		$cmd = wget_cmd($link, 'to' => extract_filename($temp_file), 'continue' => $continue);
	}
	else { return error "Unexpected download tool: '$tool'"; }
	
	$cmd .= ' >/dev/null' if $options{QUIET};
	
	my $resp = capture( "cd ".shq($dest_dir)." && ".$cmd );
	( $? == 0 ) or return error "Sorry, failed to download '$link'.", error $resp;

	if ( !-f $temp_file ) {
		return error "Sorry, failed to download '$link'.", error $resp;
	}

	# Parse the most recent header (in the case of the file has been untimely replaced, or some cache issues or whatever).
	{
		my $sp_new;
		if ( $tool eq 'curl' ) { # CURL
			$sp_new = curl_spider_parse($resp, $link);
		}
		elsif ( $tool eq 'wget' ) { # WGET
			$sp_new = wget_spider_parse($resp, $link);
		}
		if ( $sp_new ) {
			$sp->{SIZE} = $sp_new->{SIZE} if $sp_new->{SIZE};
			$sp->{TIMESTAMP} = $sp_new->{TIMESTAMP} if $sp_new->{TIMESTAMP};
			$sp->{NOTFOUND} = $sp_new->{NOTFOUND} if $sp_new->{NOTFOUND};
		}
	}
	if ( $sp->{NOTFOUND} ) {
		cmd( "rm -f ".shq($temp_file) );
		return error "Sorry, failed to download '$link'.";
	}

	# Check the file size
	if ( $sp->{SIZE} ) {
		my $size = -s $temp_file;
		if ( $size != $sp->{SIZE} ) {
			return error "Size of the downloaded file ($size) differs from the expected one ($sp->{SIZE}).";
		}
	}

	# set the file time
	if ( $sp->{TIMESTAMP} ) { # Set the file modification time as on the server
		# Wget doesn't keep the file modification time when used with the -O key.
		utime $sp->{TIMESTAMP}, $sp->{TIMESTAMP}, $temp_file;
	}

	# Move the file
	my $result = cmd( "mv -f ".shq($temp_file)." ".shq($dest) );
	if ( $result != 0 || !-f $dest ) { # Failed
		return error "Failed to move the downloaded file '$temp_file' to '$dest'.";
	}
	
	# check the md5sum, if we have a checksum file
	if ( -f $dest . ".md5" && $options{MD5SUM}) {
		my $result = cmd("cd ".shq($dest_dir)." && md5sum -c ".shq($dest).".md5");
		if ( $result != 0 ) { # Failed
			return error "Checksum failed for '$dest'.";
		} else {
			inform "Checksum OK for '$dest'.";
		}
	}

	inform "Downloaded file '".extract_filename($dest)."' ".(-s $dest)." bytes" unless $options{QUIET};
	
	return 1; # Successfully downloaded
}
#----------------------------------------------------------------------

# Converts time string of format "Wed, 25 Mar 2009 09:26:18 GMT" into unix time numeric value.
# This is used for parsing the HTTP header's "Last-Modified" field.
sub str2time {
	my ($s) = @_;
	
	#     "Wed,    25      Mar     2009     09:26:18          GMT"
	$s =~ /\w+,\s+(\d+)\s+(\w+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+GMT/ or return undef;
	#             $mday   $MON    $year   $hour $min  $sec
	my ($mday, $MON, $year, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);
	
	( $mday >= 1 && $mday <= 31 ) or return undef;
	( $year >= 1900 ) or return undef;
	
	# Find the month number
	my $mon = in_array( lc $MON, qw{jan feb mar apr may jun jul aug sep oct nov dec} ); # from 1 to 12
	( $mon > 0 ) or return undef;
	
	return timegm($sec,$min,$hour,$mday,$mon-1,$year-1900);
}

sub http_header_parse {
	my ($line, $res) = @_;

	if ( $line =~ /^\s*Last-Modified:\s*(.*)/ ) {
		my $dt = $1;
		$res->{TIMESTAMP} = str2time($dt);
	}
	elsif ( $line =~ /^\s*Content-Length:\s*(\d+)/ ) {
		$res->{SIZE} = int($1);
	}
	# Content-Range: bytes 3768320-5846490/5846491
	elsif ( $line =~ /^\s*Content-Range:\s*bytes\s+[^\/]*\/\s*(\d+)/ ) {
		$res->{SIZE} = int($1);
	}
	elsif ( $line =~ /^\s*HTTP\S*\s+(\d+)\b/ ) {
		my $code = $1;
		if ( $code =~ /^[45]\d\d$/ ) { # 4xx - client error, 5xx - server error
			$res->{NOTFOUND} = 1;
		}
	}
}

sub spider {
	my ($link) = @_;

	( $enabled ) or return error "Downloading isn't enabled.";

	my $res = {};

	if ( $tool eq 'curl' ) {
		my $resp = capture( curl_cmd($link, 'spider' => 1) );
		$res = curl_spider_parse($resp, $link);
	}
	elsif ( $tool eq 'wget' ) {
		my $resp = capture( wget_cmd($link, 'spider' => 1) );
		$res = wget_spider_parse($resp, $link);
	}
	
	return $res;
}

sub curl_spider_parse {
	my ($resp, $link) = @_;
	
	my $res = {};
	
	my @lines = split /[\r\n]+/, $resp;
	while ( @lines ) {
		my $line = shift @lines;
		
		# curl response looks similar for both http and ftp (many thanks to curl developers)
		http_header_parse($line, $res);
	}
	
	return $res;
}

# Get some information about file at a remote server specified by $link.
# Requests a HTTP header using the '--spider' option of wget.
# Returns $res->{SIZE => $size, TIMESTAMP => $last_modified, NOTFOUND => $if_404 }
sub wget_spider_parse {
	my ($resp, $link) = @_;
	
	my $res = {};
	
	my $resp_mode = '';
	if ( $link =~ /^https?:/ ) {
		$resp_mode = 'http';
	}
	elsif ( $link =~ /^s?ftp:/ ) {
		$resp_mode = 'ftp';
	}
	else {
		warning "Unsupported protocol: '$link'";
		return $res;
	}
	
	my @lines = split /[\r\n]+/, $resp;
	while ( @lines ) {
		my $line = shift @lines;
		
		if ( $resp_mode eq 'http' ) {
			# Parse an HTTP header line
			http_header_parse($line, $res);
		}
		elsif ( $resp_mode eq 'ftp' ) {
			# Parse an FTP response line
			if ( $line =~ /^\s+HTTP\b/ ) {
				# HTTP response detected (probably HTTP-FTP proxy is in use)
				# Switch to parsing HTTP response
				unshift @lines, $line;
				$resp_mode = 'http';
				redo;
			}
			elsif ( $line =~ /^--> SIZE / ) { # file size requested
				$line = $lines[0]; # the next line
				$line = $lines[1] if !$line; # skip a line if it is empty
				if ( $line && $line =~ /^213s+(\d+)/ ) { # 213 - file status code
					$res->{SIZE} = int($1);
				}
			}
		}
	}
	return $res;
}

sub download_setup() {
	if( !does_shell_know($tool) ) {
		# Try another tool
		my $prev_tool = $tool;
		if ( $tool eq 'curl' ) {
			$tool = 'wget';
		}
		elsif ( $tool eq 'wget' ) {
			$tool = 'curl';
		}
		else {
			return error "Unknown download tool: '$tool'";
		}
	
		if( does_shell_know($tool) ) {
			inform "'$prev_tool' is unavailable, using '$tool' instead.";
		} else {
			return error "Couldn't find neither 'curl' nor 'wget'.";
		}
	}
	if ( $tool eq 'curl' ) {
		# check if --retry option is available
		if ( cmd('curl --retry 0 --version') != 0 ) {
			# Complies on unknown --retry option if unsupported
			$options->{'retries'} = undef;
			$options->{'retr_wait'} = undef;
		}
	}
	
	$enabled = 1; # global var
	return 1; # OK
}

#----------------------------------------------------------------------
1; # returned value

