# Distribution Checker
# Manifest Management Module (Manifest.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Manifest;
use strict;

use File::Basename qw(basename dirname);
use FindBin;

use Common;

# Export symbols
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	MANIFEST_FORMAT_VER

	&manifest_parse &expand_aliases
	&select_file &file_link
	&check_files
	&main_module_file
	
	&load_manifest &merge_manifests
);

#----------------------------------------------------------------------

# Some enums and constants

use constant {
	MANIFEST_FORMAT_VER => 2, # Supported Manifest format version
};

#----------------------------------------------------------------------

# Parse a manifest file
sub manifest_parse {
	my ($filename, %sections) = @_;
	
	# TODO: use flock to read manifest!
	
	open FILE, $filename
		or return error "Failed to open file '$filename': $!";
	
	my $manifest = {}; # Result structure
	#
	# ->{MANIFEST}->{VERSION|URL|TIMESTAMP|DC_VERSION_AVAILABLE};
	#
	# ->{ALL}{STANDARDS} = ('LSB 4.0', 'LSB 3.2', ..., 'Moblin 2.0');
	#        {TESTSUITES} = ('apache', 'cmdchk', ...);
	#        {ARCHITECTURES} = ('x86', ...);
	#        {STATUSES} = ('cert', 'beta', 'snapshot', ...);
	
	# ->{TREE}{$standard}{$testsuite} = "<..>/<..>"
	
	# ->{STD_PROFILES}{$standard}{DISPLAY_NAME}{$profile_id} = "Display name (html)";
	#                            {DEFAULT} = $profile_id;
	#                            {TESTS}{$testsuite}{$profile_id} = 1;
	
	# ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
	
	# ->{MODULES}{$testsuite}[]->{MAIN|TEST_VER|STANDARD|FILE|FROM}
	# ->{DESCRIPTIONS}{$testsuite}[]->{STANDARD|DISPLAYNAME|DESCRIPTION};
	# ->{OPTIONS}{$testsuite}[]->{ID|Q|GLOBAL=>1|TEST_VER|STANDARD|PRE|TYPE|VARIANTS->[]|DEFAULT|VALUE};
	## 'GLOBAL' options don't have TEST_VER specified.
	
	# ->{MANUAL}{$testsuite}[]->{STEP|STANDARD|OPTIONAL}
	
	# ->{CONFLICTS}{$package_name}->{$package_name} = 1;
	# ->{SERVERS}->{$dir}{$link} = 1;
	# ->{FILES}{"$src/$filename"} = $file->{FILENAME|LOCATION|SRC|REMOVED|SIZE|TIMESTAMP|ARCH|KIND|PACKAGE|VERSION}
	# ->{FILES}{"$src/$filename"}{DOWNLOADED} = (0: not downloaded | 1: downloaded | -1: downloaded wrong file) # auxiliary
	# ->{FILES}{"$src/$filename"}{OPTIONS}{RELOAD|KEEP|REINSTALL}
	
	# ->{ALIASES}  {$alias_name} = $string;
	# ->{LOCATIONS}{$alias_name} = $location;
	
	# Auxiliary hash map
	# ->{PACKAGES}{$package_name}->{FILES}[] = $filekey; # $filekey = "$src/$filename"
	#                            ->{VERSIONS}{$arch}{$ver} = 1; # auxiliary: installed versions detected by packages_info()
	#                            ->{KEEP}{$package_ver} = 1; # auxiliary: Do not deinstall this package
	
	my $all_standards = {};
	my $all_testsuites = {};
	my $all_architectures = {};
	my $all_statuses = {};
	
	my $section = undef;
	
	## Section-local variables
	my ($testsuite, $test_ver, $standard, $status, $arch, $opt_obj, $descr_obj, $servers_dir);
	my $linenum = 0; # line number
	
	local *manifest_error = sub {
		my ($msg, @other) = @_;
		
		my $pref = "$filename:$linenum"
			.(defined $section ? " [$section]" : "")
			.": ";
		$msg = $pref.$msg;
		return error $msg, -data => $manifest, @other;
	};
	
	# common patterns
	my $re_testsuite = '[\w_\-]+';
	my $re_version = '(?i:\d[a-z\d\.\-]+)';
	my $re_verfilt = '(?:\&\s*)?(?:<=|<|=|>=|>)?\s*'.$re_version;
	my $re_standard = '(?i:[a-z]+\s*'.$re_version.')';
	my $re_arch = join "|", @all_archs;
	my $re_status = '[a-z\-]+';
	my $re_package = '[\w_\-]+';
	my $re_file = '[\w_\-\.]+';
	my $re_path = '[\/\w_\-\.]+';
	my $re_alias = '[\w_\-\.]+';
	my $re_link = '(?:https?|s?ftp):\/\/(?:[^<>!\$\^\*\(\)\|\n ]+|<arch>)+';
	
	while ( my $line = <FILE> ) {
		$linenum++;
		$line = trim($line); # Trim spaces
		next if $line =~ /^#/; # Skip comments
		next if $line =~ /^$/; # Skip empty lines

		if ( $line =~ /^\[\s*([A-Z]+)\s*\]\s*$/ ) { # [SECTION]
			my $name = $1;
			if ( !defined $section ) {
				# This is the first section
				if ( $name ne 'MANIFEST' ) {
					return error "The first section in the Manifest file should be '[MANIFEST]'.";
				}
			}
			$section = $name;
			
			# Reset all the section-local variables
			$testsuite = undef;
			$test_ver = undef;
			$standard = undef;
			$status  = undef;
			$arch    = undef;
			$opt_obj = undef;
			$descr_obj = undef;
			$servers_dir = undef;
			
			next;
		}

		if ( !defined $section ) {
			warning manifest_error("Text before any section: '$line'.");
			next;
		}
		if ( %sections && !$sections{$section} ) {
			next;
		}
		if ( $section eq 'MANIFEST' ) {
		# [MANIFEST]
			if ( $line =~ /^VERSION\s*=\s*(.*)$/ ) {
				my $ver = $1;
				$manifest->{MANIFEST}{VERSION} = $ver;
				if ( $ver ne MANIFEST_FORMAT_VER ) {
					return manifest_error("Unsupported Manifest version: '$ver'. The only supported version is ".MANIFEST_FORMAT_VER);
				}
			}
			elsif ( $line =~ /^URL\s*=\s*(.*)$/ ) {
				$manifest->{MANIFEST}{URL} = $1;
			}
			elsif ( $line =~ /^TIMESTAMP\s*=\s*(.*)$/ ) {
				$manifest->{MANIFEST}{TIMESTAMP} = $1;
			}
			elsif ( $line =~ /^DC_VERSION_AVAILABLE\s*=\s*(.*)$/ ) {
				$manifest->{MANIFEST}{DC_VERSION_AVAILABLE} = $1;
			}
			else {
				warning manifest_error("Unhandled text: '$line'.");
				next;
			}
		}
		elsif ( $section eq 'TREE' ) {
		# [TREE]
			my @line_spl = split /\//, $line;
			
			my $standard = shift @line_spl;
			$standard = $1 if $standard =~ /^<\s*($re_standard)\s*>$/; # e.g. <LSB 3.1>
			$standard =~ s/^([a-z]+)\s*(\S*)$/$1 $2/i; # "LSB3.1" -> "LSB 3.1"
			
			my $testsuite = pop @line_spl;
			if ( !defined $testsuite || $testsuite !~ /^$re_testsuite$/ ) {
				return manifest_error("Wrong test name: '$testsuite'");
			}
			
			if ( defined $manifest->{TREE}{$standard}{$testsuite} ) {
				warning manifest_error("Attempt to override tree entry");
			}
			
			$manifest->{TREE}{$standard}{$testsuite} = join "/", @line_spl;
			
			$all_standards->{$standard} = 1;
			$all_testsuites->{$testsuite} = 1;
		}
		elsif ( $section eq 'DESCRIPTIONS' ) {
		# [DESCRIPTIONS]
			if ( $line =~ /^TESTSUITE\s*END$/) {
				$testsuite = undef;
				$standard = undef;
				$descr_obj = undef;
			}
			elsif ( $line =~ /^TESTSUITE\s+($re_testsuite)\s*(.*)$/ ) {
			# TESTSUITE test_name [Standard (<=|<|=|>=|>) $ver]
				$testsuite = $1;
				my $rest = $2;
				
				$standard = '';
				
				if ( $rest =~ s/^(\w+\s*($re_verfilt\s*)*(\s*\|\s*\w+\s*($re_verfilt\s*)*)*)//i ) {
					$standard = $1;
				}
				if ( $rest ne "" ) {
					return manifest_error("Standard expected near '$rest' in '$line'.");
				}
				( $standard )
					or return manifest_error("Standard expected.");
				
				$descr_obj = { STANDARD => $standard };
				push @{$manifest->{DESCRIPTIONS}{$testsuite}}, $descr_obj;
			}
			else {
				($descr_obj) or return manifest_error("'TESTSUITE test_name ...' expected before line '$line'");
				
				if ( $line =~ /^DISPLAYNAME\s*=\s*(.*)$/ ) {
					$descr_obj->{DISPLAYNAME} = $1;
				}
				elsif ( $line =~ /^DESCRIPTION\s*=\s*(.*)$/ ) {
					$descr_obj->{DESCRIPTION} = $1;
				}
				else {
					if ( defined $descr_obj->{DESCRIPTION} ) {
						# Append to the description
						$descr_obj->{DESCRIPTION} .= " ".$line;
					}
					else {
						warning manifest_error("Unhandled text: '$line'.");
						next;
					}
				}
			}
		}
		elsif ( $section eq 'OPTIONS' ) {
		# [OPTIONS]
			if ( $line =~ /^TESTSUITE\s*END$/) {
				$testsuite = undef;
				$test_ver  = undef;
				$standard  = undef;
				$opt_obj   = undef;
			}
			elsif ( $line =~ /^TESTSUITE\s+($re_testsuite)\s*(.*)$/ ) {
			# TESTSUITE test_name [ver (<=|<|=|>=|>) $test_ver] [Standard (<=|<|=|>=|>) $ver]
				$testsuite = $1;
				my $rest = $2;
				
				$test_ver = undef;
				$standard = undef;
				
				if ( $rest =~ s/^VER\s*(($re_verfilt\s*)+(\s*\|\s*($re_verfilt\s*)+)*)//i ) {
					$test_ver = $1;
					$test_ver =~ s/\s+$//; # trim spaces
				}
				if ( $rest =~ s/^(\w+\s*($re_verfilt\s*)*(\s*\|\s*\w+\s*($re_verfilt\s*)*)*)//i ) {
					$standard = $1;
				}
				if ( $rest ne "" ) {
					return manifest_error("Test version and standard are expected near '$rest' in '$line'.");
				}
			}
			elsif ( $line =~ /^(ID|Q|GLOBAL|PRE|TYPE|DEFAULT):\s*/ ) {
				( defined $testsuite ) or return manifest_error("Testsuite isn't defined here.");
				
				my $s = $line;
				
				if ( $s =~ s/^(ID|GLOBAL):\s*// ) {
					my $is_global = ($1 eq "GLOBAL");
					
					$opt_obj = { ID => $s };
					$opt_obj->{TEST_VER} = $test_ver if defined $test_ver;
					$opt_obj->{STANDARD} = $standard if defined $standard;
					$opt_obj->{GLOBAL} = 1 if $is_global;
					
					push @{$manifest->{OPTIONS}{$testsuite}}, $opt_obj;
				}
				elsif ( $s =~ s/^(Q|PRE|DEFAULT):\s*// ) {
					( $opt_obj ) or return manifest_error("'ID:' expected before line '$line'");
					
					my $KEY = uc $1;
					$opt_obj->{$KEY} = $s;
				}
				elsif ( $s =~ s/^(TYPE):\s*// ) {
					( $opt_obj ) or return manifest_error("'ID:' expected before line '$line'");
					
					if ( $s =~ /^(string)$/i ) {
						$opt_obj->{TYPE} = lc $s;
						$opt_obj->{VARIANTS} = undef;
					}
					elsif ( $s =~ /^(bool|enum)/i ) {
						my $type = lc $1;

						if ( $s =~ s/\s*{([^}]*)}$// ) {
							foreach ( split /\s*\|\s*/, $1 ) {
								my ( $option, $display) = split /\s*=\s*/, $_;
								push @{$opt_obj->{VARIANTS}}, $option;
								
								if ( !defined $display) { $display = $option; }
								$opt_obj->{DISPLAY}{$option} = $display;
							}
						}
						else {
							return manifest_error("Wrong enum: '$s' in '$line'");
						}
						
						$opt_obj->{TYPE} = $type;
						
						if ( $type eq 'bool' && scalar @{$opt_obj->{VARIANTS}} != 2) {
							return manifest_error("Bool type requires two values here: '$line'");
						}
					}
					else {
						return manifest_error("Unknown type: '$s' in '$line'");
					}
				}
			}
			else {
				warning manifest_error("Can't understand the line: '$line'.");
				next;
			}
		}
		elsif ( $section eq 'MANUAL' ) {
		# [MANUAL]
			if ( $line =~ /^TESTSUITE\s*END$/) {
				$testsuite = undef;
				$standard = undef;
			}
			elsif ( $line =~ /^TESTSUITE\s+($re_testsuite)\s*(.*)$/ ) {
			# TESTSUITE test_name [Standard (<=|<|=|>=|>) $ver]
				$testsuite = $1;
				my $rest = $2;
				
				$standard = undef;
				
				if ( $rest =~ s/^(\w+\s*($re_verfilt\s*)*(\s*\|\s*\w+\s*($re_verfilt\s*)*)*)//i ) {
					$standard = $1;
				}
				if ( $rest ne "" ) {
					return manifest_error("Standard expected near '$rest' in '$line'.");
				}
			}
			elsif ( $line =~ /^STEP\s+(OPTIONAL\s+)?(.*)/ ) {
				( defined $testsuite ) or return manifest_error("Testsuite isn't defined here.");
				my $optional = $1;
				my $s = $2;
				
				my $step_obj = {};
				$step_obj->{STEP} = $s;
				$step_obj->{STANDARD} = $standard if defined $standard;
				$step_obj->{OPTIONAL} = 1 if $optional;

				push @{$manifest->{MANUAL}{$testsuite}}, $step_obj;
				
				$all_testsuites->{$testsuite} = 1;
			}
			else {
				my $last_step = $manifest->{MANUAL}{$testsuite}[-1];
				if ( $last_step ) {
					$last_step->{STEP} .= "\n".$line;
				} else {
					warning manifest_error("Can't understand the line: '$line'.");
					next;
				}
			}
		}
		elsif ( $section eq 'PROFILES' ) {
		# [PROFILES]
			# Display names for profiles:
			if ( $line =~ /^PROFILE\s+($re_standard)(\s+DEFAULT)?\s+([^=\s]+)\s*=\s*(.*)/ ) {
				my $standard = $1;
				my $is_default = $2;
				my $profile_id = $3;
				my $profile_name = $4;
				
				( $profile_id ) or warning(manifest_error("Profile id wasn't specified in line '$line'")), next;
				( $profile_name ) or warning(manifest_error("Profile name wasn't specified in line '$line'")), next;
				
				$standard =~ s/^([a-z]+)\s*(\S*)$/$1 $2/i; # "LSB3.1" -> "LSB 3.1"
				
				# ->{STD_PROFILES}{$standard}{DISPLAY_NAME}{$profile_id} = "Display name (html)";
				#                            {DEFAULT} = $profile_id;
				#                            {TESTS}{$testsuite}{$profile_id} = 1;
				( !defined $manifest->{STD_PROFILES}{$standard}{DISPLAY_NAME}{$profile_id} )
					or warning manifest_error("Profile id was specified twice");
				$manifest->{STD_PROFILES}{$standard}{DISPLAY_NAME}{$profile_id} = $profile_name;
				if ( $is_default ) {
					$manifest->{STD_PROFILES}{$standard}{DEFAULT} = $profile_id;
				}
			}
			# Test suite to profile references:
			elsif ( $line =~ /^($re_testsuite)\s+($re_standard)(?:\s*\:\s*(.*))?/i ) {
				my $testsuite = $1;
				my $standard = $2;
				my $profiles = $3;

				( $standard ) or warning(manifest_error("Standard wasn't specified in line '$line'")), next;
				
				$standard =~ s/^([a-z]+)\s*(\S*)$/$1 $2/i; # "LSB3.1" -> "LSB 3.1"

				if ( $profiles ) {
					foreach my $profile ( split /[\|;]/, $profiles ) {
						( !defined $manifest->{STD_PROFILES}{$standard}{TESTS}{$testsuite}{$profile} )
							or warning manifest_error("Profile was specified twice");
						
						$manifest->{STD_PROFILES}{$standard}{TESTS}{$testsuite}{$profile} = 1;
					}
				}
			}
			else {
				warning manifest_error("Can't understand the line: '$line'.");
				next;
			}
		}
		elsif ( $section eq 'MODULES' ) {
		# [MODULES]
			if ( $line =~ /^TESTSUITE\s*END$/) {
				$testsuite = undef;
				$test_ver  = undef;
				$standard  = undef;
			}
			elsif ( $line =~ /^TESTSUITE\s+($re_testsuite)\s*(.*)$/ ) {
			# TESTSUITE test_name [ver (<=|<|=|>=|>) $test_ver] [Standard (<=|<|=|>=|>) $ver]
				$testsuite = $1;
				my $rest = $2;
				
				$test_ver = undef;
				$standard = undef;
				
				if ( $rest =~ s/^VER\s*(($re_verfilt\s*)+(\s*\|\s*($re_verfilt\s*)+)*)//i ) {
					$test_ver = $1;
					$test_ver =~ s/\s+$//; # trim spaces
				}
				if ( $rest =~ s/^(\w+\s*($re_verfilt\s*)*(\s*\|\s*\w+\s*($re_verfilt\s*)*)*)//i ) {
					$standard = $1;
				}
				if ( $rest ne "" ) {
					return manifest_error("Test version and standard are expected near '$rest' in '$line'.");
				}
			}
			elsif ( $line =~ /^(module|requires)\s+(\S+)(?:\s+from\s+(.*))?/i ) {
				( defined $testsuite ) or return manifest_error("Testsuite isn't defined here.");
				
				my $mod = $1;
				my $mod_file = $2;
				my $from = $3;
				
				my $obj = { FILE => $mod_file };
				$obj->{MAIN} = 1 if $mod eq 'module';
				$obj->{FROM} = $from if defined $from;
				$opt_obj->{TEST_VER} = $test_ver if defined $test_ver;
				$opt_obj->{STANDARD} = $standard if defined $standard;
				
				push @{$manifest->{MODULES}{$testsuite}}, $obj;
			}
			else {
				warning manifest_error("Can't understand the line: '$line'.");
				next;
			}
		}
		elsif ( $section eq 'COMPOSITION' ) {
		# [COMPOSITION]
			if ( $line =~ /^TESTSUITE\s+($re_testsuite)\s+($re_version|\*)\s+($re_standard)\s+((([,]|\b)\s*($re_arch)\s*)*\s+|\*\s+|)($re_status)?$/ ) {
			# TESTSUITE  test_name  test_ver
				$testsuite = $1;
				$test_ver = $2;
				$standard = $3;
				$arch = $4; # Can be a set of architectures or '*'
				if ( !defined $arch || $arch eq "") { $arch = '*' };
				$status = $8;
				
				( defined $status ) or return manifest_error("Status undefined.");
				
				$standard =~ s/^([a-z]+)\s*(\S*)$/$1 $2/i; # "LSB3.1" -> "LSB 3.1"
				
				$all_standards->{$standard} = 1;
				$all_testsuites->{$testsuite} = 1;
				$all_statuses->{$status} = 1;
				
				# Save
				my @archs;
				if ( $arch =~ /^\s*\*\s*$/ ) {
					@archs = @all_archs;
					$arch = join ",", @archs;
				} else {
					@archs = ();
					foreach my $a ( split /[ ,]/, $arch ) {
						$a = trim($a); # Trim spaces
						next if $a eq "";
					
						( in_array($a, @all_archs) )
							or return manifest_error("Wrong arch: '$arch'");
						push @archs, $a;
					}
					$arch = join ",", @archs;
				}
				foreach my $arch ( @archs ) {
					$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver} ||= [];
					$all_architectures->{$arch} = 1;
				}
			}
			elsif ( $line =~ /^($re_file)(?:\s+($re_verfilt|\*))?(?:\s+([^\d]\S*)(?:\s+(\S+))?)?$/ ) {
			# Contents of the test suite
				my $package_name = $1;
				my $package_ver  = $2; # version specification, i.e <=1.2; may be blank (undef)
				my $location     = $3; # file location, may be undef
				my $tag         = $4; # additional tags, usually undef
				
				if ( !$testsuite || !defined $test_ver || !defined $standard 
					|| !$status || !defined $arch )
				{
					return manifest_error("Something goes wrong. My be a TESTSUITE line was missed.");
				}
				
				my $packref_obj = {
						PACKAGE => $package_name, 
						PACKAGE_VER => $package_ver,
						LOCATION => $location,
					};
				$packref_obj->{TAG} = $tag if $tag;
				
				# Save
				foreach my $arch ( split /,/, $arch ) {
					push @{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}}, $packref_obj;
				}
			}
			else {
				warning manifest_error("Can't understand this line:\n '$line'.");
				next;
			} # end of  if ( $line =~ /^TESTSUITE....) { ... } else ....
		} # end of the [COMPOSITION] section
		elsif ( $section eq 'PACKAGES' ) {
			if ( $line =~ /^CONFLICTS (($re_package)\s*(,\s*($re_package)\s*)*)$/ ) {
				my $packages = $1;
				
				my @packages = split /\s*,\s*/, $packages;
				foreach my $p1 ( @packages ) { # each
					foreach my $p2 ( @packages ) { # with each
						next if $p1 eq $p2;
					
						$manifest->{CONFLICTS}{$p1}->{$p2} = 1;
					}
				}
			}
			else {
				warning manifest_error("Can't understand this line:\n '$line'.");
				next;
			}
		}
		elsif ( $section eq 'SERVERS' ) {
		# [SERVERS]
			# DIR ftp.linuxfoundation.org/
			if ( $line =~ /^DIR\s+($re_path)$/ ) {
				$servers_dir = $1;
				
				# Append a slash to avoid matching part of directory name
				$servers_dir .= "/" if $servers_dir !~ /\/$/;
			}
			elsif ( $line =~ /^($re_link)$/ ) {
				my $link = $1;
				
				( defined $servers_dir ) or return manifest_error("DIR undefined near '$line'.");
				
				$link .= "/" if $link !~ /\/$/;
				
				$manifest->{SERVERS}->{$servers_dir}{$link} = 1;
			}
			else {
				warning manifest_error("Can't understand this line:\n '$line'.");
				next;
			}
		}
		elsif ( $section eq 'FILES' ) {
		# [FILES]
			# ALIAS runtime-3.1 = http://...
			if ( $line =~ /^ALIAS\s+($re_alias)\s*=\s*(.*)$/ ) {
				my $name = $1;
				my $text = $2;
				
				( !defined $manifest->{ALIASES}{$name} )
					or warning manifest_error("Attempt to redefine alias '$name'.");
					
				$manifest->{ALIASES}{$name} = $text;
				
				my $location = expand_aliases($text, $manifest->{LOCATIONS}, undef);
				is_ok($location) or return manifest_error("Error in: '$line'", $Error::Last);
				
				( $location !~ /<(?!arch)/ )
					or return manifest_error("Failed to expand alias '$text': '$location'");
				
				( !defined $manifest->{LOCATIONS}{$name} )
					or warning manifest_error("Attempt to redefine location '$name'.");
					
				$manifest->{LOCATIONS}{$name} = $location;
			}
			else {
				# filename | location | size [timestamp]
				#       [ arch [ package kind | package name | package version ]]
				my @line_spl = split /\s+/, $line;
				
				my $filename = shift @line_spl;
				my $location = shift @line_spl;
				my $removed = 0;
				if ( $line_spl[0] eq 'rm' ) {
					$removed = shift @line_spl;
				}
				my $size     = shift @line_spl;
				my $timestamp = undef;
				if ( $line_spl[0] =~ /^t(\d+)$/ ) { # timestamp
					$timestamp = $1;
					shift @line_spl;
				}
				$arch = shift @line_spl;
				if ( $arch eq '*' ) { $arch = 'noarch' } # this is for compatibility with previous manifest version
				
				( defined $arch )
					or return manifest_error("Arch is not defined");
				
				my $package_kind = (shift @line_spl or "file");
				my $package_name = (shift @line_spl or $filename);
				my $package_version = (shift @line_spl or "");
				my $null = shift @line_spl;
					
				if ( !defined $filename || $filename !~ /^[^\/\*\+]+$/ ) {
					return manifest_error("Wrong file name: '$filename'.");
				}
			   
				my $file = {};
				
				$file->{FILENAME} = $filename;
				
				$file->{LOCATION} = $location;
				
				my $src = expand_aliases( $location, $manifest->{LOCATIONS}, $arch);
				# Some checks
				is_ok($src) or return manifest_error("Error in '$line'", $Error::Last);
				( $src !~ /</ )  or return manifest_error("Failed to expand aliases: '$location'");
				( $src =~ /^$re_path$/ ) or return manifest_error("Wrong src: '$src' ($location)");
				
				$file->{SRC} = $src;
				
				$file->{REMOVED} = $removed;
				
				if ( defined $size && $size !~ /^[\d]+$/ ) {
					return manifest_error("Wrong size: '$size'.");
				}
				$file->{SIZE} = $size;
				$file->{TIMESTAMP} = $timestamp if defined $timestamp;
				
				if ( !defined $arch || $arch !~ /^(\*|noarch|$re_arch)$/i ) {
					return manifest_error("Wrong arch: '$arch'.");
				}
				$file->{ARCH} = $arch;
				
				if ( !defined $package_kind || $package_kind !~ /^(rpm|deb|file)$/i ) {
					return manifest_error("Wrong package kind: '$package_kind'.");
				}
				if ( $package_kind eq "rpm" || $package_kind eq "deb" ) {
					if ( !defined $package_name || $package_name !~ /^$re_package$/ ) {
						return manifest_error("Wrong package name: '$package_name'.");
					}
					if ( !defined $package_version || $package_version !~ /^$re_version$/ ) {
						return manifest_error("Wrong package version: '$package_version'.");
					}
				}
				$file->{KIND} = $package_kind;
				$file->{PACKAGE} = $package_name;
				$file->{VERSION} = $package_version;

				if ( defined $null ) {
					return manifest_error("Unexpected text: '$null'.");
				}
				
				# Add the file
				$manifest->{FILES}{$src."/".$filename} = $file;
			}
		}
		else {
			warning manifest_error("Unknown section name '$section'.");
			do { $line = <FILE> } while ( $line && $line !~ /^\[/ );
			last if !$line;
			redo;
		}
	}
	
	# Build PACKAGES hash
	foreach my $key ( keys %{$manifest->{FILES}} ) {
		my $package = $manifest->{FILES}{$key}{PACKAGE} or next;
		push @{$manifest->{PACKAGES}{$package}{FILES}}, $key;
	}
	
	$manifest->{ALL}{STANDARDS} = [reverse sort keys %$all_standards];
	$manifest->{ALL}{TESTSUITES} = [sort keys %$all_testsuites];
	$manifest->{ALL}{ARCHITECTURES} = [sort {sort_cat($a,$b,\@all_archs)} keys %$all_architectures];
	$manifest->{ALL}{STATUSES} = [sort keys %$all_statuses];
	# Keep in sync with merge_manifests()!
	
	return $manifest;
}
#----------------------------------------------------------------------

sub expand_aliases {
	my ($s, $aliases, $arch) = @_;
	
	while ( $s =~ /<([^>]+)>/g ) {
		my $pattern = $1;
		if ( $pattern eq 'arch' ) {
			next if !defined $arch;
			$s =~ s/<$pattern>/$arch/;
		}
		elsif ( $pattern =~ /^arch\d+$/ ) {
			next if !defined $arch;
			
			my $alias = $aliases->{$pattern};
			( defined $alias )
				or return error "Undefined alias: '$pattern' in '$s'.";
			
			my %arch_table = (split /,|;|=>/, $alias);
			my $replacement = $arch_table{$arch}
				or return error "Undefined architecture alias for '$arch' in '$alias' (<$pattern>).";
			
			$s =~ s/<$pattern>/$replacement/;
		}
		else {
			my $replacement = $aliases->{$pattern};
			( defined $replacement )
				or return error "Undefined alias: '$pattern' in '$s'.";
			
			$s =~ s/<$pattern>/$replacement/;
		}
	}
	return $s;
}
#----------------------------------------------------------------------

sub not_found_error {
	my ($msg, $files_arr) = @_;
	
	my $txt = $msg;
	if ( $files_arr ) {
		$txt .= " among:";
		foreach my $file ( @$files_arr ) {
			$txt .= "\n      ".$file->{FILENAME}." ".$file->{LOCATION}."";
		}
	}
	return error $txt;
};

sub select_file {
	my ($manifest, $arch, $package_manager, $req) = @_;
	
	#( in_array($arch, @all_archs) ) or return error "wrong arch '$arch'";

	my $package_ver = ($req->{PACKAGE_VER} or "*");
	
	my $src = "";
	if ( $req->{LOCATION} ) {
		# Expand <arch> templates, since $arch is known here
		$src = expand_aliases($req->{LOCATION}, $manifest->{LOCATIONS}, $arch);
		is_ok($src) or return error "failed to expand aliases", $Error::Last;
		( $src !~ /[<>]/ ) or return error "failed to expand aliases: '$req->{LOCATION}' -> '$src'";
	}
	
	## Manifest structure:
	## ->{FILES}{"$src/$filename"} = $file->{FILENAME|LOCATION|SRC|SIZE|TIMESTAMP|ARCH|KIND|PACKAGE|VERSION}
	
	my $last_error;
	
	my $files_arr = [map {$manifest->{FILES}{$_}} @{$manifest->{PACKAGES}{$req->{PACKAGE}}{FILES}}];
	
	if ( @$files_arr && $package_ver ne "*" ) {
		my $tmp = [grep { defined $_->{VERSION} && version_filter($package_ver, $_->{VERSION}) } @$files_arr];
		( @$tmp ) or return not_found_error("Can't find file '$req->{PACKAGE}' v. '$package_ver'", $files_arr);
		$files_arr = $tmp;
	}
	if ( @$files_arr && $arch ) {
		my $tmp = [grep { $_->{ARCH} eq $arch || $_->{ARCH} eq 'noarch' } @$files_arr];
		( @$tmp ) or return not_found_error("Can't find file '$req->{PACKAGE}' v. '$package_ver' '$arch'", $files_arr);
		$files_arr = $tmp;
	}
	if ( @$files_arr && $src ne "" ) {
		my $tmp = [grep { $_->{SRC} eq $src  } @$files_arr];
		( @$tmp ) or return not_found_error( "Can't find file '$req->{PACKAGE}' v. '$package_ver' '$arch'".($req->{LOCATION} ? ", in '$req->{LOCATION}'": ""), $files_arr);
		$files_arr = $tmp;
	}

	if ( @$files_arr ) {
		if ( $package_manager eq 'rpm' ) {
			@$files_arr = grep {$_->{KIND} eq 'rpm' || $_->{KIND} eq 'file'} @$files_arr;
		}
		elsif ( $package_manager eq 'dpkg' ) {
			my $tmp = [grep {$_->{KIND} eq 'deb' || $_->{KIND} eq 'file'} @$files_arr];
			if ( @$tmp ) {
				$files_arr = $tmp;
			} else {
				# If there are no deb packages, try rpm packages (they can be installed via 'alien').
				@$files_arr = grep {$_->{KIND} eq 'rpm' || $_->{KIND} eq 'file'} @$files_arr;
			}
		}
	}

	if ( !@$files_arr ) {
		return error "Can't find file '$req->{PACKAGE}' v. '$package_ver',"
				." arch: '$arch'".($req->{LOCATION} ? ", in '$req->{LOCATION}'": "");
	}

	return $files_arr;
}
#----------------------------------------------------------------------

# Returns hyperlink for downloading the $file ($file is a manifest object).
# The $i parameter is used to select a mirror.
sub file_link {
	my ($manifest, $file, $dl_method, $i) = @_;
	$i = 0 if !defined $i;
	
	my $fpath = $file->{SRC}."/".$file->{FILENAME};
	
	my $subst = undef;
	my $dir_matched;
	foreach my $dir ( reverse sort keys %{$manifest->{SERVERS}} ) {
	# Reverse sorting to match longer patterns first
		
		# $fpath should start with $dir
		next if index($fpath, $dir) != 0;

		my @mirrors = keys %{$manifest->{SERVERS}{$dir}};
		
		# Sort mirrors;
		my @ftp_mirrors = grep {/^s?ftp:/} @mirrors;
		my @http_mirrors = grep {/^https?:/} @mirrors;
		
		if ( defined $dl_method && $dl_method eq "ftp" ) {
			@mirrors = (@ftp_mirrors, @http_mirrors);
		} else {
			@mirrors = (@http_mirrors, @ftp_mirrors); # default
		}
		next if !@mirrors;
		
		if ( $#mirrors < $i ) { return error "No more mirrors"; }
		
		$subst = $mirrors[$i];
		$dir_matched = $dir;
		last;
	}
	
	if ( !defined $subst ) { # No wariants =(
		return error "URL is unknown for '$fpath'";
	}
	
	substr($fpath, 0, length($dir_matched)) = $subst; # replace /^$dir_matched/ with $subst
	
	return $fpath;
}
#----------------------------------------------------------------------

# Check if $file matches the expected parameters in the $file_obj structure.
# That are: size (SIZE) and modification time (TIMESTAMP).
# Returns an error if the file doesn't match.
# Returns $ret_code instead of an error if $ret_code is specified (it's faster).
sub check_file {
	my ($file, $file_obj, $ret_code) = @_;
	
	my ($size, $mtime) = file_info($file);
	if ( !defined $size ) {
		return $ret_code if defined $ret_code;
		return error "Failed to stat file '$file'";
	}
	
	if ( defined $file_obj->{SIZE} && $file_obj->{SIZE} != 0 ) { # Treat 0 as unknown file size
	# Check the file size
		if ( $size != $file_obj->{SIZE} ) {
			my $filename = extract_filename($file);
			return $ret_code if defined $ret_code;
			return error "$filename: file size doesn't match: $size != ".$file_obj->{SIZE};
		}
	}
	
	if ( defined $file_obj->{TIMESTAMP} ) {
	# Check the timestamp
		if ( $mtime != $file_obj->{TIMESTAMP} ) {
			my $filename = extract_filename($file);
			return $ret_code if defined $ret_code;
			return error "$filename: file timestamp doesn't match: $mtime != ".$file_obj->{TIMESTAMP};
		}
	}
	
	return 1; # OK
}
#-----------------------------------------------------------------------

# Check files in cache.
# Set the $file->{DOWNLOADED} field for each file in array @files:
#  -1: the file exists but is invalid (see check_file for criteries).
#   1: the file exists and is valid.
#   0: the file is absent.
sub check_files {
	my (@files) = @_;
	
	foreach my $file ( @files ) {
		my $file_in_cache = cache_dest($file->{SRC})."/".$file->{FILENAME};
		if ( -f $file_in_cache ) {
			if ( check_file($file_in_cache, $file, 0) ) {
				$file->{DOWNLOADED} = 1;
			} else {
				$file->{DOWNLOADED} = -1;
			}
		}
		else {
			$file->{DOWNLOADED} = 0;
		}
	}
	
	return 1;
}
#-----------------------------------------------------------------------

sub main_module_file {
	my ($manifest, $test_name, $test_ver, $standard) = @_;

	( $test_ver ) or return error "Test version isn't defined";
	( $standard ) or return error "Standard isn't defined";

	my @modules = grep {
			   version_filter($_->{TEST_VER}, $test_ver)
			&& version_filter($_->{STANDARD}, $standard)
		} @{$manifest->{MODULES}{$test_name}};
	
	my ($main_module) = grep $_->{MAIN}, @modules;
	( $main_module ) or return error "No module specifed in Manifest for '$test_name' ver $test_ver, $standard";
	
	return $main_module->{FILE};
}
#-----------------------------------------------------------------------

sub load_manifest {
	my (%sections) = @_;
	my $manif_file = $MANIFEST_DIR."/Manifest"; # The main Manifest file.
	
	( -f $manif_file ) or return error "Can't find Manifest file: '$manif_file'.";
	
	# Load the main manifest file
	debug_inform("Loading manifest '$manif_file'");
	my $manifest = manifest_parse( $manif_file, %sections );
	is_ok($manifest) or return error "Failed to parse Manifest file: '$manif_file'.", $Error::Last;
	
	# Additional manifests: *.mnf files
	for $manif_file ( glob($MANIFEST_DIR."/*.mnf") ) {
		debug_inform("Loading manifest '$manif_file'");
		my $manifest_add = manifest_parse( $manif_file, %sections );
		if ( !is_ok($manifest_add) ) {
			complain error "Failed to parse Manifest file: '$manif_file'. It will be skipped.", $Error::Last;
			next;
		}
		my $manifest_tmp = copy($manifest); # temporary copy
		if ( !is_ok merge_manifests($manifest, $manifest_add) ) {
			$manifest = $manifest_tmp; # undo merge
			complain error "Failed to use file '$manif_file'", $Error::Last;
			next;
		}
	}
	
	if ( !$manifest->{MANIFEST}{TIMESTAMP} ) {
		$manifest->{MANIFEST}{TIMESTAMP} = time();
	}
	
	return $manifest;
}
#----------------------------------------------------------------------

sub merge_manifests {
	my ($manifest, $manifest2) = @_;
	
	delete $manifest->{PACKAGES};
	delete $manifest2->{PACKAGES};
	
	is_ok merge_hash ( $manifest, $manifest2, 'overwrite_scalar' => 1, 'overwrite_array' => 1 ) 
		or return error "Failed to merge manifests", $Error::Last;
	
	# TODO: manually merge OPTIONS: add options from the $manifest that don't match any test in $manifest2.
	
	# Rebuild packages
	$manifest->{PACKAGES} = {};
	foreach my $key ( keys %{$manifest->{FILES}} ) {
		my $package = $manifest->{FILES}{$key}{PACKAGE} or next;
		push @{$manifest->{PACKAGES}{$package}{FILES}}, $key;
	}
	
	# Rebuild {ALL}
	my $all_standards = {};
	my $all_testsuites = {};
	my $all_architectures = {};
	my $all_statuses = {};
	foreach my $standard ( keys %{$manifest->{TREE}} ) {
		$all_standards->{$standard} = 1;
	foreach my $testsuite ( keys %{$manifest->{TREE}{$standard}} ) {
		$all_testsuites->{$testsuite} = 1;
	}}
	# ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{...}
	foreach my $standard ( keys %{$manifest->{COMPOSITION}} ) {
		$all_standards->{$standard} = 1;
	foreach my $testsuite ( keys %{$manifest->{COMPOSITION}{$standard}} ) {
		$all_testsuites->{$testsuite} = 1;
	foreach my $arch ( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}} ) {
		$all_architectures->{$arch} = 1;
	foreach my $status ( keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}} ) {
		$all_statuses->{$status} = 1;
	}}}}
	$manifest->{ALL}{STANDARDS} = [reverse sort keys %$all_standards];
	$manifest->{ALL}{TESTSUITES} = [sort keys %$all_testsuites];
	$manifest->{ALL}{ARCHITECTURES} = [sort {sort_cat($a,$b,\@all_archs)} keys %$all_architectures];
	$manifest->{ALL}{STATUSES} = [sort keys %$all_statuses];
	
	return 1; # OK
}
#-----------------------------------------------------------------------

#----------------------------------------------------------------------
1; # Return value;
