# Distribution Checker
# Apache FVT Module (Apache_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Apache_test;
use strict;

use Misc;
use Download;
use SimpleJournal;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

sub check_port_80 {
	does_shell_know("netstat")
		or return error "No 'netstat' command";
	
	my $result = capture ("netstat -an 2>/dev/null | grep -e ':80 \\+[0-9:]'");
	if ( $result =~ /\busage:/m ) {
		return error "netstat failed", $Error::Last;
	}
	# try `netstat -anp`, ignore the result if error
	my $result_ext = capture ("netstat -anp 2>/dev/null | grep -e ':80 \\+[0-9:]'");
	if ( $result_ext && ( $result_ext !~ /\busage:/m ) ) {
		$result = $result_ext;
	}
	
	return $result;
}

sub stop_httpd {
	my ($process) = @_;
	
	if ( !defined $process ) {
		# Look for httpd process (can be named differently)
		for my $name ("apache", "apache2", "httpd", "lsb-apache") {
			if ( cmd("ps -A | grep -e '\\<".$name."\$'") == 0 ) {
				$process = $name; # Gotcha!
				last;
			}
		}
	}
	( $process ) or return ""; # Return ok if process not found
	
	# Trying to stop the daemon gracefully
	my $script = "";
	for my $scriptname ($process, "apache", "apache2", "httpd", "lsb-apache") {
		if ( -f "/etc/init.d/$scriptname" ) {
			inform("Stopping $scriptname daemon");
			cmd("/etc/init.d/$scriptname stop");
			sleep 2; # Wait a bit allowing the daemon to stop completely.
			
			if ( cmd("ps -A | grep -e '\\<".$process."\$'") != 0 ) {
				$script = $scriptname;
				last;
			}
		}
	}

	# Check the result
	return $script if cmd("ps -A | grep -e '\\<".$process."\$'") != 0;
	
	# Not stopped. Kill the process.
	cmd ("killall $process");
	sleep 2; # Wait a bit allowing apache to stop completely.
	
	# Check the result
	return $script if cmd("ps -A | grep -e '\\<".$process."\$'") != 0;
	
	# Still not stopped. Use kill -9
	cmd ("killall -9 $process");
	sleep 2; # Wait a bit allowing apache to stop completely.
	
	# Check the result
	if ( cmd("ps -A | grep -e '\\<".$process."\$'") == 0 ) {
		# Still not stopped.
		return error "Failed to stop '$process'";
	}
	
	return $script;
}

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;

	# Check whether port 80 is free
	my $result = check_port_80();
	if ( !is_ok($result) || !defined $result ) {
	# Can't identify whether the port 80 is free.
		$self->report_warning(warning "Can't identify whether the port 80 is free. Please, make sure that no one process is listening on the port 80. If any, stop it and restart the test.");
	}

	return 1;
}

sub prepare {
	my ($self) = @_;

	# 1. Download the lsb-apache application package that corresponds to the LSB
	#    version and processor architecture for which certification is being 
	#    sought from the Download Page. 
	# 2. Install package on system.
	# NOTE: Application MUST be installed without use of --nodeps option,
	#  the dependency checks are part of the test.
	
	# Download and install required packages
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory", $Error::Last;
	
	# Run the command: grep "^nobody" /etc/passwd 
	# Expected result: Output shows user "nobody" already exists. 
	my $result = capture('grep -e "^nobody\b" /etc/passwd');
	if ( $result =~ /^\s*$/ ) {
		# If it is not present on the system, run the command: "useradd nobody" 
		# to create the nobody user. 
		cmd("useradd nobody") == 0
			or return error "Failed to add user 'nobody'";
	}
	
	# Run the command: grep "^nobody" /etc/group 
	# Expected result: Output shows group "nobody" already exists. 
	$result = capture('grep -e "^nobody\b" /etc/group');
	if ( $result =~ /^\s*$/ ) {
		# If it is not present on the system, run the command: "groupadd nobody" 
		# to create the nobody group. 
		cmd("groupadd nobody") == 0
			or return error "Failed to add group 'nobody'";
	}
	
	# Run the command: grep "^http" /etc/services
	# Expected result: Output shows tcp and udp port definitions for the http
	# server using port 80. 
	$result = capture('grep -e "^http\b" /etc/services');
	if ( $result =~ /^\s*$/ ) {
		# If definitions are not present, perform a second 
		# check: grep "^www" /etc/services. Some systems name the entry one way,
		# some the other (either is correct, as IANA provides both in the 
		# canonical list). 
		$result = capture('grep -e "^www\b" /etc/services');
	}
	if ( $result =~ /^\s*$/ ) {
		# In the unlikely event that there is no services entry for a webserver
		# on port 80, add the following two lines to /etc/services: 
		#   http     80/tcp
		#   http     80/udp
		is_ok append_string_to_file( "/etc/services", <<HEREDOC ) or return error "Failed to write to /etc/services", $Error::Last;
http     80/tcp
http     80/udp
HEREDOC
	}
	
	$self->{JOURNAL_FILE} = $self->name.".simplejournal";
	$self->{JOURNAL_TYPE} = 'SIMPLE';
	
	# If the system under test is already running a webserver listening on the
	# standard port (80), stop it before proceeding with the FVT. 
	
	# Check whether port 80 is free
	$result = check_port_80();
	if ( is_ok($result) && $result ) {
		# The port is occupied
		my ($process_name);

		if ( $result =~ m/\s+(\d+)\/(\S+)$/ ) {
			#$pid = $1;
			$process_name = $2;
		}
		else {
			# Can't determine what process uses port 80, but most likely it's httpd.
		}

		# Trying to stop httpd
		$self->{'stopped_script'} = stop_httpd($process_name);
		( is_ok $self->{'stopped_script'} )
			or return error "Failed to stop Apache server daemon", $Error::Last;
		
		sleep(1); # Sleep a bit, to let the 80 port to be released.
		
		# Check again
		$result = check_port_80();
	}
	if ( is_ok($result) && $result ) {
		return error $self->name." test requires port 80 to be free."
				."\nPlease, stop the process listening on the port 80."
				."\nMore info: $result";
	}

	return 1;
}

sub run {
	my ($self) = @_;
	
	my $result;
	
	my $simpjournal = SimpleJournal->new( $self->journal_file );
	is_ok $simpjournal or return error "Failed to open journal.", $Error::Last;
	
	$simpjournal->test_name( $self->name );
	
	### Functional Verification Test for lsb-apache ###
	# http://www.linuxfoundation.org/appbat/fvt-3.0.0/lsb-apache_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.1.0/lsb-apache_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.2.0/lsb-apache_fvt.html
	
	###  Check installation
	$simpjournal->test_case_start("Check installation");
	
	$simpjournal->test_description(<<HEREDOC);
- Expected result: application should be installed to 
/opt/lsb/appbat
/var/opt/lsb/appbat
/etc/opt/lsb/appbat
HEREDOC
	
	$result = "passed";
	for my $file ( "/etc/opt/lsb/appbat/httpd.conf",
				"/opt/lsb/appbat/sbin/httpd",
				"/var/opt/lsb/appbat/run",
				"/opt/lsb/appbat/lib/pkgconfig" )
	{
		if ( !-e $file ) {
			$result = "failed";
			$simpjournal->test_comment("'$file' is expected to exist after installation, but not found.");
		}
	}
	
	$simpjournal->test_case_result($result);
	#---------
	
	###  Check init.d command
	if ( version_filter("LSB 3.2", $globals->{'standard'}) ) {
	# LSB 3.2 only
		my $test_descr = <<HEREDOC;
The lsb-apache package utilizes the install_initd/remove_initd commands. 
After the package install, lsb-apache should be listed as a system service.
- Expected result: lsb-apache should listed as enabled in runlevels 3,4,5
HEREDOC

		# To verify this, use one of the following commands: 
		#   chkconfig --list lsb-apache (Red Hat and other systems providing chkconfig)
		#   sysv-rc-conf --list lsb-apache (Debian and derivatives)
		#   ls /etc/rc[0-6].d/*lsb-apache (Debian and derivatives, if sysv-rc-conf is not installed)

		if (   (my $is_chkconfig    = does_shell_know("chkconfig"))
			|| (my $is_sysv_rc_conf = does_shell_know("sysv-rc-conf"))  ) 
		{
			my $msg = "";
			if ( $is_chkconfig ) {
				$msg = capture("LC_ALL=C chkconfig --list lsb-apache");
			}
			elsif ( $is_sysv_rc_conf ) {
				$msg = capture("LC_ALL=C sysv-rc-conf --list lsb-apache");
			}
			
			# Runlevels 0, 1 and 6 are usually disabled, but aren't prohibited.
			for my $runlevel ("0", "1", "6") {
				$simpjournal->test_case_start("Check init.d command - runlevel $runlevel");
				$simpjournal->test_description($test_descr);
				$simpjournal->test_message($msg);
				if ( $msg =~ /\b$runlevel:on\b/ ) {
					$simpjournal->test_comment("Warning: lsb-apache is enabled at runlevel $runlevel");
					$simpjournal->test_case_result("warning");
				} else {
					$simpjournal->test_case_result("passed");
				}
			}
			
			# Runlevel 2 is sometimes enabled and sometimes disabled.
			# Don't check. 
		   
			for my $runlevel ("3", "4", "5") {
				$simpjournal->test_case_start("Check init.d command - runlevel $runlevel");
				$simpjournal->test_description($test_descr);
				$simpjournal->test_message($msg);
				if ( $msg =~ /\b$runlevel:on\b/ ) {
					$simpjournal->test_case_result("passed");
				} else {
					$simpjournal->test_comment("lsb-apache isn't enabled at runlevel $runlevel");
					$simpjournal->test_case_result("failed");
				}
			}
		}
		else {
			my $msg = capture("ls /etc/rc[0-6].d/*lsb-apache");
			
			# Runlevels 0, 1 and 6 are usually disabled, but aren't prohibited.
			for my $runlevel ("0", "1", "6") {
				$simpjournal->test_case_start("Check init.d command - runlevel $runlevel");
				$simpjournal->test_description($test_descr);
				$simpjournal->test_message($msg);
				if ( $msg =~ /\brc${runlevel}\.d\b/ ) {
					$simpjournal->test_comment("Warning: lsb-apache is enabled at runlevel $runlevel");
					$simpjournal->test_case_result("warning");
				} else {
					$simpjournal->test_case_result("passed");
				}
			}
			
			# Runlevel 2 is sometimes enabled and sometimes disabled.
			# Don't check. 
		   
			for my $runlevel ("3", "4", "5") {
				$simpjournal->test_case_start("Check init.d command - runlevel $runlevel");
				$simpjournal->test_description($test_descr);
				$simpjournal->test_message($msg);
				if ( $msg =~ /\brc${runlevel}\.d\b/ ) {
					$simpjournal->test_case_result("passed");
				} else {
					$simpjournal->test_comment("lsb-apache isn't enabled at runlevel $runlevel");
					$simpjournal->test_case_result("failed");
				}
			}
		}
	}
	#---------
	
	###  Run apache daemon
	$simpjournal->test_case_start("Run apache daemon");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/sbin/apachectl start 
- Expected result: The web server daemon starts as background process
  (check this with the ps command)
HEREDOC
	
	# Run the command: /opt/lsb/appbat/sbin/apachectl start 
	my $cmd = "/opt/lsb/appbat/sbin/apachectl start";
	my $msg = capture( "$cmd 2>&1" );
	$result = $?;
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message("%> $cmd");
		$simpjournal->test_message($msg);
	}
	
	# Expected result: The web server daemon starts as background process
	#  (check this with the ps command)
	if ( cmd("ps -A | grep -e '\\<httpd\$' >/dev/null") != 0 ) {
		$simpjournal->test_comment("http daemon didn't start");
		$simpjournal->test_case_result("failed");
	}
	else {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	}
	#---------
	
	###  Check port 80
	$simpjournal->test_case_start("Check port 80");
	$simpjournal->test_description(<<HEREDOC);
Run the comand: netstat -an | grep ":80"
- Expected result: The command output should include a line like:
tcp        0      0 :::80                       :::*                        LISTEN
Note: netstat is not a required command in the LSB. It is not an FVT failure if the command itself is unavailable.
HEREDOC
	
	if ( does_shell_know("netstat" ) ) {
		# Run the comand: netstat -an | grep ":80"
		$cmd = 'netstat -an | grep -e ":80\b"';
		my $msg = capture($cmd);
		
		$simpjournal->test_message('%> '.$cmd);
		$simpjournal->test_message($msg);
		
		# Expected result: The command output should include a line like:
		# tcp        0      0 :::80                       :::*                        LISTEN
		if ( $msg =~ m/^tcp6?\s+0\s+0\s+(::|0\.0\.0\.0):80\s+\1:\*\s+LISTEN\s*$/m ) {
			$simpjournal->test_case_result("passed");
		}
		else {
			$simpjournal->test_case_result("failed");
		}
	}
	else {
		$simpjournal->test_case_result("untested");
	}
	#---------
	
	###  Check the test page
	$simpjournal->test_case_start("Check the test page");
	$simpjournal->test_description(<<HEREDOC);
Open a web browser using the URL: http://localhost 
- Expected result: Test page stating "It works!"
HEREDOC

	# (also disable http_proxy just in case)
	if ( $Download::tool eq 'curl' ) {
		$msg = capture('http_proxy= curl http://localhost -s');
	} else {
		$msg = capture('http_proxy= wget http://localhost -q -O -');
	}
	if ( $msg =~ /It works!/ ) {
		$simpjournal->test_case_result("passed");
	}
	elsif ( $msg =~ /apache_pb\.gif/ ) {
		# Apache 2.0.53 (for LSB 3.0) generates another test page
		$simpjournal->test_case_result("passed");
	}
	else {
		$simpjournal->test_message($msg);
		$simpjournal->test_case_result("failed");
	}
	#---------
	
	$simpjournal->close();
	
	return 1; # Ok
}

sub Cleanup {
	my ($self) = @_;
	
	# Shut down the lsb-apache webserver with the following command: 
	# /opt/lsb/appbat/sbin/apachectl stop
	cmd( "/opt/lsb/appbat/sbin/apachectl stop" );
	sleep 2; # Wait a bit allowing apache to stop completely.

	my $result = check_port_80();
	if ( is_ok($result) && $result ) {
		stop_httpd();
	}
	
	# Deinstall apache, because it adds system startup links to /etc/init.d.
	if ( $self->{REQUIREMENTS} ) {
		my $main_package_req = $self->{REQUIREMENTS}[-1];
		if ( $main_package_req && $main_package_req->{PACKAGE} ) {
			inform "Deinstalling ".$self->name."...";
			remove_package($main_package_req->{PACKAGE});
		}
	}
	
	if ( $self->{'stopped_script'} ) {
		# Trying to re-enable stopped web-server.
		inform("Starting previously stopped ".$self->{'stopped_script'}." daemon");
		if ( cmd("/etc/init.d/".$self->{'stopped_script'}." start") == 0 ) {
			sleep 1; # Wait a bit
		}
	}
}

1; # return value
