# Distribution Checker
# Expect FVT Test Module (Expect_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Expect_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common

#----------------------------------------------------------------------

my $test_user = "expect4lsb";

my $checkpoints = undef;

#----------------------------------------------------------------------

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory.", $Error::Last;
	
	# Copy test data file to the temporary directory
	is_ok $self->copy_files_from_reqs_to( $self->temp_dir )
		or return error "Failed to prepare test data file.", $Error::Last;
	
	# Unpack the test data file.
	cmd("cd ".shq($self->temp_dir)." && tar xvf expect-tests.tar >/dev/null") == 0
		or return error "Failed to unpack test data tarball.";
	
	# Check the unpacked files
	( -f $self->temp_dir."/all.tcl" )
		or return error "Failed to unpack test data tarball or tarball is wrong."; 

	# Create a test user.
	if ( cmd( "id $test_user >/dev/null" ) != 0 ) {
		cmd("useradd -m $test_user") == 0
			or return error "Failed to create test user '$test_user'.";
	}
	
	# Chown all the temp direrctory
	cmd( "chown ${test_user}:users -R ".shq($self->temp_dir) ) == 0
		or return error "Failed to change owner of '".$self->temp_dir."'.";
	
	$self->{JOURNAL_FILE} = $self->name.".tcljournal";
	$self->{JOURNAL_TYPE} = 'TCL';
	
	# Expected number of test cases
	if ( version_filter("LSB <= 3.2", $globals->{'standard'}) ) {
		$self->{TESTS_EXPECTED} = 7;
	}
	
	init_checkpoint_data($self->{VERSION});
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	# Open a subshell
	my $subshell = $self->Spawn_subshell($test_user);
	is_ok $subshell or return $Error::Last;
	
	# Go to the test temp directory
	$subshell->Settle();
	$subshell->Send("cd ".shq($self->temp_dir)."\n");
	
	# Use $self->temp_dir for journal because the subshell is run under test user.
	my $cmd = "/opt/lsb/appbat/bin/expect all.tcl | tee ".shq($self->temp_dir."/".$self->{JOURNAL_FILE});
	$subshell->Settle();
	$subshell->Send("$cmd; exit\n"); # Run the test
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(3) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	return 1; # Ok
}

sub after_run {
	my ($self) = @_;
	
	is_ok $self->copy_file( $self->temp_dir."/".$self->{JOURNAL_FILE}, $self->journal_file )
		or $self->report_error(complain $Error::Last);
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /(\S+\.test)[\r\n]/ ) {
		my $testcase = $1;
		$self->progress_profiled($checkpoints, $testcase);
		return;
	}
	# else:
	$self->progress_profiled();
}

sub Cleanup {
	my ($self) = @_;
	
	if ( cmd( "id $test_user >/dev/null" ) == 0 ) {
		cmd( "userdel -r ".$test_user );
	}
}

sub init_checkpoint_data {
	my ($ver) = @_;
	
	if ( $ver lt '5.43.0-5' ) {
		# expect v. 5.43.0-3 for LSB 3.0 or earlier
		$checkpoints = [
			[ 10, 'cat.test' ],
			[ 73, 'logfile.test' ],
			[ 100, 'spawn.test' ],
		];
	} 
	elsif ( $ver eq '5.43.0-5' ) {
		# expect v. 5.43.0-5 for LSB 3.1
		$checkpoints = [
			[ 10, 'cat.test' ],
			[ 68, 'logfile.test' ],
			[ 100, 'spawn.test' ],
		];
	}
	elsif ( $ver eq '5.43.0-6' ) {
		# expect v. 5.43.0-6 for LSB 3.2
		$checkpoints = [
			[ 23, 'cat.test' ],
			[ 82, 'logfile.test' ],
			[ 100, 'spawn.test' ],
		];
	}
	else {
		# expect v. 5.43.0-7 for LSB 4.0
		$checkpoints = [
			[ 12, 'cat.test' ],
			[ 75, 'logfile.test' ],
			[ 93, 'spawn.test' ],
			[ 100, 'stty.test' ],
		];
	}
}

1; # return value
