# Distribution Checker
# Groff FVT Test Module (Groff_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Groff_test;
use strict;

use Misc;
use SimpleJournal;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

sub prepare {
	my ($self) = @_;

	# 1. Download the lsb-groff application package that corresponds to the LSB
	#    version and processor architecture for which certification is being 
	#    sought from the Download Page. 
	# 2. Install package on system.
	# NOTE: Application MUST be installed without use of --nodeps option,
	#  the dependency checks are part of the test.
	
	# Download and install required packages
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory.", $Error::Last;
	
	$self->{JOURNAL_FILE} = $self->name.".simplejournal";
	$self->{JOURNAL_TYPE} = 'SIMPLE';

	my $dir = "/opt/lsb/appbat/man";
	if ( vercmp($self->{VERSION}, "1.20") >= 0 ) { # version >= 1.20
		$dir = "/opt/lsb/appbat/share/man";
	}
	chdir( $dir )
		or return error "Failed to chdir to $dir: $!";
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $simpjournal = SimpleJournal->new( $self->journal_file );
	is_ok $simpjournal or return error "Failed to open journal.", $Error::Last;
	
	$simpjournal->test_name( $self->name );
	
	my $tests_expected = 6;
	$simpjournal->tests_expected( $tests_expected ); # Number of tests
	
	### Functional Verification Test for lsb-groff ###
	# http://www.linuxfoundation.org/appbat/fvt-3.0.0/lsb-groff_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.1.0/lsb-groff_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.2.0/lsb-groff_fvt.html
	
	###  Check installation
	$simpjournal->test_case_start("Check installation");
	
	$simpjournal->test_description(<<HEREDOC);
- Expected result: application should be installed to 
/opt/lsb/appbat
HEREDOC
	
	my $result = "passed";
	for my $file ( "/opt/lsb/appbat/bin/groff",
				"/opt/lsb/appbat/share/groff",
				"/opt/lsb/appbat/lib/groff" )
	{
		if ( !-e $file ) {
			$result = "failed";
			$simpjournal->test_comment("'$file' is expected to exist after installation, but not found.");
		}
	}
	
	$simpjournal->test_case_result($result);
	#---------

	###  groff -v 
	$simpjournal->test_case_start("groff -v");

	my $ver = '1.19';
	my $copyright_year = '2002';
	if ( vercmp($self->{VERSION}, "1.20") >= 0 ) { # version >= 1.20
		$ver = '1.20.1';
		$copyright_year = '2009';
	}
	$simpjournal->test_description(<<HEREDOC);
Run the command /opt/lsb/appbat/bin/groff -v 
- Expected result: Version information for groff will be displayed.
It should look something like this: 
  GNU groff version $ver
  Copyright (C) $copyright_year Free Software Foundation, Inc.
  GNU groff comes with ABSOLUTELY NO WARRANTY.
  You may redistribute copies of groff and its subprograms
  under the terms of the GNU General Public License.
  For more information about these matters, see the file named COPYING.

  called subprograms:

  GNU troff (groff) version $ver
  GNU grops (groff) version $ver
HEREDOC
	
	# Change to the /opt/lsb/appbat/share/man directory 
	#   cd /opt/lsb/appbat/share/man
	# Run the command /opt/lsb/appbat/bin/groff -v 
	my $cmd = "/opt/lsb/appbat/bin/groff -v";
	my $msg = capture( "$cmd 2>&1" );
	$result = $?;
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message("%> $cmd");
		$simpjournal->test_message($msg);
	}

	#Expected result: Version information for groff will be displayed.
	$ver =~ s/([\.\?\*\+\$\^\(\)\|\[\]\\])/\\$1/g;
	if (   $msg =~ /GNU groff version $ver/
		&& $msg =~ /Copyright \(C\) 200\d Free Software Foundation/
		&& $msg =~ /GNU General Public License/
		&& $msg =~ /GNU grops \(groff\) version $ver/
		&& $msg =~ /GNU troff \(groff\) version $ver/
		 ) 
	{
		$simpjournal->test_case_result("passed");
	} else {
		$simpjournal->test_comment("Output doesn't look familiar to the test");
		if ( $result == 0 ) {
			$simpjournal->test_message("%> $cmd");
			$simpjournal->test_message($msg);
		}
		$simpjournal->test_case_result("failed");
	}
	#---------

	###  groff -mandoc
	$simpjournal->test_case_start("groff -mandoc");
	
	$simpjournal->test_description(<<HEREDOC);
Run the command /opt/lsb/appbat/bin/groff -mandoc man1/groff.1
- Expected result: The groff man page will be formatted for a Postscript printer
 and displayed on the screen.
HEREDOC

	# Run the command /opt/lsb/appbat/bin/groff -mandoc man1/groff.1
	$cmd = "/opt/lsb/appbat/bin/groff -mandoc man1/groff.1";
	$msg = capture( "$cmd 2>&1" );
	$result = $?;
	my $output_file = $self->name.".output1";
	write_string_as_file( $globals->{'test_result_dir'}."/".$output_file, $msg );
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
		$simpjournal->test_message("%> $cmd");
	}

	if (   $msg =~ /^%!PS/
		&& $msg =~ /%%Creator: groff/
		&& $msg =~ /%%EOF$/
		 ) 
	{
		$simpjournal->test_case_result("passed");
	} 
	else {
		$simpjournal->test_comment("Output doesn't look familiar to the test");
		if ( $result == 0 ) {
			$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
			$simpjournal->test_message("%> $cmd");
		}
		$simpjournal->test_case_result("failed");
	}
	#---------
	
	###  groff -mandoc -Tascii
	$simpjournal->test_case_start("groff -mandoc -Tascii");
	
	$simpjournal->test_description(<<HEREDOC);
Run the command /opt/lsb/appbat/bin/groff -mandoc -Tascii man1/nroff.1
- Expected result: The nroff man page will be formatted for an ASCII device
 and displayed on the screen.
HEREDOC

	# Run the command /opt/lsb/appbat/bin/groff -mandoc -Tascii man1/nroff.1
	$cmd = "/opt/lsb/appbat/bin/groff -mandoc -Tascii man1/nroff.1";
	$msg = capture( "$cmd 2>&1" );
	$result = $?;
	$output_file = $self->name.".output2";
	write_string_as_file( $globals->{'test_result_dir'}."/".$output_file, $msg );
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
		$simpjournal->test_message("%> $cmd");
	}

	if (   $msg =~ /^NROFF\(1\)/
		&& $msg =~ /NROFF\(1\)$/
		&& $msg =~ /\033\[\d+m/ # That are escape-codes that expected to appear in -Tascii format
		 ) 
	{
		$simpjournal->test_case_result("passed");
	} 
	else {
		$simpjournal->test_comment("Output doesn't look familiar to the test");
		if ( $result == 0 ) {
			$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
			$simpjournal->test_message("%> $cmd");
		}
		$simpjournal->test_case_result("failed");
	}
	#---------
	
	###  groff -mandoc -Tlatin1
	$simpjournal->test_case_start("groff -mandoc -Tlatin1");
	
	$simpjournal->test_description(<<HEREDOC);
Run the command /opt/lsb/appbat/bin/groff -mandoc -Tlatin1 man1/pic.1
- Expected result: The pic man page will be formatted for the ISO 8859-1 character set
	and displayed on the screen.
HEREDOC

	# Run the command /opt/lsb/appbat/bin/groff -mandoc -Tlatin1 man1/pic.1
	$cmd = "/opt/lsb/appbat/bin/groff -mandoc -Tlatin1 man1/pic.1";
	$msg = capture( "$cmd 2>&1" );
	$result = $?;
	$output_file = $self->name.".output3";
	write_string_as_file( $globals->{'test_result_dir'}."/".$output_file, $msg );
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
		$simpjournal->test_message("%> $cmd");
	}

	if (   $msg =~ /^PIC\(1\)/
		&& $msg =~ /PIC\(1\)$/
		&& $msg =~ /\033\[\d+m/ # That are escape-codes that expected to appear in -Tlatin1 format
		 ) 
	{
		$simpjournal->test_case_result("passed");
	} 
	else {
		$simpjournal->test_comment("Output doesn't look familiar to the test");
		if ( $result == 0 ) {
			$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
			$simpjournal->test_message("%> $cmd");
		}
		$simpjournal->test_case_result("failed");
	}
	#---------
	
	###  groff -mandoc -Thtml
	$simpjournal->test_case_start("groff -mandoc -Thtml");
	
	$simpjournal->test_description(<<HEREDOC);
Run the command /opt/lsb/appbat/bin/groff -mandoc -Thtml man1/eqn.1
- Expected result: The eqn man page will be formatted for HTML 
	and displayed on the screen.
HEREDOC

	# Run the command /opt/lsb/appbat/bin/groff -mandoc -Thtml man1/eqn.1
	$cmd = "/opt/lsb/appbat/bin/groff -mandoc -Thtml man1/eqn.1";
	$msg = capture( "$cmd 2>&1" );
	$result = $?;
	$output_file = $self->name.".output4";
	write_string_as_file( $globals->{'test_result_dir'}."/".$output_file, $msg );
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
		$simpjournal->test_message("%> $cmd");
	}

	if (   $msg =~ /^<!--\s+Creator\s+:\s+groff /
		&& $msg =~ /<\/html>$/
		 ) 
	{
		$simpjournal->test_case_result("passed");
	} 
	else {
		$simpjournal->test_comment("Output doesn't look familiar to the test");
		if ( $result == 0 ) {
			$simpjournal->test_comment("See the {noreplace}".Report::local_href("results/".$output_file, "output")."{/noreplace};");
			$simpjournal->test_message("%> $cmd");
		}
		$simpjournal->test_case_result("failed");
	}
	#---------
	
	$simpjournal->close();
	
	return 1; # Ok
}

#----------------------------------------------------------------------
1; # return value
