# Distribution Checker
# Libstdcpp Test Module (Libstdcpp_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Libstdcpp_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

my $checkpoints = undef;

#----------------------------------------------------------------------

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;

	( defined $self->{OPTIONS} )
		or return error "Answers are not defined"; # Should not happen
	
	return 1;
}

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory.", $Error::Last;
	
	# Options & questions
		
	# profile option
	my $q = 'profile';
	if ( defined $self->{OPTIONS}{$q} ) {
		if ( !defined $self->{OPTIONS}{$q}{VALUE} ) {
			if ( defined $globals->{'std_profile'} ) {
				$self->{OPTIONS}{$q}{VALUE} = $globals->{'std_profile'};
			}
		}
	}
	
	my $testsuite_dir = $self->option('TESTSUITE_DIR') || "/opt/lsb/test/libstdcpp_4.1.0";
	$self->{TESTSUITE_DIR} = $testsuite_dir;
	
	init_checkpoint_data($self->{VERSION});
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $subshell = $self->Spawn_subshell();
	is_ok $subshell or return $Error::Last;
	
	# Change to the temporary directory
	$subshell->Settle();
	$subshell->Send( "cd ".shq($self->temp_dir)."\n" );
	
	# Prepare the command
	my $command = shq($self->{TESTSUITE_DIR}."/run_tests");
	if ( defined $self->option('profile') ) {
		$command .= " -T".$self->option('profile');
	}
	
	$subshell->Settle();
	$subshell->Send("$command; exit\n"); # Run the test
	
	my $questions_wait_time = 60*10; # In seconds
	my $questions_deadline = time() + $questions_wait_time;
	
	my $started = 0;
	
	my $line;
	DIALOG_LOOP:
	while ( $line = $subshell->Read($questions_wait_time) ) {
		last if !$line; #required for 'redo'
		
		if ( $line =~ /Executing the test suites/ ) {
			$started = 1;
			last;
		}
		
		# Answer the prompts
		
		$subshell->Settle($line) or redo DIALOG_LOOP;
		
		if ( $line =~ /Where do you want to create the test database[^\?]*\?/ ) {
			$subshell->Settle($line) or redo DIALOG_LOOP;
			$subshell->Send( $self->temp_dir."\n" );
		}
		
		# Formalized questions
		foreach my $opt ( values %{$self->{OPTIONS}} ) {
			next if !defined $opt->{PRE} || !defined $opt->{VALUE};

			my $pre = $opt->{PRE};
			if ( $line =~ /$pre/  ) {
				
				$subshell->Send( $opt->{VALUE}."\n" );
				next DIALOG_LOOP;
			}
		}
	} continue {
		
		# Check the questions timeout
		if ( time() > $questions_deadline ) {
			inform "Time for questions is over.";
			is_ok $subshell->ExpectLogout() or return $Error::Last;
			return error "Timeout";
		}
	}
	if ( $subshell->{TIMEOUT} ) { inform "!!! Warning: timeout at subshell"; }
	
	$self->{ESTIM_START_TIME} = time();
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(3) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	if ( !$started ) {
		return error "The test hasn't started";
	}
	
	return 1; # Ok
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /^  (.*\.cc) +:.*[\r\n]/ ) { # Something like "  thread/pthread1.cc    : PASS"
		my $testcase = $1;
		$self->progress_profiled($checkpoints, $testcase);
		return;
	}
	# else:
	$self->progress_profiled();
}

sub after_run {
	my ($self) = @_;
	
	$self->{JOURNAL_FILE} = $self->name.".journal";
	$self->{JOURNAL_TYPE} = 'TET';
	
	# Find the journal file and copy it to where it should be.
	is_ok $self->copy_file_globbed( $self->temp_dir."/v3db/journal*", $self->journal_file )
		or $self->report_error(complain $Error::Last);
	
	return 1;
}

sub init_checkpoint_data {
	my ($ver) = @_;
	
	if ( $ver le '3.4.3-6' ) { # 3.4.3-6, LSB 3.1
		$checkpoints = [
			[ 2, '17_intro/header_cassert.cc' ],
			[ 87, '24_iterators/istreambuf_iterator/2.cc' ],
			[ 93, '27_io/basic_ifstream/3.cc' ],
			[ 99, 'thread/pthread3.cc' ],
		];
	}
	elsif ( $ver ge '3.4.3-7' && $ver lt "4" ) { # 3.4.3-7, LSB 3.2
		$checkpoints = [
			[ 2, '17_intro/header_cassert.cc' ],
			[ 8, '22_locale/num_get/get/wchar_t/wrapped_env.cc' ],
			[ 75, '24_iterators/istreambuf_iterator/2.cc' ],
			[ 81, '27_io/basic_filebuf/underflow/char/10097.cc' ],
			[ 88, '27_io/objects/wchar_t/7.cc' ],
			[ 96, 'thread/pthread1.cc' ],
		];
	}
	else { # 4.1.0-4, LSB 4.0
		$checkpoints = [
			[ 4, '17_intro/header_cassert.cc' ],
			[ 10, '22_locale/codecvt/out/wchar_t/4.cc' ],
			[ 16, '22_locale/messages/members/char/3.cc' ],
			[ 22, '22_locale/num_put/put/wchar_t/wrapped_env.cc' ],
			[ 28, '26_numerics/numeric/sum_diff.cc' ],
			[ 35, '27_io/basic_ostream/cons/wchar_t/9827.cc' ],
			[ 41, '27_io/ios_base/storage/11584.cc' ],
			[ 90, 'thread/pthread1.cc' ],
			[ 97, 'tr1/6_containers/array/capacity/empty.cc' ],
		];
	}
}

1; # return value
