# Distribution Checker
# Olver Test Module (Olver_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Olver_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;

	my @errors = ();

	# Check for required perl modules
	my @missing_modules = ();
	# !!! Keep this list in sync with the suggest_packages() sub !!!
	for my $module ( qw/ Data::Dumper  File::Basename  File::Spec  Getopt::Long  Getopt::Std URI::file/ )
	{
		if ( cmd("perl -e 'use ".$module."'") != 0 ) {
			push @missing_modules, $module;
		}
	}
	if ( @missing_modules ) {
		push @errors, error missing_packages_message("Can't find some required perl modules, please install them:\n ", @missing_modules);
	}

	# Search for a XML parser.
	{
		my $found = 0;
		# Any one of them will be enough:
		my @sax_modules = qw/ XML::SAX::Expat  XML::SAX::ExpatXS  XML::LibXML::SAX XML::Parser::PerlSAX /;
		for my $module ( @sax_modules ) {
			if ( cmd("perl -e 'use ".$module."'") == 0 ) {
				$found = 1;
				last;
			}
		}
		if ( !$found ) {
			push @errors, error missing_packages_message("Can't find any suitable XML-SAX parser module.\nPlease install the package which provides XML::LibXML::SAX perl module,\nOR any of ALTERNATIVE modules: 'XML::SAX::Expat', 'XML::SAX::ExpatXS' or 'XML::Parser::PerlSAX'.\n ", @sax_modules);
		}
	}
	
	if ( @errors ) {
		return join_errors(@errors);
	}
	
	return 1;
}

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	my $testsuite_dir = $self->option('TESTSUITE_DIR') || "/opt/lsb/test/olver-core";
	$self->{TESTSUITE_DIR} = $testsuite_dir;
	$self->{RESULTS_DIR} = $self->option('RESULTS_DIR') || "/var/opt/lsb/test/olver-core";
	
	# Remove old results
	cmd("rm -rf ".shq($self->{RESULTS_DIR})."/*");
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $subshell = $self->Spawn_subshell();
	is_ok $subshell or return $Error::Last;
	
	# Change to the test suite directory
	$subshell->Settle();
	$subshell->Send( "cd ".shq($self->{TESTSUITE_DIR})."\n" );
	
	# Run the test
	$subshell->Settle();
	$subshell->Send("./bin/olver_run_tests; exit\n");
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(2) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} && $subshell->{EXIT_CODE} != 1 ) {
		# Olver returns 1 if there were problems found.
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	return 1; # Ok
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /^Running scenario .* \((\d+) of (\d+)\).*[\r\n]/ ) {
		my $n = $1;
		my $total = $2;
		$self->progress($n/$total);
	}
}

sub after_run {
	my ($self) = @_;
	
	# Journal
	$self->{JOURNAL_FILE} = $self->name.".journal";
	$self->{JOURNAL_TYPE} = 'TET';
	
	# Find the journal file and copy it to where it should be.
	is_ok $self->copy_file_globbed( $self->{RESULTS_DIR}."/*/tet/tetreport.log", $self->journal_file )
		or $self->report_error(complain $Error::Last);
	
	is_ok $self->copy_file_globbed( "/var/opt/lsb/test/olver-core/*", $globals->{"test_result_dir"}."/olver_full" )
		or $self->report_error(complain $Error::Last);

	$self->report( '<p>Open '.Report::local_href( "results/olver_full/summary.htm", 'full report' ).'.</p>' );

	return 1;
}

1; # return value
