# Distribution Checker
# Perl Test Module (Perl_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Perl_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

my $checkpoints = undef;

#----------------------------------------------------------------------

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;

	# Check for required perl modules
	# !!! Keep this list in sync with the suggest_packages() sub !!!
	my @modules_to_check = qw/Tie::Hash Time::HiRes Test::Harness Test::Simple Test::More Test::Builder File::Spec ExtUtils::MakeMaker/;
	my @missing_modules = ();
	if ( version_filter("< 4.0", $self->{VERSION}) ) {
		push @modules_to_check, qw/Test::Harness::Assert Test::Harness::Straps/;
	}
	
	for my $module ( @modules_to_check )
	{
		if ( cmd("perl -e 'use ".$module."'") != 0 ) {
			push @missing_modules, $module;
		}
	}
	if ( @missing_modules ) {
		return error missing_packages_message("Can't find some required perl modules, please install them:\n ", @missing_modules);
	}
	
	return 1;
}

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	my $testsuite_dir = $self->option('TESTSUITE_DIR') || "/opt/lsb/test/perl";
	$self->{TESTSUITE_DIR} = $testsuite_dir;
	$self->{RESULTS_DIR} = $self->option('RESULTS_DIR') || "/opt/lsb/test/perl/results";
	
	# Remove old results
	cmd("rm -rf ".shq($self->{RESULTS_DIR})."/*");
	
	init_checkpoint_data($self->{VERSION});
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $subshell = $self->Spawn_subshell();
	is_ok $subshell or return $Error::Last;
	
	# Change to the test suite directory
	$subshell->Settle();
	$subshell->Send( "cd ".shq($self->{TESTSUITE_DIR})."\n" );
	
	# Run the test
	$subshell->Settle();
	$subshell->Send("./run_tests; exit\n");
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(3) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	return 1; # Ok
}

my $TC_last = "";
sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /^([^\/\s]+\/[^\s\.]+)\.\.\..*[\r\n]/ ) {
		if ( $1 ne $TC_last ) {
			my $testcase = $TC_last = $1;
			$self->progress_profiled($checkpoints, $testcase);
			return;
		}
	}
	# else:
	$self->progress_profiled();
}

sub after_run {
	my ($self) = @_;

	$self->{JOURNAL_FILE} = $self->name.".journal";
	$self->{JOURNAL_TYPE} = 'TET';
	
	# Find the journal file and copy it to where it should be.
	is_ok $self->copy_file_globbed( $self->{RESULTS_DIR}."/0*/journal", $self->journal_file )
		or $self->report_error(complain $Error::Last);
	
	return 1;
}

sub init_checkpoint_data {
	my ($ver) = @_;
	
	if ( $ver lt "4.0" ) {
		# perl v. 3.2.1-1 for LSB 3.2
		$checkpoints = [
			[ 0, 'base/cond' ],
			[ 6, 'io/pipe' ],
			[ 12, 'op/fork' ],
			[ 19, 'op/utftaint' ],
			[ 28, '../lib/Memoize/t/expmod_t' ],
			[ 34, '../lib/Test/Simple/t/is_deeply_fail' ],
			[ 40, '../lib/Net/Ping/t/400_ping_syn' ],
			[ 46, '../lib/Net/Ping/t/200_ping_tcp' ],
			[ 53, '../lib/warnings' ],
			[ 59, '../lib/Math/BigInt/t/fallback' ],
			[ 65, '../lib/ExtUtils/t/Embed' ],
			[ 76, '../lib/Benchmark' ],
			[ 83, '../ext/Time/HiRes/t/HiRes' ],
			[ 90, '../ext/B/t/concise-xs' ],
			[ 96, '../ext/IO/t/io_pipe' ],
		];
	} else {
		# perl v. 4.0.5-1 for LSB 4.0
		$checkpoints = [
			[ 0, 'base/cond' ],
			[ 8, 'io/through' ],
			[ 14, 'op/pack' ],
			[ 21, 'uni/class' ],
			[ 27, '../ext/Devel/PPPort/t/ppphtest' ],
			[ 34, '../ext/IO/t/io_sock' ],
			[ 40, '../ext/Time/HiRes/t/HiRes' ],
			[ 47, '../lib/Benchmark' ],
			[ 56, '../lib/Benchmark' ],
			[ 62, '../lib/Getopt/Long/t/gol-linkage' ],
			[ 69, '../lib/Math/BigInt/t/use_lib2' ],
			[ 75, '../lib/Memoize/t/expmod_t' ],
			[ 82, '../lib/Memoize/t/speed' ],
			[ 88, '../lib/Net/Ping/t/400_ping_syn' ],
			[ 94, '../lib/Tie/File/t/28_mtwrite' ],
		];
	}
}

1; # return value
