# Distribution Checker
# Printing Test Module (Printing_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Printing_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

my $checkpoints = undef;

#----------------------------------------------------------------------

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;
	
	my @missing_libs = ();
	for my $lib ( "libcups.so.2",  "libcupsimage.so.2") {
		my $lib_grep = "\\<".$lib." ";
		$lib_grep =~ s/\./\\./g;
		if ( cmd("ldconfig -p | grep ".shq($lib_grep)) != 0 ) {
			push @missing_libs, $lib;
		}
	}
	if ( @missing_libs ) {
		$self->report_warning( warning missing_packages_message("Can't find some 'cups' libraries, please install them:\n ", @missing_libs));
	}

	my @missing_apps = ();
	for my $app ( "foomatic-rip", "gs" ) {
		if ( ! -x "/usr/bin/$app" ) {
			push @missing_apps, $app;
		}
	}
	if ( @missing_apps ) {
		$self->report_warning( warning missing_packages_message("Can't find some tools, please install them:\n ", @missing_apps));
	}

	return 1;
}

sub stop_cups {
	my ($process) = @_;
	
	if ( !defined $process ) {
		# Look for httpd process (can be named differently)
		for my $name ("cupsd") {
			if ( cmd("ps -A | grep -e '\\<".$name."\$'") == 0 ) {
				$process = $name; # Gotcha!
				last;
			}
		}
	}
	( $process ) or return ""; # Return ok if process not found
	
	# Trying to stop the daemon gracefully
	my $script = "";
	for my $scriptname ($process, "cupsys", "cups", "cupsd") {
		if ( -f "/etc/init.d/$scriptname" ) {
			inform("Stopping $scriptname daemon");
			cmd("/etc/init.d/$scriptname stop");
			sleep 2; # Wait a bit allowing the daemon to stop completely.
			
			if ( cmd("ps -A | grep -e '\\<".$process."\$'") != 0 ) {
				$script = $scriptname;
				last;
			}
		}
	}

	# Check the result
	return $script if cmd("ps -A | grep -e '\\<".$process."\$'") != 0;
	
	# Not stopped. Kill the process.
	cmd ("killall $process");
	sleep 2; # Wait a bit allowing apache to stop completely.
	
	# Check the result
	return $script if cmd("ps -A | grep -e '\\<".$process."\$'") != 0;
	
	# Still not stopped. Use kill -9
	cmd ("killall -9 $process");
	sleep 2; # Wait a bit allowing apache to stop completely.
	
	# Check the result
	if ( cmd("ps -A | grep -e '\\<".$process."\$'") == 0 ) {
		# Still not stopped.
		return error "Failed to stop '$process'";
	}
	
	return $script;
}

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	init_checkpoint_data($self->{VERSION});
	
	my $testsuite_dir = $self->option('TESTSUITE_DIR') || "/opt/lsb/test/printing";
	$self->{TESTSUITE_DIR} = $testsuite_dir;
	$self->{RESULTS_DIR} = $self->option('RESULTS_DIR') || "/opt/lsb/test/printing/results";
	
	# Remove old results
	cmd("rm -rf ".shq($self->{RESULTS_DIR})."/*");
	
	# Stop the cups if it is running
	$self->{'stopped_script'} = stop_cups();
	( is_ok $self->{'stopped_script'} )
		or return error "Failed to stop cups daemon", $Error::Last;
		
	return 1;
}

sub run {
	my ($self) = @_;
   
	my $subshell = $self->Spawn_subshell("root");
	is_ok $subshell or return $Error::Last;
	
	# Change to the test suite directory
	$subshell->Settle();
	$subshell->Send( "cd ".shq($self->{TESTSUITE_DIR})."\n" );
	
	$subshell->Settle();
	$subshell->Send( "export USER=root\n" ); # /testfoomaticrip/testfoomaticrip test script uses variable $USER
	
	# Run the test
	$subshell->Settle();
	$subshell->Send("./run_tests ; exit\n");
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(2) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	return 1; # Ok
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /\d+:\d+:\d+\s+Execute (.*)[\r\n]/ ) {
		my $testcase = $1;
		$self->progress_profiled($checkpoints, $testcase);
		return;
	}
	# else:
	$self->progress_profiled();
}

sub after_run {
	my ($self) = @_;
	
	my @errors = ();
	
	$self->{JOURNAL_FILE} = $self->name.".journal";
	$self->{JOURNAL_TYPE} = 'TET';
	
	# Find the journal file and copy it to where it should be.
	is_ok $self->copy_file_globbed( $self->{RESULTS_DIR}."/0*/journal", $self->journal_file )
		or $self->report_error(complain $Error::Last);
		
	return 1;
}

sub Cleanup {
	my ($self) = @_;
	
	stop_cups();
	
	cmd("killall -9 foomatic-rip"); # foomatic-rip often remains hanging up after the test.
	
	if ( $self->{'stopped_script'} ) {
		# Trying to restart the server.
		inform("Starting previously stopped ".$self->{'stopped_script'}." daemon");
		cmd("/etc/init.d/".$self->{'stopped_script'}." start");
	}
}

sub init_checkpoint_data {
	my ($ver) = @_;
	
	# $ver <= 3.2.*
	if ( $ver lt "4.0" ) {
		$checkpoints = [
			[ 1, '/convenience/cupsConvenience' ],
			[ 3, '/ppd/cupsPPD' ],
			[ 11, '/testfoomaticrip/testfoomaticrip' ],
		];
	}
	else { # 4.0
		$checkpoints = [
			[ 2, '/convenience/cupsConvenience' ],
			[ 11, '/testfoomaticrip/testfoomaticrip' ],
			[ 98, '/fhs/share-ppd/share-ppd' ],
		];
	}
}

1; # return value
