# Distribution Checker
# Python FVT Test Module (Python_app_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Python_app_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

my $test_user = "python4lsb";

my $checkpoints = undef;

#----------------------------------------------------------------------

sub prepare {
	my ($self) = @_;

	# 1. Download the lsb-python application package that corresponds to the LSB
	#    version and processor architecture for which certification is being 
	#    sought from the Download Page. 
	# 2. Install package on system.
	# NOTE: Application MUST be installed without use of --nodeps option,
	#  the dependency checks are part of the test.
	
	# Download and install required packages
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory.", $Error::Last;
	
	# Change to the python library directory: 
	#   cd /opt/lsb/appbat/lib/python2.4
	chdir( "/opt/lsb/appbat/lib/python2.4" )
		or return error "Can't chdir to /opt/lsb/appbat/lib/python2.4: $!";
	
	# Change permissions on the test directory to allow non-root write access 
	#   chmod 777 test
	chmod( 0777, "test" )
		or return error "Failed to chmod 'test': $!";
	
	# Create a non-root user called "python4lsb" 
	#   useradd -m python4lsb
	if ( cmd( "id $test_user >/dev/null" ) != 0 ) {
		cmd("useradd -m $test_user") == 0
			or return error "Failed to create user '$test_user'";
	}

	# Chown the temp direrctory
	cmd("chown ".$test_user.":users -R ".shq($self->temp_dir) ) == 0
		or return error "Failed to change owner of '".$self->temp_dir."'."; 
	
	# Change to the python library directory: 
	#   cd /opt/lsb/appbat/lib/python2.4
	chdir( "/opt/lsb/appbat/lib/python2.4/test" )
		or return error "Can't chdir to /opt/lsb/appbat/lib/python2.4/test: $!";
	
	# Move aside the pre-compiled python scripts by running the following command sequence: 
	#   mkdir SAVE
	#   mv *.pyc *.pyo SAVE
	unless ( -d "SAVE" ) {
		mkdir( "SAVE" )
			or return error "Failed to mkdir SAVE: $!";
		cmd( "mv *.pyc *.pyo SAVE" ) == 0
			or return error "Failed to save *.pyc and *.pyo files to directory SAVE";
	}
	
	$self->{JOURNAL_FILE} = $self->name.".pythonjournal";
	$self->{JOURNAL_TYPE} = 'PYTHON';
	
	init_checkpoint_data($self->{VERSION});
	
	return 1;
}

sub run {
	my ($self) = @_;

	# Open a subshell
	my $subshell = $self->Spawn_subshell($test_user);
	is_ok $subshell or return $Error::Last;
	
	# Change to the python test directory 
	#   cd /opt/lsb/appbat/lib/python2.4/test
	$subshell->Settle();
	$subshell->Send("cd '/opt/lsb/appbat/lib/python2.4/test'\n");
	
	# Use $self->temp_dir for journal because the subshell is run under test user.
	my $cmd = "/opt/lsb/appbat/bin/python -E regrtest.py -u curses"
				." | tee ".shq($self->temp_dir."/".$self->{JOURNAL_FILE});
	
	$subshell->Settle();
	$subshell->Send("$cmd; exit\n"); # Run the test
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(3) ) {
		$self->check_progress($line);
	}
	
	# Just wait for the test finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	return 1; # Ok
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /^(test_\S*)[\r\n]/ ) {
		my $testcase = $1;
		$self->progress_profiled($checkpoints, $testcase);
		return;
	}
	# else:
	$self->progress_profiled();
}

sub after_run {
	my ($self) = @_;
	
	is_ok $self->copy_file( $self->temp_dir."/".$self->{JOURNAL_FILE}, $self->journal_file )
		or $self->report_error(complain $Error::Last);
}

sub Cleanup {
	my ($self) = @_;
	
	# Restore the original pre-compiled python scripts by running the following command sequence: 
	#   cd /opt/lsb/appbat/lib/python2.4/test
	#   rm -f *.pyc *.pyo
	#   mv SAVE/* .
	#   rmdir SAVE
	#   (Note: in the mv command, the final argument is a dot) 
	chdir "/opt/lsb/appbat/lib/python2.4/test"
	and do {
		if ( -d "SAVE" ) {
			cmd( "rm -f *.pyc *.pyo; mv SAVE/* .; rmdir SAVE" );
		}
	};
	
	# Remove the python4lsb user and its home directory 
	#   userdel -r python4lsb
	if ( cmd( "id $test_user >/dev/null" ) == 0 ) {
		cmd( "userdel -r $test_user" );
	}
	
	# Restore the original permissions to the python test directory 
	#   chmod 755 /opt/lsb/appbat/lib/python2.4/test
	chmod 0755, "/opt/lsb/appbat/lib/python2.4/test";
}

sub init_checkpoint_data {
	my ($ver) = @_;
	
	# LSB 3.1 - 4.0
	$checkpoints = [
		[ 3, 'test_grammar' ],
		[ 9, 'test_codeop' ],
		[ 16, 'test_descrtut' ],
		[ 22, 'test_gdbm' ],
		[ 28, 'test_pkg' ],
		[ 35, 'test_popen' ],
		[ 45, 'test_set' ],
		[ 59, 'test_site' ],
		[ 65, 'test_sunaudiodev' ],
		[ 89, 'test_threaded_import' ],
		[ 95, 'test_ucn' ],
	];
}

#----------------------------------------------------------------------
1; # return value
