# Distribution Checker
# Rsync FVT Module (Rsync_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Rsync_test;
use strict;

use Misc;
use SimpleJournal;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

sub prepare {
	my ($self) = @_;

	# 1. Download the lsb-rsync application package that corresponds to the LSB
	#    version and processor architecture for which certification is being 
	#    sought from the Download Page. 
	# 2. Install package on system.
	# NOTE: Application MUST be installed without use of --nodeps option,
	#  the dependency checks are part of the test.
	
	# Download and install required packages
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory.", $Error::Last;
	
	# Create the file /etc/opt/lsb/appbat/rsyncd.conf with the following content: 
	# [rsync]
	#    path = /tmp/rsync
	#    comment = rsync server directory
	#    read only = false

	my $path = $self->temp_dir."/rsync";
	is_ok write_string_as_file ( "/etc/opt/lsb/appbat/rsyncd.conf", <<HEREDOC ) or return error "Failed to write /etc/opt/lsb/appbat/rsyncd.conf", $Error::Last;
[rsync]
	path = $path
	comment = rsync server directory
	read only = false
HEREDOC
	
	#Create a directory called /tmp/rsync
	mkdir( $self->temp_dir."/rsync" )
		or return error "Failed to create '".$self->temp_dir."/rsync'";
	#Create a directory called /tmp/rsync.test
	mkdir( $self->temp_dir."/rsync.test" )
		or return error "Failed to create '".$self->temp_dir."/rsync.test'";
	
	# Run the command: grep "^nobody" /etc/passwd 
	# Expected result: Output shows user "nobody" already exists. 
	my $result = capture('grep -e "^nobody" /etc/passwd');
	if ( $result =~ /^\s*$/ ) {
		# If it is not present on the system, run the command: "useradd nobody" 
		# to create the nobody user. 
		cmd("useradd nobody") == 0
			or return error "Failed to add user 'nobody'";
	}
	
	# Run the command: grep "^nobody" /etc/group 
	# Expected result: Output shows group "nobody" already exists. 
	$result = capture('grep -e "^nobody" /etc/group');
	if ( $result =~ /^\s*$/ ) {
		# If it is not present on the system, run the command: "groupadd nobody" 
		# to create the nobody group. 
		cmd("groupadd nobody") == 0
			or return error "Failed to add group 'nobody'";
	}
	
	# Run the command: grep rsync /etc/services
	# - Expected result: Output shows tcp and udp port definitions for the rsync server.
	$result = capture('grep -e "^rsync" /etc/services');
	if ( $result =~ /^\s*$/ ) {
		# If definitions are not present, add the following two lines to /etc/services: 
		#   rsync   873/tcp
		#   rsync   873/udp
		is_ok append_string_to_file( "/etc/services", <<HEREDOC ) or return error "Failed to write /etc/services", $Error::Last;

rsync   873/tcp
rsync   873/udp
HEREDOC
	}
	
	# Run the command: chmod 777 /tmp/rsync /tmp/rsync.test
	# - Expected result: All users can write to both directories 
	chmod( 0777, $self->temp_dir."/rsync", $self->temp_dir."/rsync.test" ) == 2
		or return error "Failed to chmod /rsync or /rsync.test: $!";
		
	# Run the command: cp /etc/opt/lsb/appbat/rsyncd.conf /tmp/rsync 
	# - Expected result: Config file copied to /tmp/rsync directory 
	cmd( "cp /etc/opt/lsb/appbat/rsyncd.conf ".shq($self->temp_dir)."/rsync" ) == 0
		or return error "Failed to copy /etc/opt/lsb/appbat/rsyncd.conf";

	# Change to /tmp/rsync.test directory
	chdir( $self->temp_dir."/rsync.test" )
		or return error "Failed to chdir to '".$self->temp_dir."/rsync.test': $!";
	
	$self->{JOURNAL_FILE} = $self->name.".simplejournal";
	$self->{JOURNAL_TYPE} = 'SIMPLE';
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $result;

	my $simpjournal = SimpleJournal->new( $self->journal_file );
	is_ok($simpjournal) or return error "Failed to open journal.", $Error::Last;
	
	$simpjournal->test_name( $self->name );
	
	my $tests_expected = 7;
	$simpjournal->tests_expected( $tests_expected ); # Number of tests
	
	### Functional Verification Test for lsb-apache ###
	# http://www.linuxfoundation.org/appbat/fvt-3.0.0/lsb-rsync_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.1.0/lsb-rsync_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.2.0/lsb-rsync_fvt.html
	
	###  Check installation
	$simpjournal->test_case_start("Check installation");
	
	$simpjournal->test_description(<<HEREDOC);
- Expected result: application should be installed to 
/opt/lsb/appbat
/etc/opt/lsb/appbat
/var/opt/lsb/appbat
HEREDOC
	# Actually, not to '/var/opt/lsb/appbat'
	# And nothing is being installed to '/etc/opt/lsb/appbat'
	
	$result = "passed";
	for my $file ( "/opt/lsb/appbat/bin/rsync",
				 )
	{
		if ( !-e $file ) {
			$result = "failed";
			$simpjournal->test_comment("'$file' is expected to exist after installation, but not found.");
		}
	}
	$simpjournal->test_case_result($result);
	#---------
	
	$simpjournal->test_case_start("Copy rsync binary");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/rsync /opt/lsb/appbat/bin/rsync . (ends with dot) 
 - Expected result: rsync executable should be in /tmp/rsync.test directory
HEREDOC
	# Run the command: /opt/lsb/appbat/bin/rsync /opt/lsb/appbat/bin/rsync . (ends with dot) 
	$result = cmd("/opt/lsb/appbat/bin/rsync /opt/lsb/appbat/bin/rsync .");
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message($last_cmd_output);
	}
	# - Expected result: rsync executable should be in /tmp/rsync.test directory
	if ( -x $self->temp_dir."/rsync.test/rsync" ) {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	} else {
		$simpjournal->test_case_result("failed");
	}
	#---------

	$simpjournal->test_case_start("Run rsync daemon");
	$simpjournal->test_description(<<HEREDOC);
As root run the command: /opt/lsb/appbat/bin/rsync --daemon 
 - Expected result: rsync server running on test system
HEREDOC
	# As root run the command: /opt/lsb/appbat/bin/rsync --daemon
	$result = cmd("/opt/lsb/appbat/bin/rsync --daemon");
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message($last_cmd_output);
	}
	# - Expected result: rsync server running on test system
	if ( cmd( "ps -A | grep -e '\\brsync\\b'") == 0 ) {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	} else {
		$simpjournal->test_case_result("failed");
	}
	#---------

	$simpjournal->test_case_start("List  localhost::rsync");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/rsync localhost::rsync 
 - Expected result: Files available on local rsync server are listed
HEREDOC
	# Run the command: /opt/lsb/appbat/bin/rsync localhost::rsync 
	my $msg = capture("/opt/lsb/appbat/bin/rsync localhost::rsync");
	if ( $? != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $?");
		$simpjournal->test_message($msg);
	}
	# - Expected result: Files available on local rsync server are listed
	if ( $msg =~ m/\brsyncd.conf\b/ ) {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	} else {
		$simpjournal->test_case_result("failed");
	}
	#---------

	$simpjournal->test_case_start("Retrieve localhost::rsync/rsyncd.conf");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/rsync localhost::rsync/rsyncd.conf . (ends with dot) 
- Expected result: rsync config file copied to /tmp/rsync.test directory
HEREDOC
	# Run the command: /opt/lsb/appbat/bin/rsync localhost::rsync/rsyncd.conf . (ends with dot)
	$result = cmd("/opt/lsb/appbat/bin/rsync localhost::rsync/rsyncd.conf .");
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message($last_cmd_output);
	}
	# - Expected result: rsync config file copied to /tmp/rsync.test directory
	if ( -f $self->temp_dir."/rsync.test/rsyncd.conf" ) {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	} else {
		$simpjournal->test_case_result("failed");
	}
	#---------
	
	$simpjournal->test_case_start("Upload localhost::rsync/rsyncd.conf");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/rsync rsyncd.conf localhost::rsync/rsyncd.new 
- Expected result: rsync config file copied back to local rsync server directory with a new name.
HEREDOC
	# Run the command: /opt/lsb/appbat/bin/rsync rsyncd.conf localhost::rsync/rsyncd.new 
	$result = cmd("/opt/lsb/appbat/bin/rsync rsyncd.conf localhost::rsync/rsyncd.new");
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message($last_cmd_output);
	}
	# - Expected result: rsync config file copied back to local rsync 
	# server directory with a new name.
	if ( -f $self->temp_dir."/rsync/rsyncd.new" ) {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	} else {
		$simpjournal->test_case_result("failed");
	}
	#---------

	$simpjournal->test_case_start("Compare uploaded file with the original one");
	$simpjournal->test_description(<<HEREDOC);
Run the command: diff /tmp/rsync/rsyncd.conf /tmp/rsync/rsyncd.new to compare the two files copied to and from the rsync server 
- Expected result: Both files are identical
HEREDOC
	# Run the command: diff /tmp/rsync/rsyncd.conf /tmp/rsync/rsyncd.new
	$result = cmd("diff ".shq($self->temp_dir)."/rsync/rsyncd.conf ".shq($self->temp_dir)."/rsync/rsyncd.new");
	# - Expected result: rsync config file copied back to local rsync 
	# server directory with a new name.
	$simpjournal->test_case_result( $result == 0 ? "passed" : "failed");
	#---------
	
	$simpjournal->close();
	
	return 1; # Ok
}

sub Cleanup {
	
	# Stop the rsync daemon.
	cmd("killall rsync"); 
}

1; # return value
