# Distribution Checker
# Samba FVT Module (Samba_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Samba_test;
use strict;

use Misc;
use SimpleJournal;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

my $test_user = "smbuser";
my $test_user_passw = crypt("smb4lsb", time() );

my $service_to_be_restarted = 0;

#----------------------------------------------------------------------

sub prepare {
	my ($self) = @_;

	# 1. Download the lsb-samba application package that corresponds to the LSB
	#    version and processor architecture for which certification is being 
	#    sought from the Download Page. 
	# 2. Install package on system.
	# NOTE: Application MUST be installed without use of --nodeps option,
	#  the dependency checks are part of the test.
	
	# Download and install required packages
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory.", $Error::Last;
	
	# Create a test user for the FVT. Run the command: useradd -m smbuser
	if ( cmd( "id $test_user >/dev/null" ) == 0 ) {
		cmd("userdel -r $test_user");
	}
	cmd("useradd -m $test_user") == 0
		or return error "Failed to create test user '$test_user'.";
	
	# Use the passwd command to set the password for smbuser to smb4lsb
	
	# Open a subshell to set passwords for test user.
	my $subshell = $self->Spawn_subshell();
	is_ok $subshell or return $Error::Last;
	
	# Set passwords for testing users
	$subshell->Settle();
	$subshell->Send( "passwd $test_user; exit"."\n" );
	is_ok $subshell->expect_and_reply("assword:", $test_user_passw."\n") or return $Error::Last;
	is_ok $subshell->expect_and_reply("assword:", $test_user_passw."\n") or return $Error::Last;
	is_ok $subshell->ExpectLogout() or return $Error::Last;
	
	# Populate the /home/smbuser directory.
	# Run the command: cp /opt/lsb/appbat/man/man1/* /home/smbuser
	cmd ("cp /opt/lsb/appbat/man/man1/* /home/smbuser") == 0
		or return error "Failed to populate the /home/smbuser directory";
	
	# Create the file /etc/opt/lsb/appbat/smb.conf with the following contents: 
	# NOTE: The smb.conf file assumes eth0 is the network interface available 
	# on the test system. If a different interface is being used, such as eth1,
	# then it should be used on the interfaces line.
	is_ok write_string_as_file ( "/etc/opt/lsb/appbat/smb.conf", <<HEREDOC) or return error "Failed to write smb.conf", $Error::Last;
[global]
	workgroup = WORKGROUP
[homes]
	guest ok = no
	read only = no
HEREDOC
	# 'interfaces eth0' parameter was omitted.
	# By default Samba will query the kernel for the list of all  active
	# interfaces and use any interfaces except 127.0.0.1 that are broadcast capable.

	# If it does not already exist, create the file /etc/pam.d/samba with the following contents: 
	unless ( -f "/etc/pam.d/samba" ) {
		is_ok write_string_as_file ( "/etc/pam.d/samba", <<HEREDOC ) or return error "Failed to write /etc/pam.d/samba", $Error::Last;
#%PAM-1.0
auth       required     pam_unix.so
account    required     pam_unix.so
HEREDOC
	}
	
	# Run the command: grep "^nobody" /etc/passwd 
	# Expected result: Output shows user "nobody" already exists. 
	my $result = capture('grep "^nobody\b" /etc/passwd');
	if ( $result =~ /^\s*$/ ) {
		# If it is not present on the system, run the command: "useradd nobody" 
		# to create the nobody user. 
		cmd("useradd nobody") == 0
			or return error "Failed to add user 'nobody'";
	}

	# If the Samba SMB/CIFS server (smbd) and NetBIOS name server (nmbd)
	# are already running on the system under test, they should be stopped
	# so that they don't interfere with the running of the test.
	if ( cmd("ps -A | grep -e '\\bsmbd\$'") == 0 ) {
		$service_to_be_restarted = 1;
		
		# Trying to stop samba using init.d script
		foreach my $script ("smb", "samba") {
			if ( -f "/etc/init.d/$script" ) {
				inform("Stop samba daemon");
				cmd("/etc/init.d/$script stop");
			}
		}
	}
	# Check the result
	if ( cmd("ps -A | grep -e '\\bsmbd\$'") == 0 ) {
	# Failed to stop samba via script
		cmd("killall smbd");
		cmd("killall nmbd");
		cmd("rm -f /var/run/samba/smbd.pid");
	}
	
	# If the TMPDIR environment variable is set to some value, it should be
	# cleared to avoid problems with some of the tests.
	delete $ENV{"TMPDIR"};
	
	$self->{JOURNAL_FILE} = $self->name.".simplejournal";
	$self->{JOURNAL_TYPE} = 'SIMPLE';
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $result;

	my $simpjournal = SimpleJournal->new( $self->journal_file );
	is_ok($simpjournal) or return error "Failed to open journal.", $Error::Last;
	
	$simpjournal->test_name( $self->name );
	
	my $tests_expected = 7;
	$simpjournal->tests_expected( $tests_expected ); # Number of tests
	
	### Functional Verification Test for lsb-apache ###
	# http://www.linuxfoundation.org/appbat/fvt-3.0.0/lsb-samba_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.1.0/lsb-samba_fvt.html
	# http://www.linuxfoundation.org/appbat/fvt-3.2.0/lsb-samba_fvt.html
	
	###  Check installation
	$simpjournal->test_case_start("Check installation");
	
	$simpjournal->test_description(<<HEREDOC);
- Expected result: application should be installed to 
/opt/lsb/appbat
/etc/opt/lsb/appbat
/var/opt/lsb/appbat
HEREDOC
	# No files being installed to /etc/opt/lsb/appbat
	
	$result = "passed";
	for my $file ( "/opt/lsb/appbat/bin/smbpasswd",
				"/var/opt/lsb/appbat/locks" )
	{
		if ( !-e $file ) {
			$result = "failed";
			$simpjournal->test_comment("'$file' is expected to exist after installation, but not found.");
		}
	}
	
	$simpjournal->test_case_result($result);
	#---------
	
	$simpjournal->test_case_start("testparm");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/testparm 
- Expected result: The following output should be seen: 
	Load smb config files from /etc/opt/lsb/appbat/smb.conf
	Processing section "[homes]"
	Loaded services file OK.
	Server role: ROLE_STANDALONE
	Press enter to see a dump of your service definitions
After pressing "ENTER" you should see a list of the service definitions.
HEREDOC
	# Open a subshell
	my $output_file = $self->name.".output1";
	my $subshell = $self->Spawn_subshell(undef, $output_file);
	is_ok $subshell or return $Error::Last;
	
	# Set passwords for testing users
	# Run the command: /opt/lsb/appbat/bin/testparm 
	$subshell->Settle();
	$subshell->Send( "/opt/lsb/appbat/bin/testparm; exit"."\n" );
	# - Expected result: The following output should be seen: 
	#  ...
	#  Press enter to see a dump of your service definitions
	$subshell->{EXPECT_NO_COMPLAIN} = 1;
	is_ok $subshell->expect_and_reply("Press enter.*", "\n", 10) # 10 seconds maximum
		or $result = 0;
	is_ok $subshell->WaitForSubshell(10) # 10 seconds maximum
		or $result = 0;
	$subshell = undef;
	
	# Check the result
	my $output = read_file( $globals->{'test_result_dir'}."/".$output_file );
	if ( !is_ok($output) ) {
		$simpjournal->test_comment("Can't obtain the command output.");
		$simpjournal->test_case_result("failed");
	} else {
		$simpjournal->test_message( $output );
		my $result = 1;
		foreach my $regexp ( q{Load smb config files from /etc/opt/lsb/appbat/smb\.conf},
					q{Processing section "\[homes\]"},
					q{Loaded services file OK\.},
					q{Server role: ROLE_STANDALONE},
					q{\[global\]},
					q{\[homes\]},
					q{read only = No},
			 ) 
		{
			if ( $output !~ /$regexp/ ) {
				$simpjournal->test_comment("Expected: /$regexp/");
				$result = 0;
			}
		}
		
		$simpjournal->test_case_result($result ? "passed" : "failed");
	}
	#---------
	
	$simpjournal->test_case_start("smbd -D");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/sbin/smbd -D 
- Expected result: The samba SMB/CIFS server starts as a daemon 
 (This can be verified by running the ps command)
HEREDOC
	# Run the command: /opt/lsb/appbat/sbin/smbd -D
	my $cmd = "unset TMPDIR; /opt/lsb/appbat/sbin/smbd -D";
	$result = cmd( $cmd );
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message("%> $cmd");
	}
	# - Expected result: The samba SMB/CIFS server starts as a daemon 
	# (This can be verified by running the ps command)
	if ( cmd( "ps -A | grep -e '\\bsmbd\\b'") == 0 ) {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	} else {
		$simpjournal->test_case_result("failed");
	}
	#---------
	
	$simpjournal->test_case_start("nmbd -D");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/sbin/nmbd -D 
- Expected result: The samba NetBIOS name server starts as a daemon
 (This can be verified by running the ps command)
HEREDOC
	# Run the command: /opt/lsb/appbat/sbin/nmbd -D
	$cmd = "unset TMPDIR; /opt/lsb/appbat/sbin/nmbd -D";
	$result = cmd( $cmd );
	if ( $result != 0 ) {
		$simpjournal->test_comment("The command has returned non-zero code: $result");
		$simpjournal->test_message("%> $cmd");
	}
	#- Expected result: The samba NetBIOS name server starts as a daemon
	# (This can be verified by running the ps command)
	if ( cmd( "ps -A | grep -e '\\bnmbd\\b'") == 0 ) {
		$simpjournal->test_case_result( $result == 0 ? "passed" : "warning");
	} else {
		$simpjournal->test_case_result("failed");
	}
	#---------

	$simpjournal->test_case_start("smbpasswd");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/smbpasswd -a smbuser 
- Expected result: You will be prompted twice to enter the SMB password
 for the smbuser. Use the same password as in setup step #4 above, smb4lsb.
HEREDOC
	# Open a subshell
	$output_file = $self->name.".output2";
	$subshell = $self->Spawn_subshell(undef, $output_file);
	is_ok $subshell or return $Error::Last;
	
	# Run the command: /opt/lsb/appbat/bin/smbpasswd -a smbuser 
	$subshell->Settle();
	$subshell->Send( "/opt/lsb/appbat/bin/smbpasswd -a $test_user; exit"."\n" );
	# - Expected result: You will be prompted twice to enter the SMB password
	# for the smbuser. Use the same password as in setup step #4 above, smb4lsb.
	$result = 1;
	$subshell->{EXPECT_NO_COMPLAIN} = 1;
	is_ok $subshell->expect_and_reply("assword:", $test_user_passw."\n", 10) or $result=0;
	is_ok $subshell->expect_and_reply("assword:", $test_user_passw."\n", 10) or $result=0;
	is_ok $subshell->WaitForSubshell(10) # 10 seconds maximum
		or $result = 0;
	$subshell = undef;
	
	$simpjournal->test_case_result($result ? "passed" : "failed");
	#---------

	$simpjournal->test_case_start("smbclient -L");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/smbclient -L localhost -U smbuser 
- Expected result: After entering the password for smbuser it will list
 all the items being shared by the system.
HEREDOC
	# Open a subshell
	$output_file = $self->name.".output3";
	$subshell = $self->Spawn_subshell(undef, $output_file);
	is_ok $subshell or return $Error::Last;

	# Run the command: /opt/lsb/appbat/bin/smbclient -L localhost -U smbuser 
	# [ 'localhost' was replaced with 127.0.0.1, because on some distros
	#  'localhost' is mapped to an IPv6 address, and smb can't handle it. ]
	$subshell->Settle();
	$subshell->Send( "/opt/lsb/appbat/bin/smbclient -L 127.0.0.1 -U $test_user; exit"."\n" );
	# - Expected result: After entering the password for smbuser it will list
	# all the items being shared by the system.
	$result = 1;
	$subshell->{EXPECT_NO_COMPLAIN} = 1;
	is_ok $subshell->expect_and_reply("assword:", $test_user_passw."\n", 10) or $result=0;
	is_ok $subshell->WaitForSubshell(10) # 10 seconds maximum
		or $result=0;
	$subshell = undef;
	
	# Check the result
	$output = read_file( $globals->{'test_result_dir'}."/".$output_file );
	if ( !is_ok($output) ) {
		$simpjournal->test_comment("Can't obtain the command output.");
		$simpjournal->test_case_result("failed");
	}
	elsif ( !$result ) {
		$simpjournal->test_comment("smbclient hasn't asked a password");
		$simpjournal->test_case_result("failed");
	}
	else {
		$simpjournal->test_message( $output );
		
		if ( $output =~ m{\bhomes\b}
			&& $output =~ m{\b$test_user\b} ) 
		{
			$simpjournal->test_case_result("passed");
		} else {
			$simpjournal->test_comment("Output doesn't look familiar to the test");
			$simpjournal->test_case_result("failed");
		}
	}
	#---------

	$simpjournal->test_case_start("smbclient browsing");
	$simpjournal->test_description(<<HEREDOC);
Run the command: /opt/lsb/appbat/bin/smbclient //localhost/smbuser -U smbuser 
- Expected result: After entering the password for smbuser you will be presented
 with the smb prompt. If you type "dir" at the prompt it will list all the files
  in the /home/smbuser directory. Type "exit" to disconnect from the samba server.
HEREDOC
	# Open a subshell
	$output_file = $self->name.".output4";
	$subshell = $self->Spawn_subshell(undef, $output_file);
	is_ok $subshell or return $Error::Last;

	# Run the command: /opt/lsb/appbat/bin/smbclient //localhost/smbuser -U smbuser 
	# [ 'localhost' was replaced with 127.0.0.1, because on some distros
	#  'localhost' is mapped to an IPv6 address, and smb can't handle it. ]
	$subshell->Settle();
	$subshell->Send( "/opt/lsb/appbat/bin/smbclient //127.0.0.1/smbuser -U $test_user; exit"."\n" );
	# - Expected result: After entering the password for smbuser you will be presented
	#  with the smb prompt. If you type "dir" at the prompt it will list all the files
	#  in the /home/smbuser directory. Type "exit" to disconnect from the samba server.
	$result = 1;
	$subshell->{EXPECT_NO_COMPLAIN} = 1;
	is_ok $subshell->expect_and_reply("assword:", $test_user_passw."\n", 10)
		or $result=0;
	$subshell->Settle(undef, 10, 2);
	$subshell->Send( "dir"."\n" );
	$subshell->Settle(undef, 10, 2);
	$subshell->Send( "exit"."\n" );
	is_ok $subshell->WaitForSubshell(10) # 10 seconds maximum
		 or $result=0;
	$subshell = undef;
	
	# Check the result
	$output = read_file( $globals->{'test_result_dir'}."/".$output_file );
	if ( !is_ok($output) ) {
		$simpjournal->test_comment("Can't obtain the command output.");
		$simpjournal->test_case_result("failed");
	}
	elsif ( !$result ) {
		$simpjournal->test_comment("smbclient hasn't asked a password");
		$simpjournal->test_case_result("failed");
	}
	else {
		$simpjournal->test_message( $output );
		
		if ( $output =~ m{\.\.\s+D\s+0} ) 
		{
			$simpjournal->test_case_result("passed");
		} else {
			$simpjournal->test_comment("Output doesn't look familiar to the test");
			$simpjournal->test_case_result("failed");
		}
	}
	#---------

	$simpjournal->close();
	
	$self->{'sslogfile'} = undef;
	
	return 1; # Ok
}

sub Cleanup {
	
	# Stop the test smbd and nmdb daemons. Use ps and kill for this.
	cmd("killall smbd");
	cmd("killall nmbd");
	
	if ( $service_to_be_restarted ) {
		# Trying to restart samba using init.d script
		foreach my $script ("smb", "samba") {
			if ( -f "/etc/init.d/$script" ) {
				inform("Restarting samba daemon...");
				cmd("/etc/init.d/$script restart");
			}
		}
	}
}

1; # return value
