# Distribution Checker
# Function Signature Checker Integration Module (Sigchk_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Sigchk_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

my $expected_testcases = {};

#----------------------------------------------------------------------

sub check_libs {
	my ($self) = @_;

	# Check for dev files (-check option of lsb-sigchk)
	my $check_cmd = "cd ".shq($self->{TESTSUITE_DIR})
			." && ".shq($self->{RUN_SCRIPT})." -check"
			." -src ".shq($self->{TEST_DATA})
			." -d ".shq($self->{DISTR_NAME});
	my $res = capture($check_cmd);
	if ( $res ) {
		my @missing_dev_files = ();
		my $i = 0; # lib counter for $expected_testcases
		foreach my $line ( split /\n/, $res ) {
			if ( $line =~ /^\[?([^ :\]]+)\]?[:\s]\s*(yes|no)/i ) {
				my $libname = $1;
				my $status = lc$2;
				if ( $status eq 'yes' ) {
					$i++;
					$expected_testcases->{$libname} = $i;
				} else {
					push @missing_dev_files, $libname;
				}
			}
		}
		if ( @missing_dev_files ) {
			my $n = scalar(@missing_dev_files);
			$self->report_warning(
				warning "The following libraries aren't checked because"
				." dev files for them wasn't found (".$n." librar".(($n==1)?"y":"ies")."):\n"
				.join(", ", @missing_dev_files)
			);
			return 0;
		}
	}
	return 1;
}

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;

	is_ok $self->prepare_options() or return $Error::Last;

	if ( -f $self->{TESTSUITE_DIR}.'/'.$self->{RUN_SCRIPT}
		&& -d $self->{TEST_DATA} )
	{
		is_ok $self->check_libs();
	}

	return 1;
}

sub prepare_options {
	my ($self) = @_;

	my $testsuite_dir = $self->option('TESTSUITE_DIR') || "/opt/lsb/test/lsb-sigchk";
	$self->{TESTSUITE_DIR} = $testsuite_dir;
	my $run_script = $self->option('RUN_SCRIPT') || "./lsb-sigchk.pl";
	$self->{RUN_SCRIPT} = $run_script;
	my $results_dir = $self->option('RESULTS_DIR') || $testsuite_dir."/results";
	$self->{RESULTS_DIR} = $results_dir;
	$self->{TEST_DATA} = $self->option('TEST_DATA') || "/opt/moblin/test/moblin-sigchk-data";

	$self->{DISTR_NAME} = $host_info->{'OS'};
	$self->{DISTR_NAME} =~ s/[^\w\_\-.,^=]/-/g; # TODO: remove when fixed in lsb-sigchk
	$self->{DISTR_NAME} =~ s/^-+//; # remove '-' at the beginning for not be mistaken with an option.
	$self->{DISTR_NAME} =~ s/-+$//; # trailing '-' are unnecessary too.

	return 1;
}

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	is_ok $self->prepare_options() or return $Error::Last;

	# The package was just installed. Run check libs to fill $expected_testcases.
	if ( -f $self->{TESTSUITE_DIR}.'/'.$self->{RUN_SCRIPT}
		&& -d $self->{TEST_DATA} )
	{
		if ( !%$expected_testcases ) {
			is_ok $self->check_libs();
		}
	}

	# Remove old results
	cmd("rm -rf ".shq($self->{RESULTS_DIR})."/*");

	return 1;
}

sub run {
	my ($self) = @_;
	
	# Open a subshell.
	my $subshell = $self->Spawn_subshell();
	is_ok $subshell or return $Error::Last;
	
	# Go to the test temp directory
	$subshell->Settle();
	$subshell->Send("cd ".shq($self->{TESTSUITE_DIR})."\n");

	# Prepare the command
	my $command = shq($self->{RUN_SCRIPT})
		." -run"
		." -src ".shq($self->{TEST_DATA})
		." -d ".shq($self->{DISTR_NAME})
		;

	$subshell->Settle();
	$subshell->Send("$command; exit\n"); # Run the test
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(2) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	return 1; # Ok
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /^testing for (\S+) ABI compliance.*[\r\n]/ ) {
		my $testcase = $1;
		my $n = $expected_testcases->{$testcase};
		if ( $n ) {
			my $total = scalar keys %$expected_testcases;
			$self->progress($n/$total);
		}
	}
}

sub after_run {
	my ($self) = @_;

	$self->{JOURNAL_FILE} = $self->name.".simplejournal";
	$self->{JOURNAL_TYPE} = 'SIMPLE';

	# Find the journal file and copy it to where it should be.
	is_ok $self->copy_file( $self->{RESULTS_DIR}."/journal", $self->journal_file )
		or $self->report_error(complain $Error::Last);
	
	is_ok $self->copy_file( $self->{RESULTS_DIR}."/abi_compliance_report.html", $globals->{"test_result_dir"} )
		or $self->report_error(complain $Error::Last);
	is_ok $self->copy_file( $self->{RESULTS_DIR}."/compat_reports", $globals->{"test_result_dir"} )
		or $self->report_error(complain $Error::Last);

	$self->report( '<p>Open '.Report::local_href( "results/abi_compliance_report.html", 'full report' ).'.</p>' );

	return 1;
}

1; # return value
