# Distribution Checker
# Tcl FVT Test Module (Tcl_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Tcl_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common
#----------------------------------------------------------------------

my $test_user = "tcl4lsb";

my $checkpoints = undef;

#----------------------------------------------------------------------

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	# Create temporary directory
	is_ok $self->create_temp_dir()
		or return error "Failed to create temporary directory.", $Error::Last;

	if ( version_filter("LSB < 5.0", $globals->{'standard'}) ) {
		# Copy test data file to the temporary directory
		is_ok $self->copy_files_from_reqs_to( $self->temp_dir )
			or return error "Failed to prepare test data file.", $Error::Last;
	
		# Unpack the test data file.
		cmd("cd ".shq($self->temp_dir)." && tar xvf tcl*-tests.tar >/dev/null") == 0
			or return error "Failed to unpack test data tarball.", $Error::Last;
	} else {
		# Use the test files from the package
		cmd("cd ".shq($self->temp_dir)." && cp -ar /opt/lsb/appbat/lib/tcl/tests/* . >/dev/null") == 0
			or return error "Failed to copy test data.", $Error::Last;
	}

	# Check the unpacked files
	( -f $self->temp_dir."/all.tcl" )
		or return error "Failed to unpack test data tarball or tarball is wrong."; 

	# Create a test user.
	if ( cmd( "id $test_user >/dev/null" ) != 0 ) {
		cmd("useradd -m $test_user") == 0
			or return error "Failed to create test user '$test_user'.";
	}
	
	# Chown all the temp direrctory
	cmd( "chown ".$test_user.":users -R ".shq($self->temp_dir) ) == 0
		or return error "Failed to change owner of '".$self->temp_dir."'.";
	
	init_checkpoint_data($self->{VERSION});
	
	$self->{JOURNAL_FILE} = $self->name.".tcljournal";
	$self->{JOURNAL_TYPE} = 'TCL';
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	# Open a subshell
	my $subshell = $self->Spawn_subshell($test_user);
	is_ok $subshell or return error $Error::Last;
	
	# Go to the test temp directory
	$subshell->Settle();
	$subshell->Send("cd ".shq($self->temp_dir)."\n");
	
	my $cmd = "";
	if ( $self->{VERSION} =~ /^8\.4\.11.*/ ) {
		# Use $self->temp_dir for journal because the subshell is run under test user.
		$cmd = "/opt/lsb/appbat/bin/tclsh8.4 all.tcl | tee ".shq($self->temp_dir."/".$self->{JOURNAL_FILE});
	}
	elsif ( $self->{VERSION} =~ /^8\.5([^\d]|$)/ ) {
		# Use $self->temp_dir for journal because the subshell is run under test user.
		$cmd = "/opt/lsb/appbat/bin/tclsh8.5 all.tcl | tee ".shq($self->temp_dir."/".$self->{JOURNAL_FILE});
	}
	else {
		return error "Unhandled version: ".$self->{VERSION};
	}
	$subshell->Settle();
	$subshell->Send("$cmd; exit\n"); # Run the test
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(3) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	return 1; # Ok
}

sub after_run {
	my ($self) = @_;
	
	is_ok $self->copy_file( $self->temp_dir."/".$self->{JOURNAL_FILE}, $self->journal_file )
		or $self->report_error(complain $Error::Last);
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /(\S+\.test)[\r\n]/ ) {
		my $testcase = $1;
		$self->progress_profiled($checkpoints, $testcase);
		return;
	}
	# else:
	$self->progress_profiled();
}

sub Cleanup {
	my ($self) = @_;
	
	if ( cmd( "id $test_user >/dev/null" ) == 0 ) {
		cmd( "userdel -r ".$test_user );
	}
}

sub init_checkpoint_data {
	my ($ver) = @_;
	
	if ( $ver lt '8.5' ) {
		# tcl v. 8.4.11-2 for LSB 3.0
		# tcl v. 8.4.11-4 for LSB 3.1
		$checkpoints = [
			[ 0, 'l.*.test' ],
			[ 3, 'env.test' ],
			[ 8, 'execute.test' ],
			[ 42, 'ioCmd.test' ],
			[ 45, 'ioUtil.test' ],
			[ 48, 'scan.test' ],
			[ 95, 'source.test' ],
			[ 99, 'unixNotfy.test' ],
		];
	}
	elsif ( $ver eq '8.5a4-1' ) {
		# tcl v. 8.5a4-1 for LSB 3.2
		$checkpoints = [
			[ 0, 'l.*.test' ],
			[ 4, 'cmdIL.test' ],
			[ 8, 'exec.test' ],
			[ 11, 'execute.test' ],
			[ 17, 'io.test' ],
			[ 46, 'ioCmd.test' ],
			[ 49, 'ioUtil.test' ],
			[ 52, 'misc.test' ],
			[ 95, 'source.test' ],
			[ 99, 'unixNotfy.test' ],
		];
	}
	else {
		# tcl v. 8.5.1-2 for LSB 4.0
		$checkpoints = [
			[ 0, 'l.*.test' ],
			[ 12, 'clock.test' ],
			[ 18, 'concat.test' ],
			[ 26, 'execute.test' ],
			[ 36, 'io.test' ],
			[ 48, 'ioCmd.test' ],
			[ 55, 'mathop.test' ],
			[ 95, 'source.test' ],
		];
	}
}

1; # return value
