# Distribution Checker
# Vsw4 Test Module (Vsw4_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Vsw4_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common

use X11_common qw/ guess_font_dir /;
#----------------------------------------------------------------------

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;

	# Vsw4 depends on the xlsfonts command
	if ( !does_shell_know ('xlsfonts') ) {
		my $packages = suggest_packages('xlsfonts');
		my $error = "Can't find the 'xlsfonts' command required by "
					.$self->name." test. Please install this tool.";
		if ( $packages->{'xlsfonts'} ) {
			$error .= " Package may be named '".$packages->{'xlsfonts'}."'.";
		}
		return error $error;
	}
	
	if ( version_filter("LSB <= 3.1", $globals->{'standard'}) ) {
	# For LSB 3.1 and older releases
		# Check Xvfb
		does_shell_know ("Xvfb")
			or return error "Can't find the 'Xvfb' tool required by "
					.$self->name." test."
					." Note that lsb-xvfb can't be used instead Xvfb, because its path differs."
					." Please install Xvfb.";
	}
	
	return 1;
}

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;
	
	cmd ("killall -q Xvfb"); # Should be done before calling guess_display()
	sleep 1;
	
	( defined $self->{OPTIONS} )
		or return error "Answers are undefined"; # Should not happen
	
	my $testsuite_dir = $self->option('TESTSUITE_DIR') || "/opt/lsb/test/vsw4";
	$self->{TESTSUITE_DIR} = $testsuite_dir;
	$self->{RESULTS_DIR} = $self->option('RESULTS_DIR') || "/opt/lsb/test/vsw4/xtest/results";
	
	my $q;
	
	my $x11_fonts_dir = undef;
	local *prepare_x11_fonts_dir = sub {
		if ( defined $x11_fonts_dir ) {
			return $x11_fonts_dir;
		}
		$x11_fonts_dir = X11_common::guess_font_dir();
		is_ok $x11_fonts_dir
			or return error $self->name." test requires certain X11 fonts to be installed.", $Error::Last;
		
		debug_inform "X11 font dir: '$x11_fonts_dir';";
		
		return $x11_fonts_dir; # OK
	};

	$q = 'XT_FONTPATH';
	if ( defined $self->{OPTIONS}{$q} ) {
		if ( !defined $self->{OPTIONS}{$q}->{VALUE} ) {
			is_ok prepare_x11_fonts_dir() or return $Error::Last;
			
			$self->{OPTIONS}{$q}->{VALUE} = 
					$testsuite_dir."/xtest/fonts/,".$x11_fonts_dir."/misc/";
		}
	}

	$q = 'XT_FONTPATH_GOOD';
	if ( defined $self->{OPTIONS}{$q} ) {
		if ( !defined $self->{OPTIONS}{$q}->{VALUE} ) {
			is_ok prepare_x11_fonts_dir() or return $Error::Last;
			
			$self->{OPTIONS}{$q}->{VALUE} =
					$x11_fonts_dir."/100dpi/,".$x11_fonts_dir."/75dpi/";
		}
	}

	$q = 'XVFB_DISPLAY';
	if ( defined $self->{OPTIONS}{$q} ) {
		if ( !defined $self->{OPTIONS}{$q}->{VALUE} ) {
			$self->{OPTIONS}{$q}->{VALUE} = X11_common::guess_display();
		}
	}
	
	# Remove old config files
	cmd("rm -f ".shq($self->{TESTSUITE_DIR})."/.configured");
	
	# Remove old results
	cmd("rm -rf ".shq($self->{RESULTS_DIR})."/*");
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $subshell = $self->Spawn_subshell();
	is_ok $subshell or return $Error::Last;
	
	# Change to the directory of the testsuite
	$subshell->Settle();
	$subshell->Send( "cd ".shq($self->{TESTSUITE_DIR})."\n" );
	
	# Prepare the command
	$subshell->Settle();
	$subshell->Send("./run_vsw4.sh; exit\n"); # Run the test
	
	my $questions_wait_time = 60*10; # In seconds
	my $questions_deadline = time() + $questions_wait_time;
	
	my $started = 0;
	
	my $line;
	DIALOG_LOOP:
	while ( $line = $subshell->Read($questions_wait_time) ) {
		last if !$line; # required for 'redo'
		
		if ( $line =~ /Starting X Virtual Frame Buffer/ ) {
			$started = 1;
			last;
		}
		
		# Answer the prompts
		
		$subshell->Settle($line) or redo DIALOG_LOOP;
		
		# Formalized questions
		foreach my $opt ( values %{$self->{OPTIONS}} ) {
			next if !defined $opt->{PRE} || !defined $opt->{VALUE};
			
			my $pre = $opt->{PRE};
			if ( $line =~ /$pre/  ) {
				
				$subshell->Send( $opt->{VALUE}."\n" );
				next DIALOG_LOOP;
			}
		}
		
	} continue {
		
		# Check the questions timeout
		if ( time() > $questions_deadline ) {
			inform "Time for questions is over.";
			is_ok $subshell->ExpectLogout() or return $Error::Last;
			return error "Timeout";
		}
	}
	if ( $subshell->{TIMEOUT} ) { inform "!!! Warning: timeout at subshell"; }
	
	# Just wait for the test finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	if ( !$started ) {
		return error "The test hasn't started";
	}
	
	return 1; # Ok
}

sub after_run {
	my ($self) = @_;

	$self->{JOURNAL_FILE} = $self->name.".journal";
	$self->{JOURNAL_TYPE} = 'TET';
	
	# Find the journal file and copy it to where it should be.
	is_ok $self->copy_file_globbed( $self->{RESULTS_DIR}."/0*/journal", $self->journal_file )
		or $self->report_error(complain $Error::Last);

	return 1;
}

sub Cleanup {
	my ($self) = @_;
	
	cmd ("killall Xvfb");
}

1; # return value
