# Distribution Checker
# Common X11 Test's Functionality (X11_common.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package X11_common;
use strict;

use Misc;

# Export symbols
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
	&guess_font_dir
);
#----------------------------------------------------------------------

# Find fonts needed by X11 tests.
# Returns font directory or undef.
sub guess_font_dir {
	# Try fonts from package 'lsb-xvfb'.
	my $x_font_dir = "/opt/lsb/share/fonts";
	check_font_dir($x_font_dir) and return $x_font_dir;
	
	# Try to grep the font path out from /etc/X11/xorg.conf
	if ( -f "/etc/X11/xorg.conf") {
		$x_font_dir = capture( 'grep "FontPath" /etc/X11/xorg.conf | grep 75dpi | head -n1' );
		chomp $x_font_dir;
		if ( $x_font_dir =~ s/^\s*FontPath\s+\"?// ) {
			if ( $x_font_dir =~ s/\/75dpi.*// ) {
				check_font_dir($x_font_dir) and return $x_font_dir;
			}
		}
	}
	
	# Else, try to grep the font path out from /etc/X11/fs/config
	if ( -f "/etc/X11/fs/config") {
		$x_font_dir = capture('cat /etc/X11/fs/config | grep 75dpi | head -n1');
		chomp $x_font_dir;
		my $re_path = qr{/\w_\-}; # Symbols that can compose a path.
		if ( $x_font_dir =~ m{(^|[^$re_path])([$re_path]+/75dpi)} ) { # .../75dpi directory in a list
			$x_font_dir = $1;
			if ( $x_font_dir =~ s{/75dpi$}{} ) { # remove "/75dpi" in the end of the string
				check_font_dir($x_font_dir) and return $x_font_dir;
			}
		}
	}
	
	# Try other variants
	$x_font_dir = "/usr/share/fonts";
		check_font_dir($x_font_dir) and return $x_font_dir;
	$x_font_dir = "/usr/share/fonts/X11";
		check_font_dir($x_font_dir) and return $x_font_dir;
	$x_font_dir = "/usr/share/X11/fonts";
		check_font_dir($x_font_dir) and return $x_font_dir;
	$x_font_dir = "/usr/X11R7/lib/X11/fonts";
		check_font_dir($x_font_dir) and return $x_font_dir;
	$x_font_dir = "/usr/X11R6/lib/X11/fonts";
		check_font_dir($x_font_dir) and return $x_font_dir;

	# Couldn't find anything =(
	return error "Couldn't find the X11 fonts directory. See README - Troubleshooting.";
}

# Auxiliary function for guess_font_dir()
# Checks if a given directory is a good font directory
sub check_font_dir {
	my ($x_font_dir) = @_;
	
	$x_font_dir =~ s/(^\s+|\s+$|\n)//g; #trim
	
	return 0 unless ( -d $x_font_dir );
	return 0 unless ( -d "$x_font_dir/misc/" );
	return 0 unless ( -d "$x_font_dir/100dpi/" );
	return 0 unless ( -d "$x_font_dir/75dpi/" );
	
	return 1;
}
#----------------------------------------------------------------------

# Guess the Desktop Environment
sub guess_DE {

	# This doesn't work in SUDO mode
	return "kde" if ( $ENV{"KDE_FULL_SESSION"} && $ENV{"KDE_FULL_SESSION"} eq "true" ); # eq true
	return "gnome" if ( $ENV{"GNOME_DESKTOP_SESSION_ID"} ); # just not empty is enough

	# This works only if xprop is present (not included in LSB)
	if ( does_shell_know("xprop" ) ) {
		return "kde" if  0 == cmd("xprop -root 2>/dev/null | grep '_KDE_' >/dev/null 2>&1");
		return "kde" if  0 == cmd ("xprop -root 2>/dev/null | grep 'QT_DESKTOP' >/dev/null 2>&1");
		return "gnome" if  0 == cmd ("xprop -root 2>/dev/null | grep '_METACITY_' >/dev/null 2>&1");
		return "gnome" if  0 == cmd ("xprop -root 2>/dev/null | grep 'NAUTILUS_DESKTOP' >/dev/null 2>&1");
	}
	
	# This may work wrong if two DEs are running
	return "kde" if  0 == cmd ("ps -A | grep '\\<kdeinit\$' >/dev/null 2>&1");
	return "gnome" if  0 == cmd ("ps -A | grep '\\<gnome-session\$' >/dev/null 2>&1");

	return ""; # Can't detect DE, use what is suggested
}

# Find first unused X display.
sub guess_display {

	my $res = ":2.0";
	
	if ( does_shell_know("xprop") ) { # xprop is not included in LSB
		# Try first 10 displays
		debug_inform "Looking for a free display...";
		for ( my $i=2; $i<10;$i++ ) {
			if ( 0 != cmd("xprop -display :".$i.".0 -root >/dev/null 2>&1") ) {
				# No properties for this display, which means it's free.
				return ":${i}.0";
			}
		}
	}

	return $res; # return default value
}

#----------------------------------------------------------------------
1; # return value
