# Distribution Checker
# Xts5 Test Module (Xts5_test.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Xts5_test;
use strict;

use Misc;
use Subshell;

use Test_common;
our @ISA = qw(Test_common); # Inherit Test_common

use X11_common;
#----------------------------------------------------------------------

my $x11_fonts_dir = undef;

my $checkpoints = undef;

#----------------------------------------------------------------------

sub check {
	my ($self) = @_;

	# Call the parent's check() function
	is_ok $self->Test_common::check() or return $Error::Last;

	if ( version_filter("LSB <= 3.1", $globals->{'standard'}) ) {
	# For LSB 3.1 and older releases
		# Check Xvfb
		does_shell_know ("Xvfb")
			or return error "Can't find the 'Xvfb' tool required by "
					.$self->name." test."
					." Note that lsb-xvfb can't be used instead Xvfb, because its path differs."
					." Please install Xvfb.";
	}
	
	return 1;
}

sub prepare {
	my ($self) = @_;

	# Call the parent's prepare function
	is_ok $self->Test_common::prepare() or return $Error::Last;

	cmd ("killall -q Xvfb"); # Should be done before calling guess_display()
	sleep 1;
	
	( defined $self->{OPTIONS} )
		or return error "Answers are undefined"; # Should not happen
	
	my $testsuite_dir = $self->option('TESTSUITE_DIR') || "/opt/lsb/test/xts5";
	$self->{TESTSUITE_DIR} = $testsuite_dir;
	$self->{RESULTS_DIR} = $self->option('RESULTS_DIR') || "/opt/lsb/test/xts5/xts5/results";
	
	my $q;
	
	my $x11_fonts_dir = undef;
	local *prepare_x11_fonts_dir = sub {
		if ( defined $x11_fonts_dir ) {
			return $x11_fonts_dir;
		}
		$x11_fonts_dir = X11_common::guess_font_dir();
		is_ok $x11_fonts_dir
			or return error $self->name." test requires certain X11 fonts to be installed.", $Error::Last;
		
		debug_inform "X11 font dir: '$x11_fonts_dir';";
		
		return $x11_fonts_dir; # OK
	};

	$q = 'XT_FONTPATH';
	if ( defined $self->{OPTIONS}{$q} ) {
		if ( !defined $self->{OPTIONS}{$q}->{VALUE} ) {
			if ( is_ok prepare_x11_fonts_dir() ) {
				$self->{OPTIONS}{$q}->{VALUE} = 
					$testsuite_dir."/xts5/fonts/,".$x11_fonts_dir."/misc/";
			} else {
				$self->{OPTIONS}{$q}->{ERROR} = $Error::Last;
			}
		}
	}

	$q = 'XT_FONTPATH_GOOD';
	if ( defined $self->{OPTIONS}{$q} ) {
		if ( !defined $self->{OPTIONS}{$q}->{VALUE} ) {
			if ( is_ok prepare_x11_fonts_dir() ) {
				$self->{OPTIONS}{$q}->{VALUE} =
					$x11_fonts_dir."/100dpi/,".$x11_fonts_dir."/75dpi/";
			} else {
				$self->{OPTIONS}{$q}->{ERROR} = $Error::Last;
			}
		}
	}
	
	$q = 'XVFB_DISPLAY';
	if ( defined $self->{OPTIONS}{$q} ) {
		if ( !defined $self->{OPTIONS}{$q}->{VALUE} ) {
			$self->{OPTIONS}{$q}->{VALUE} = X11_common::guess_display();
		}
	}
	
	init_checkpoint_data($self->{VERSION});
	
	# Remove old config files
	cmd("rm -f ".shq($testsuite_dir)."/.configured");
	cmd("rm -f ".shq($testsuite_dir)."/lock/.configured");
		
	# Remove locks that could leave if a previous run was terminated in the middle.
	cmd("rm -f `find ".shq($testsuite_dir)."/xts5/tset/Xlib10/ -name tet_lock`");
	
	# Remove old results
	cmd("rm -rf ".shq($self->{RESULTS_DIR})."/*");
	
	return 1;
}

sub run {
	my ($self) = @_;
	
	my $subshell = $self->Spawn_subshell();
	is_ok $subshell or return $Error::Last;
	
	# Change to the directory of the testsuite
	$subshell->Settle();
	$subshell->Send( "cd ".shq($self->{TESTSUITE_DIR})."\n" );
	
	# Prepare the command
	$subshell->Settle();
	$subshell->Send("./run_xts5.sh; exit\n"); # Run the test
	
	my $questions_wait_time = 60*10; # In seconds
	my $questions_deadline = time() + $questions_wait_time;
	
	my $started = 0;
	
	my $line;
	DIALOG_LOOP:
	while ( $line = $subshell->Read($questions_wait_time) ) {
		last if !defined $line; # required for 'redo'
		
		if ( $line =~ /Starting X Virtual Frame Buffer/ ) {
			$started = 1;
			last;
		}
		
		# Answer the prompts
		
		$subshell->Settle($line) or redo DIALOG_LOOP;
		
		# Formalized questions
		foreach my $opt ( values %{$self->{OPTIONS}} ) {
			next if !defined $opt->{PRE};
			
			my $pre = $opt->{PRE};
			if ( $line =~ /$pre/  ) {
				if ( !defined $opt->{VALUE} ) {
					is_ok $subshell->ExpectLogout() or $self->report_error(complain $Error::Last);
					return $opt->{ERROR} if is_err($opt->{ERROR});
					return error $self->name.": can't answer question: '".$line."'.";
				}
				$subshell->Send( $opt->{VALUE}."\n" );
				next DIALOG_LOOP;
			}
		}
	} continue {
		
		# Check the questions timeout
		if ( time() > $questions_deadline ) {
			inform "Time for questions is over.";
			is_ok $subshell->ExpectLogout() or return $Error::Last;
			return error "Timeout";
		}
	}
	if ( $subshell->{TIMEOUT} ) { inform "!!! Warning: timeout at subshell"; }
	
	# Just wait for the test finished
	while ( my $line = $subshell->Read(2) ) {
		$self->check_progress($line);
	}
	
	# Just confirm that the subshell has finished
	$subshell->WaitForSubshell();
	
	if ( $subshell->{EXIT_CODE} ) {
		$self->report_warning(warning "subshell exited with code ".$subshell->{EXIT_CODE}
			.".\n".$subshell->get_last_lines);
	}
	
	if ( !$started ) {
		return error "The test hasn't started";
	}
	
	return 1; # Ok
}

sub check_progress {
	my ($self, $line) = @_;
	
	defined $line or $line = "";
	
	if ( $line =~ /\d+:\d+:\d+\s+Execute (.*)[\r\n]/ ) {
		my $testcase = $1;
		$self->progress_profiled($checkpoints, $testcase);
		return;
	}
	# else:
	$self->progress_profiled();
}

sub after_run {
	my ($self) = @_;
	
	$self->{JOURNAL_FILE} = $self->name.".journal";
	$self->{JOURNAL_TYPE} = 'TET';
	
	# Find the journal file and copy it to where it should be.
	is_ok $self->copy_file_globbed( $self->{RESULTS_DIR}."/0*/journal", $self->journal_file )
		or $self->report_error(complain $Error::Last);
	
	return 1;
}

sub Cleanup {
	my ($self) = @_;
	
	cmd ("killall -q Xvfb");
}

sub init_checkpoint_data {
	my ($ver) = @_;
	
	# xts5 v. 5.1.5-10 for LSB 3.2
	# xts5 v. 5.1.5-16 for LSB 4.0
	$checkpoints = [
		[ 1, '/tset/Xproto/allcclr/Test' ],
		[ 9, '/tset/Xproto/opnfnt/Test' ],
		[ 16, '/tset/Xlib3/cnnctnnmbr/Test' ],
		[ 22, '/tset/Xlib3/mxcmpsofsc/MTest' ],
		[ 28, '/tset/Xlib7/strclr/Test' ],
		[ 35, '/tset/Xlib9/drwarcs/Test' ],
		[ 41, '/tset/Xlib9/fllrctngl/Test' ],
		[ 47, '/tset/Xlib10/rstscrnsvr/Test' ],
		[ 61, '/tset/Xlib12/flsh/Test' ],
		[ 68, '/tset/Xlib12/pkevnt/Test' ],
		[ 74, '/tset/Xlib12/ptbckevnt/Test' ],
		[ 80, '/tset/Xlib13/stpntrmppn/Test' ],
		[ 86, '/tset/Xlib15/strlstttxt/Test' ],
		[ 92, '/tset/Xlib17/prsgmtry/Test' ],
		[ 98, '/tset/XIproto/gtdvctl/Test' ],
	];
}

1; # return value
