#!/usr/bin/perl -w

# Distribution Checker
# Journal Comparer (jdiff.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use FindBin;

sub BEGIN {
	# Add current directory to the @INC to be able to include modules
	# from the same directory where this script is located.
	unshift @INC, $FindBin::Bin;
	
	$| = 1; # Immediate flush of all output
}

use Common;
use Report;

#----------------------------------------------------------------------

my $show_equal = 0;  # Show matching test cases.
my $cmp_undef = 1;   # Show 'skipped' test cases.
my $cmp_content = 1; # Compare error messages.

my $report_diff = undef;

my $webui = undef;
my $report_link = undef;

#----------------------------------------------------------------------

my $box_id = 0;

#======================================================================

sub print_help {
	
	print <<HEREDOC;
Usage: $0 [-r] [-u] [-e] [-c] file1.journal file2.journal file3.journal ...

HEREDOC
	
	print "    -r       Compare reports. Result directories should be specified with this option.\n\n";
	
	unless ( $show_equal ) {
		print "    -e       Show matching test cases.\n";
	} else {
		print "    -e       Don't show matching test cases.\n";
	}
	unless ( $cmp_undef ) {
		print "    -u       Show 'skipped' test cases.\n";
	} else {
		print "    -u       Don't show 'skipped' test cases.\n";
	}
	unless ( $cmp_content ) {
		print "    -c       Compare error messages.\n";
	} else {
		print "    -c       Don't compare error messages.\n";
	}
}
#-----------------------------------------------------------------------

# This function adds $jstruct by merging it into the $hash & $order structures.
# It is assumed, that $jstruct contains test cases in the same order as in
# other (already added) journals, though it may contain new test cases.
# The goal is to place test cases alongside with ones already added to $hash,
# and insert new test cases into the $order if any.
sub add_jstruct {
	my ($hash, $order, $jstruct, $n) = @_;
	
	# $hash->{$testcase}->[$n] == $journal_entry
	# @$order - united list of test cases from all the journals.
	
	# So what is the idea?
	# We iterate testcases from the $jstruct one by one
	#  jointly with testcases in the @$order.
	# As long as testcases in the $jstruct coinide with ones from the @$order,
	#  we place them alongside to $hash->{$testcase}->[$n] and push to @$new_order.
	# If the testcase in $jstruct isn't equal to the next in the @$order,
	#  we bypass some elements in the @$order until meet this testcase.
	# We do know if the testcase will be met in the @$order or not from the $hash.
	# All bypassed testcases are marked in the hash with a "?" symbol.
	# If we then meet a testcase that is marked with the "?" symbol in the hash,
	# we don't search it forward in the @$order, but do know it has been already bypassed,
	# and just put it in place of the "?". It happens when a testcase goes
	# not in the same order as in already added journals.
	
	my $new_order = [];
	
	my $i = 0; # position in the @$order.
	foreach my $entry ( @{$jstruct->{ARR}} ) {
		my $testcase;
		my $val;
		if ( ref($entry) eq "" ) {
			if ( $entry =~ /^([^:]*):(.*)$/ ) {
				$val = $1;
				$testcase = $2;
			}
		} else {
			( ref($entry) eq "HASH" ) or return error "Wrong entry type in jstruct";
			
			$testcase = $entry->{TEST_NAME} || "***";
			
			$testcase .= " ".$entry->{TEST_POINT} if defined $entry->{TEST_POINT};
			
			$val = $entry;
		}

		FOR_TESTCASE: {
			if ( !defined $hash->{$testcase} ) {
				# This element hasn't ever been met in previous journals.
				push @$new_order, $testcase;
				$hash->{$testcase}->[$n] = $val;
				next;
			}
			
			# else: known test case
			# This test case has been already met in previously added journals.
			
			if ( defined $hash->{$testcase}->[$n] ) { # already counted
				# This test case has been previously bypassed.
				# This happens when a test case goes not in the same order
				#  as in already added journals. A "?" symbol should be here,
				# otherwise it means that this testcase was encountered twice.
				if ( $hash->{$testcase}->[$n] eq "?" ) { # unpaired
					$hash->{$testcase}->[$n] = $val;
				} else { # paired
					# Double testcase name. Add '@' chars until unique
					print STDERR "Warning (journal ".($n+1)."): Double: $testcase\n";
					$testcase .= '@' while (defined $hash->{$testcase} && defined $hash->{$testcase}->[$n]);

					# The $testcase has changed, so redo the algorithm
					redo FOR_TESTCASE;
				}
			}
			else { # not counted
				# This test case is present in the $hash, therefore
				# it should be present in the @$order.
				# Since the $hash->{$testcase}->[$n] isn't defined,
				# this testcase is somewhere further in the @$order.
				# So we bypass some test cases, marking them with "?".
				# If we meet such test case later, we will put it in place of the "?". 
				
				while ( $order->[$i] ne $testcase ) {
					# Bypass elements until have met the $testcase.
					my $testcase_b = $order->[$i];
					push @$new_order, $testcase_b;
					$hash->{$testcase_b}->[$n] = "?"; # Mark as unpaired
					$i++;
					die if $i > $#$order; # Should not happen
				}
				
				# add the $testcase
				push @$new_order, $testcase;
				$hash->{$testcase}->[$n] = $val;
				$i++;
			}
		}
	}
	
	# Clean up all "?" marks
	foreach my $testcase ( keys %$hash ) {
		local $_ = $hash->{$testcase}->[$n];
		if ( defined $_ && $_ eq "?" ) {
			$hash->{$testcase}->[$n] = undef;
		}
	}
	
	# There may left some elements in the @$order.
	# We should put them back to the @$new_order.
	while ( $i < @$order ) {
		push @$new_order, $order->[$i];
		$i++;
	}
	
	# Swap $order and $new_order
	$order = $_[1] = $new_order;
}
#-----------------------------------------------------------------------

sub cmp_jentries ($$) {
	my ( $a, $b ) = @_;
	
	if ( !defined $a ) {
		if ( defined $b ) {
			return 1; # not equal
		} else {
			return 0; # equal
		}
	}
	else {
		if ( !defined $b ) {
			return 1; # not equal
		}
	}
	
	if ( ref($a) eq "" ) {
		if ( ref($b) ne "" ) { return 1 } # not equal
		
		return ($a ne $b);
	}
	else {
		if ( ref($b) eq "" ) { return 1 } # not equal
		
		if ( $cmp_content ) {
			foreach my $key ( 'SEVERITY', 'MESSAGE', 'COMMENT' ) {
				if ( defined $a->{$key} ) {
					if ( !defined $b->{$key} ) { return 1 } # not equal
					if ( $a->{$key} ne $b->{$key} ) { return 1 }
				} else {
					if ( defined $b->{$key} ) { return 1 } # not equal
				}
			}
		}
	}
	
	return 0; #equal
}

# Returns 0 if no differences.
# Returns 1 if there are differences.
# Returns -1 if there are no differences between defined elements, and no details to show in a popup.
sub line_differs {
	my ($line, $n) = @_;
	
	my $undef_diff = 0;
	my $failures = 0;
	
	my $last_defined = $line->[0];
	# Yes, at the first iteration it compares $line->[0] with itself.
	# It's required, for example, to set '$failures' flag.
	for ( my $i = 0; $i < $n; $i++ ) {
		if ( !defined $line->[$i] ) {
			if ( defined $last_defined ) {
				$undef_diff = 1;
			}
		} else {
			if ( ref($line->[$i]) ne "" ) {
				$failures = 1;
			}
			if ( !defined $last_defined ) {
				$undef_diff = 1;
			} else {
				if ( cmp_jentries($last_defined, $line->[$i]) ) {
					return 1;
				}
			}
			$last_defined = $line->[$i];
		}
	}
	
	if ( $cmp_undef && $undef_diff ) {
		if ( $failures ) {
			return 1;
		} else {
			return -1;
		}
	}
	return 0;
}
#---------------

sub diff_line_popup {
	my ($row_title, $testcase, $line, $col_title, @titles) = @_;
	
	my $res = "";
	
	$res .= '<div style="display:none" id="b'.$box_id.'">';
	$res .= "<h3><b>$row_title:</b> ".$testcase.'</h3>';
	$res .= '<hr/>';
	for ( my $i = 0; $i < @titles; $i++ ) {
		$res .= '<div style="padding:.5em;margin:0;'
				.( ($i % 2) ? 'background:#F0F0F0;' : '').'">';
		$res .= "<p><b>$col_title:</b> ".$titles[$i].'</p>';
		if ( defined $line->[$i] ) {
			my $entry = $line->[$i];
			my $severity = "";
			if ( ref($entry) eq "" ) {
				$severity = ($entry or "passed");
			}
			else {
				( ref($entry) eq "HASH" ) or die "Wrong entry type in jstruct" if DEBUG;
				
				$severity = $entry->{SEVERITY};
			}
			if ( $entry->{WAIVE} ) {
				if ( defined $Report::severity_map{lc $entry->{WAIVE}} ) {
					$severity = $Report::severity_map{lc $entry->{WAIVE}};
				} else {
					$severity .= ' (waived)' unless $severity eq 'passed';
				}
			}
			$res .= '<p><b>Verdict:</b> '.html($severity).'</p>';
			
			if ( ref($entry) ne "" ) {
				if ( defined $entry->{RESOLUTION} && $entry->{RESOLUTION} ne "" ) {
					$res .= '<p><b>Resolution:</b></p>'
						.'<pre style="overflow:auto">'.Report::html_excl($entry->{RESOLUTION}).'</pre>';
				}
				if ( defined $entry->{DESCRIPTION} && $entry->{DESCRIPTION} ne "" ) {
					$res .= '<p><b>Test Description:</b></p>'
						.'<pre style="overflow:auto">'.Report::html_excl($entry->{DESCRIPTION}).'</pre>';
				}
				if ( defined $entry->{MESSAGE} && $entry->{MESSAGE} ne "" ) {
					$res .= '<p><b>Messages from the test:</b></p>'
						.'<pre style="overflow:auto">'.Report::html_excl($entry->{MESSAGE}).'</pre>';
				}
				if ( defined $entry->{COMMENT} && $entry->{COMMENT} ne "" ) {
					$res .= '<p><b>Comment:</b></p>'
						.'<pre style="overflow:auto">'.Report::html_excl($entry->{COMMENT}).'</pre>';
				}
				if ( defined $entry->{PROBLEM_INFO} ) { # from problem DB
					$res .= '<p><b>Problem info:</b></p>'
						.'<p>'.Report::html_excl($entry->{PROBLEM_INFO}, sub{$_[0]=~s/\n/<br \/>\n/g}).'</p>';
				}
				# "LINE_NUM"
			}
		}
		else {
			$res .= '<p><b>-- Skipped --</b></p>';
		}
		$res .= '</div>';
		$res .= '<hr/>';
	}
	$res .= '</div>';
	
	return $res;
}

sub diff_line {
	my ($differs, $row_title, $testcase, $line, $col_title, @titles) = @_;
	
	$box_id++;

	my $double = 0;
	if ( $testcase =~ s/(\@+)$// ) { # remove added '@' characters (for double test cases)
		$double = length($1);
	}
	
	my $is_link = ( $testcase =~ /^<a/ );
	if ( !$is_link ) { $testcase = html($testcase); }
	
	my $res = "";
	
	if ( $differs == 1 ) {
		$res .= '<tr class="ct" onclick="return sd('.$box_id.')">';
		if ( $is_link ) {
			$res .= '<td style="cursor:default" onclick="event.cancelBubble=true">';
		} else {
			$res .= '<td class="ct_white">';
		}
		# Print an invisible div with detailed info on differences.
		$res .= "\n\n";
		$res .= diff_line_popup($row_title, $testcase, $line, $col_title, @titles);
		$res .= "\n\n";
		$res .= $testcase;
	} else {
		$res .= '<tr>';
		$res .= '<td style="color:gray">'.$testcase;
	}
	if ( $double ) {
		$res .= '<span style="color:red">(dup '.$double.')</span>';
	}
	$res .= "</td>\n";
	
	for ( my $i = 0; $i < @titles; $i++ ) {
		my $severity = undef;
		if ( defined $line->[$i] ) {
			my $entry = $line->[$i];
			if ( ref($entry) eq "" ) {
				$severity = ($entry or "passed");
			}
			else {
				( ref($entry) eq "HASH" ) or die "Wrong entry type in jstruct" if DEBUG;
				
				$severity = $entry->{SEVERITY};
			}
			if ( $entry->{WAIVE} ) {
				if ( defined $Report::severity_map{lc $entry->{WAIVE}} ) {
					$severity = $Report::severity_map{lc $entry->{WAIVE}};
				} else {
					$severity .= ' (waived)' unless $severity eq 'passed';
				}
			}
		}
		
		my $color_class = severity_color_class($severity);
		$color_class = "white" if !$color_class;
		$color_class = "ct_".$color_class;
		
		$res .= '<td class="'.$color_class.'">';
		$res .= $severity if defined $severity;
		$res .= "</td>\n";
	}
	$res .= "</tr>\n";
	
	return $res;
}

sub html_header {
	my ($title) = @_;
	
	my $res = "";
	
	unless ( $webui ) {
		$res .= <<HEREDOC;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en" dir="ltr">
<head>
	<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
	<title>$title</title>
<style type="text/css">
table.wikitable {
  border: 1px #aaaaaa solid;
  border-collapse: collapse;
}

table.wikitable th, table.wikitable td {
  border: 1px #aaaaaa solid;
  padding: 0.2em 0.4em;
  white-space: nowrap;
}

tr.ct {
	cursor:pointer;
}

.ct_gray       { background: #E6E6E6;}
tr.ct:hover td.ct_gray  { background: #D6D6F0;}

.ct_green       { background: #CDFFCD;}
tr.ct:hover td.ct_green { background: #A0FFA0;}

.ct_yellow       { background: #FFFFCD;}
tr.ct:hover td.ct_yellow { background: #FFFF90;}

.ct_red       { background: #FFCDCD;}
tr.ct:hover td.ct_red { background: #FFA0A0;}

.ct_white       { background: ;}
tr.ct:hover .ct_white { background: #E5E5FF;}

</style>
</head>

<body>
HEREDOC
	}
	
	$res .= <<HEREDOC;

<script language="javascript" type="text/javascript">
// <![CDATA[
function sd (box_id) {
	var w = window.open( '', 'diffdetails', 'toolbar=0,menubar=0,height=750,width=850,dependent=yes,dialog=yes');
	w.moveTo(100,0);
	w.document.body.innerHTML = document.getElementById('b' + box_id).innerHTML;
	
	w.onkeyup = function (event) {
		if (event.keyCode == 27) { // Close on ESC
			w.close ();
		}
	}

	w.focus();
	return false;
}
// ]]>
</script>

HEREDOC
	
	return $res;
}

sub table_header {
	my (@titles) = @_;
	
	my $res = "";
	
	$res .= '<table class="wikitable">'."\n";

	$res .= "<tr>\n";
	foreach my $title_html ( @titles ) {
		$res .= '<th>'. $title_html.'</th>'."\n";
	}
	$res .= "</tr>\n";

	return $res;
}

sub table_footer {
	return "</table>\n";
}

sub html_footer {
	return "" if $webui;
	
	return "</body></html>\n";
}

sub severity_color_class {
	my ($severity) = @_;
	
	if ( !defined $severity ) {
		return "";
	}
	
	if ( defined $Report::severity_map{$severity} ) {
		$severity = $Report::severity_map{$severity};
	}
	
	if ( $severity eq "" || $severity eq "passed" || $severity eq "untested" ) {
		return 'green';
	}
	elsif ( $severity =~ /(waived|quelled)/ 
		|| $severity eq "fip" || $severity eq "notice" ) 
	{
		return 'gray';
	}
	elsif ( $severity =~ /(warning)/ ) {
		return 'yellow';
	}
	else {
		return 'red';
	}
	
	return "";
}

sub compare_journals {
	my (@journals) = @_;
	
	my $hash = {};
	my $order = [];
	my @added = ();
	foreach ( my $n = 0; $n < @journals; $n++ ) {
		my $journal = $journals[$n];
		my $journal_kind = guess_journal_type($journal)
			or warning("Can't reconize journal type for ".$journal), next;
		
		my $jstruct = {
			FILE => $journal,
			KIND => $journal_kind,
		};

		is_ok Report::ParseJournal( $jstruct )
		or do {
			warning "Failed to parse journal ".$jstruct->{FILE}, $Error::Last;
			next;
		};
		
		add_jstruct($hash, $order, $jstruct, $n);
		push @added, $journal;
	}
	
	my ($result_linenames, $result_lines) = filter_results($hash, $order, scalar @added);
	$hash = $order = undef;
	
	my @titles = map '<a target="_blank" href="'.$added[$_].'">Journal #'.($_+1).'</a>', 0..$#added;
	
	# Print result
	print html_header("Compare Journals");
	
	print table_header("Test Case", @titles);

	while ( @$result_linenames ) {
		my $testcase = shift @$result_linenames;
		my $line = shift @$result_lines;
		
		my $differs = line_differs($line, scalar @titles);
		print diff_line( $differs, "Test Case", $testcase, $line, "Journal", @titles );
	}

	print table_footer();
	
	print html_footer();
}

my $skip_limit = 2;

sub filter_results {
	my ($hash, $order, $n) = @_;
	
	my @result_linenames = ();
	my @result_lines = ();
	for my $testcase ( @$order ) {
		my $differs = line_differs($hash->{$testcase}, $n);
		next if !$differs && !$show_equal;
		
		push @result_linenames, $testcase;
		push @result_lines, $hash->{$testcase};
	}
	return \@result_linenames, \@result_lines;
}

sub compare_reports {
	my (@reports) = @_;

	# Gather journals from the result directories.
	
	my $journals = {};
	for ( my $n = 0; $n < @reports; $n++ ) {
		my $report = $reports[$n];
		my $run_config = read_config($report."/runconfig");
		my @files = glob($report."/results/*");
		foreach my $file ( @files ) {
			next if !guess_journal_type($file); # Skip non-journal files.
			my $filename = extract_filename($file);
			my $jstruct = $journals->{$filename}{$report} = {};
			$jstruct->{FILE} = $file;
			$jstruct->{KIND} = guess_journal_type($filename);

			if ( (my $info_file=$file) =~ s/\.[\w]+$/.info/ ) {
				my $test_info = read_config( $info_file );
				$jstruct->{TESTSUITE} = $test_info->{NAME};
				$jstruct->{TESTSUITE_VER} = $test_info->{VERSION};
			}
			if ( $run_config ) {
				$jstruct->{STANDARD} = $run_config->{STANDARD};
				$jstruct->{ARCHITECTURE} = $run_config->{HOST_architecture};
			}
		}
	}
	
	# Prepare jornals compare tables.
	my $tables = {};
	foreach my $filename ( keys %$journals ) {
		
		my $hash = {};
		my $order = [];
		my @added = ();
		foreach my $report ( @reports ) {
			next if !$journals->{$filename}{$report};
			
			my $jstruct = {%{$journals->{$filename}{$report}}};

			is_ok Report::ParseJournal( $jstruct )
			or do {
				warning "Failed to parse journal ".$jstruct->{FILE}, $Error::Last;
				next;
			};
			
			$journals->{$filename}{$report}{STATS} = $jstruct->{STATS};
			
			add_jstruct($hash, $order, $jstruct, scalar @added);
			push @added, $report;
		}
		
		my ($result_linenames, $result_lines) = filter_results($hash, $order, scalar @added);
		
		$hash = $order = undef;
		
		$tables->{$filename} = {
			REPORTS => \@added,
			LINENAMES => $result_linenames,
			LINES => $result_lines,
		};
	}
	
	# Make reports' pseudo jstructs
	my $hash = {};
	my $order = [];
	for ( my $n = 0; $n < @reports; $n++ ) {
		my $report = $reports[$n];
		
		my $jstruct = {
				FILE => $report,
			};
		foreach my $filename ( sort keys %$journals ) {
			my $stats = $journals->{$filename}{$report}{STATS} or next;
			my $verdict = Report::verdict( $stats );
			
			my $description = 'Tests Total: '.$stats->{TESTS_TOTAL}."\n"
				.'Passed: '.$stats->{PASSED}."\n"
				.'Failed: '.$stats->{FAILURES}."\n"
				.'Warnings: '.$stats->{WARNINGS}."\n"
				.'FIPs: '.$stats->{FIPS}."\n"
				.'Test suite errors: '.$stats->{TESTSUITE_ERRORS}."\n"
				;
			
			my $jentry = {
					'TEST_NAME' => $filename,
					'SEVERITY' => $verdict,
					'COMMENT' => $description,
				};
			push @{$jstruct->{ARR}}, $jentry;
		}
		add_jstruct($hash, $order, $jstruct, $n);
	}
	
	print html_header("Compare Reports");
	
	# Print TOC
	print "<h1>Compare Reports</h1>\n\n";
	
	my @titles = map {
			'<a target="_blank" href="'
				.( $report_link ? $report_link.extract_filename($reports[$_]) : $reports[$_] )
				.'">'.extract_filename($reports[$_]).'</a>'
		} 0..$#reports;
	
	print table_header("Journal", @titles);
	
	foreach my $testcase ( @$order ) {
		my $line = $hash->{$testcase};
		
		my $testcase_html = '<a href="#'.html($testcase).'"><b>'.html($testcase).'</b></a>';
		print diff_line( 1, "Journal", $testcase_html, $line, "Report", @titles );
	}
	
	print table_footer();
	
	# Print compare tables for each journal
	foreach my $filename ( sort keys %$tables ) {
		print "<hr/>\n";
		print '<a name="'.$filename.'"></a>'."\n";
		print "<h3>Journal ".html($filename)."</h3>\n\n";
		
		my $linenames = $tables->{$filename}{LINENAMES};
		my $lines = $tables->{$filename}{LINES};
		my $added = $tables->{$filename}{REPORTS};
		
		if ( !@$linenames ) {
			print '<span style="color:#008800">No differences</span>'."\n\n";
			next;
		}
		
		my @titles = map {
			'<a target="_blank" href="'
				.( $report_link ? $report_link.extract_filename($added->[$_]) : $added->[$_] )
				.'">'.extract_filename($added->[$_]).'</a>'
		} 0..$#$added;
		
		print table_header("Test Case", @titles);
		
		while ( @$linenames ) {
			my $testcase = shift @$linenames;
			my $line = shift @$lines;
			
			my $differs = line_differs($line, scalar @titles);
			print diff_line( $differs, "Test Case", $testcase, $line, "Report", @titles );
		}

		print table_footer();
	}
	
	print html_footer();
}

#======================================================================

#### BEGIN ####

my @journals = ();

my @args = @ARGV;

while ( @args ) {
	my $arg = shift @args;
	
	if ( $arg eq "-u" ) {
		$cmp_undef = !$cmp_undef;
	}
	elsif ( $arg eq "-e" ) {
		$show_equal = !$show_equal;
	}
	elsif ( $arg eq "-c" ) {
		$cmp_content = !$cmp_content;
	}
	elsif ( $arg eq "-r" ) {
		$report_diff = 1;
	}
	elsif ( $arg eq "--webui" ) {
		$webui = 1;
	}
	elsif ( $arg eq "--link" ) {
		$report_link = shift @args;
	}
	elsif ( $arg eq "--help" ) {
		print_help(); exit;
	}
	elsif ( $arg =~ /^-/ ) {
		print_help(); exit;
	}
	else {
		push @journals, $arg;
	}
}

if ( @journals == 1 ) {
	# DBG: Start
	#my $journal = $journals[0];
	#my $jstruct = {
		#FILE => $journal,
		#KIND => guess_journal_type($journal)||die "Can't reconize journal type for ".$journal,
	#};

	#is_ok Report::ParseJournal( $jstruct )
		#or die "Failed to parse journal ".$jstruct->{FILE}, $Error::Last;
	
	#use Print_ref;
	#print "Journal structure:\n";
	#print_ref $jstruct;
	# DBG: End
	print_help();
	exit 0;
}
if ( @journals == 0 ) {
	print_help();
	exit 0;
}

if ( $report_diff ) {
	is_ok Report::problem_db2_load($PROBLEM_DB_FILE)
		or warning "Failed to load problem database", $Error::Last;
	compare_reports(@journals);
}
else {
	compare_journals(@journals);
}

exit 0;

