#!/usr/bin/perl -w

# Distribution Checker
# Script for listing files from Manifest (list_files.pl)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use FindBin;

sub BEGIN {
	# Add current directory to the @INC to be able to include modules
	# from the same directory where this script is located.
	unshift @INC, $FindBin::Bin, $FindBin::Bin."/..";
	
	$| = 1; # Immediate flush of all output
}

use Common;
use Manifest;
use Download;

#-----------------------------------------------------------------------

my $template = "\%{LINK}\t\%{SRC}/\%{FILENAME}";
my $manifest_file = undef;
my $strip = 1;
my $standards = undef;
my $archs = undef;
my $statuses = undef;
my $package_manager = '';
my $dl_method = undef;
my $tests = undef;
my $exec = 0;
my $list = 0;
my $mirror = 0;
my $verbose = 1;
my $ignore_errors = 0;
my $base_dir = undef;
my $list_old = 0;
my $purge_old = 0;
my $force_time = 0;

#-----------------------------------------------------------------------

# Prepare list of files as defined by global options.
sub list_files {
	my ($manifest) = @_;
	
	my $res = [];
	
	# hash
	my $standards_hash = { map { $_ => 1 } split /,/, $standards } if $standards;
	my $archs_hash = { map { $_ => 1 } split /,/, $archs } if $archs;
	my $statuses_hash = { map { $_ => 1 } split /,/, $statuses } if $statuses;
	my $tests_hash = { map { $_ => 1 } split /,/, $tests } if $tests;
	
	my $added_files = {};
	
	# ->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]->{PACKAGE|PACKAGE_VER|LOCATION} = ".."
	my $composition = $manifest->{COMPOSITION};
	foreach my $standard ( keys %$composition ) {
		next if $standards_hash && !$standards_hash->{$standard};
	foreach my $testsuite ( keys %{$composition->{$standard}} ) {
		next if $tests_hash && !$tests_hash->{$testsuite};
	foreach my $arch ( keys %{$composition->{$standard}{$testsuite}} ) {
		next if $archs_hash && !$archs_hash->{$arch};
	foreach my $status ( keys %{$composition->{$standard}{$testsuite}{$arch}} ) {
		next if $statuses_hash && !$statuses_hash->{$status};
	foreach my $test_ver ( keys %{$composition->{$standard}{$testsuite}{$arch}{$status}} ) {
	foreach my $req ( @{$composition->{$standard}{$testsuite}{$arch}{$status}{$test_ver}} ) {
		my $files = Manifest::select_file($manifest, $arch, $package_manager, $req);
		if ( !is_ok($files) ) {
			warning "No files for '$req->{PACKAGE}'"
					.($req->{PACKAGE_VER} ? " v. '$req->{PACKAGE_VER}'" : "")
					.($req->{LOCATION} ? " @ '$req->{LOCATION}'" : "");
			next;
		}
		foreach my $file ( @$files ) {
			my $link = Manifest::file_link($manifest, $file, $dl_method);
			( !$added_files->{$link} ) or next;
			$added_files->{$link} = 1;
			$file->{LINK} = $link;
			push @$res, $file;
		}
	}}}}}}
	
	return $res;
}
#-----------------------------------------------------------------------

sub print_usage {
	print <<HEREDOC;
Usage: $0 OPTIONS

Examples:

 Mirror all files needed for LSB 4.0 certification:
  $0 -m Manifest -s 'LSB 4.0' -S cert -P rpm  --mirror --dir /srv/ftp/pub/lsb --purge-old

 List all files needed for LSB 4.0 certification:
  $0 -m Manifest -s 'LSB 4.0' -S cert -P rpm  --list

Options:
 -m <Manifest>  Manifest file.
 --list         List files.
 -t <template>  Output template. Default is '\%{LINK}\\t\%{SRC}/\%{FILENAME}'.
                Other possible parameters are: LOCATION, REMOVED, SIZE,
                TIMESTAMP, ARCH, KIND, PACKAGE, VERSION.
 -p <N>         Strip N leading components from SRC. One by default.

The following options restrict file list:
 -s <standard>,..  Standards (comma separated). All, if not specified.
 -a <arch>,..      Architectures (comma separated). All, if not specified.
 -S cert,..        Statuses (comma separated). All, if not specified.
 -P 'rpm'          Package manager: 'rpm' or 'dpkg' (preferred package format).
 -D 'ftp'          Preferred download protocol: 'http' or 'ftp'.
 --tests libchk,.. Chosen test suites.

 --exec    Execute command specified by '-t' template for each file.
           Exits if the command fails unless --ignore was specified.

Mirroring options:
 --mirror      Start mirroring. Synchronizes files, specified by options above.
 --dir <dir>   Mirror root directory.
 -q            Be quiet.
 --ignore      Ignore errors and continue with a next file.

 --list-old    List obsolete files.
 --purge-old   Delete obsolete files.
 --force-time  Set modification time for each file in accordance with Manifest.
 
Proxy Settings:
 -x,--proxy [<user>:<password>@]<host>:<port>[,<auth>][,notunnel]   Setup proxy.
                         <auth> - authentication method (see 'man curl')
                         'notunnel' - see '--proxytunnel' on the curl manpage.
 --http-proxy <...>    Proxy settings can be specified separately for HTTP
 --ftp-proxy  <...>      and FTP.
 --no-proxy            Don't use proxy at all.
HEREDOC
}
#-----------------------------------------------------------------------

## BEGIN ##

my @args = @ARGV;

while ( @args ) {
	my $arg = shift @args;
	
	if ( $arg eq '-t' ) {
		$template = shift @args;
	}
	elsif ( $arg eq '-m' ) {
		$manifest_file = shift @args;
	}
	elsif ( $arg eq '-p' ) {
		$strip = shift @args;
	}
	elsif ( $arg eq '-s' ) {
		$standards = shift @args;
	}
	elsif ( $arg eq '-a' ) {
		$archs = shift @args;
	}
	elsif ( $arg eq '-S' ) {
		$statuses = shift @args;
	}
	elsif ( $arg eq '-P' ) {
		$package_manager = shift @args;
	}
	elsif ( $arg eq '-D' ) {
		$dl_method = shift @args;
	}
	elsif ( $arg eq '--tests' ) {
		$tests = shift @args;
	}
	elsif ( $arg eq '--exec' ) {
		$exec = 1;
	}
	elsif ( $arg eq '--list' ) {
		$list = 1;
	}
	elsif ( $arg eq '--mirror' ) {
		download_setup();
		$mirror = 1;
	}
	elsif ( $arg eq '--dir' ) {
		$base_dir = shift @args;
	}
	elsif ( $arg eq '--ignore' ) {
		$ignore_errors = 1;
	}
	elsif ( $arg eq '-q' ) { # quiet
		$verbose = 0;
	}
	elsif ( $arg eq '--list-old' ) {
		$list_old = 1;
	}
	elsif ( $arg eq '--purge-old' ) {
		$purge_old = 1;
	}
	elsif ( $arg eq '--force-time' ) {
		$force_time = 1;
	}
	elsif ( $arg eq '--proxy' || $arg eq '--http-proxy' || $arg eq '--ftp-proxy' ) {
		local *parse_proxy_options = sub {
			my ($s, $proxy_protocol) = @_;
			
			local *set_proxy_option = sub {
				my ($key, $value) = @_;
				if ( !$proxy_protocol || $proxy_protocol eq "http" ) {
					$Download::options->{'http_'.$key} = $value;
				}
				if ( !$proxy_protocol || $proxy_protocol eq "ftp" ) {
					$Download::options->{'ftp_'.$key} = $value;
				}
			};
			
			# $s = "[user[:password]@]server[:port][,<auth>][,notunnel]"
			my ($proxy_user, $proxy_password);
			my ($proxy_host, $proxy_port);
			if ( $s =~ s/^(?:([A-Za-z0-9\-_.\@]+)(?::(.*))?\@)?([A-Za-z0-9\-_.]+)(?::(\d+))?// ) {
				$proxy_user = $1;
				$proxy_password = $2;
				$proxy_host = $3;
				$proxy_port = $4;
			}
			my ($proxy_auth, $proxy_notunnel);
			if ( $s =~ s/^,(\w+)// ) {
				if ( $1 eq 'notunnel' ) {
					$proxy_notunnel = 1;
				} else {
					$proxy_auth = $1;
					if ( $s =~ s/^,notunnel// ) {
						$proxy_notunnel = 1;
					}
				}
			}
			if ( $s ne "" ) {
				fail "Wrong --proxy option. Can't parse '$s'";
			}
			if ( !$proxy_host ) {
				fail "Wrong --proxy option. Expected: [http://][user[:password]\@]server[:port]";
			}
			set_proxy_option('proxy_host', $proxy_host);
			
			$proxy_port ||= 1080; # default proxy port
			set_proxy_option('proxy_port', $proxy_port);
			
			set_proxy_option('proxy_user', $proxy_user);
			set_proxy_option('proxy_password', $proxy_password);
			
			set_proxy_option('proxy_auth', $proxy_auth);
			set_proxy_option('proxy_notunnel', $proxy_notunnel);
		}; # end of local *parse_proxy_options()
		
		my $val = shift @args;
		
		if ( $arg eq '--proxy' ) {
			parse_proxy_options( $val );
		}
		if ( $arg eq '--http-proxy' ) {
			parse_proxy_options( $val, 'http' );
		}
		if ( $arg eq '--ftp-proxy' ) {
			parse_proxy_options( $val, 'ftp' );
		}
	}
	elsif ( $arg eq '--no-proxy' ) {
		delete $Download::options->{'http_proxy_host'};
		delete $Download::options->{'ftp_proxy_host'};
		delete $ENV{$_} for ( qw/http_proxy ftp_proxy HTTP_PROXY FTP_PROXY HTTPS_PROXY ALL_PROXY/ );
	}
	else {
		print STDERR "Unknown option: '$arg'\n\n";
		print_usage();
		exit 1;
	}
}


my $manifest;
if ( defined $manifest_file ) {
	$manifest = manifest_parse( $manifest_file );
} else {
	$manifest = load_manifest();
}
is_ok $manifest or fail "Failed to load Manifest.", $Error::Last;

my $files = list_files($manifest); # also uses global options

if ( !$base_dir || !-d $base_dir ) {
	if ( !$mirror ) {
		$base_dir = cache_dest();
	} else {
		fail "Mirror root directory (--dir) is wrong or isn't specified.";
	}
}

# Make file hash
my $file_hash = {};
foreach my $file_obj ( @$files ) {
	if ( defined $file_obj->{SRC_ORIG} ) {
		$file_obj->{SRC} = $file_obj->{SRC_ORIG}; # restore original SRC value
	} else {
		$file_obj->{SRC_ORIG} = $file_obj->{SRC}; # save original SRC value
	}
	$file_obj->{SRC} =~ s{^/?[^/]*/}{} for 1 .. $strip; # strip leading components of the path
	my $file = $base_dir."/".$file_obj->{SRC}."/".$file_obj->{FILENAME};
	$file_hash->{$file} = $file_obj;
}
my $passed_dirs = {};

foreach my $file ( sort keys %$file_hash ) {
	my $file_obj = $file_hash->{$file};
	my $file_src = $file_obj->{SRC};
	my $filename = $file_obj->{FILENAME};
	my $file_dir = extract_dir($file);
	
	# Expand -t $template
	my $out;
	if ( $list || $exec ) {
		$out = $template; # expand the template:
		$out =~ s/(?<!\\)\%{(\w+)}/(defined $file_obj->{$1} ? $file_obj->{$1} : $1)/ge;
		my $repl = { 'n' => "\n", 't' => "\t", '\\' => '\\', '%' => '%' };
		$out =~ s/\\(.)/ (defined $repl->{$1} ? $repl->{$1} : '\\'.$1) /ge; # process escaped chars

		print $out."\n";
	}
	
	# Obsolete files
	if ( $mirror || $purge_old || $list_old ) {
		if ( -d $file_dir && !$passed_dirs->{$file_dir} ) {
			$passed_dirs->{$file_dir} = 1;
			
			if ( opendir(DIR, $file_dir) ) {
				my @files_in_dir = grep { $_ !~ /^\./ } readdir(DIR);
				closedir DIR;
				
				for my $tmp_file ( @files_in_dir ) {
					next if !-f $file_dir."/".$tmp_file;
					if ( !defined $manifest->{FILES}{$file_obj->{SRC_ORIG}."/".$tmp_file} ) {
						if ( $purge_old ) {
							print "Remove obsolete '$file_src/$tmp_file'.\n" if $verbose;
							unlink($file_dir."/".$tmp_file)
								or warning("Can't delete '$file_dir/$tmp_file': $!"), next;
						} else {
							print "Obsolete file: '$file_src/$tmp_file'.\n" if $verbose;
						}
					}
				}
			} else {
				warning("Can't read dir '$file_dir': $!");
			}
		}
	}
	
	if ( $mirror ) {
		# Get this file
		# The download_file() function doesn't re-download file if it is already downloaded and doesn't differ from one on the server.
		
		if ( !-d $file_dir ) {
			system "mkdir -p ".shq($file_dir);
		}
		is_ok download_file( $file_obj->{LINK}, $file, RESUME => 1, QUIET => !$verbose )
			or $ignore_errors ? do{complain($Error::Last); next} : fail($Error::Last);
	}
	if ( $exec ) {
		system($out) == 0 or $ignore_errors or exit $?;
	}
	if ( $force_time ) {
		my $timestamp = $file_obj->{TIMESTAMP};
		if ( $timestamp ) {
			if ( -f $file ) {
				print "> $file  $timestamp\n" if $verbose;
				utime $timestamp, $timestamp, $file;
			}
			else {
				print "> $file  MISSING\n" if $verbose;
			}
		}
	}
}

exit 0;

