# Distribution Checker
# Module for Generating Lists of Tests (BuildList.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package BuildList;
use strict;
use Exporter;
use Manifest;
use Packages;
use Common;
use Error;

@BuildList::ISA = 'Exporter';
@BuildList::EXPORT = (
	qw(
		$manifest
		@supported_archs
		@all_std_versions
		&generate_modlist
		&generate_std_profiles
		&generate_js_init_data
		&generate_profile_init
	),
	@Manifest::EXPORT
);

# $manifest: reference to the hash containing Manifest data
our $manifest;

# List of error messages
our @buildlist_errors = ();

# Supported architectures
our @supported_archs;

our @all_std_versions;

# %tests_tree: Categorized list of tests
# {std_version => {category => {'SUBCAT' => {subcategory => {'TESTS' => [test1, test2, ...] } } } } }
my %tests_tree;

# Builds the tree of tests from manifest data
sub build_tests_tree() {
	@buildlist_errors = ();
	%tests_tree = ();
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
		return;
	}
	foreach my $stdv (keys %{$manifest->{'TREE'}}) {
		$tests_tree{$stdv} = {} if (!exists($tests_tree{$stdv}));
		foreach my $ts (keys %{$manifest->{'TREE'}->{$stdv}}) {
			my @path_elems = split(m/\//, $manifest->{'TREE'}->{$stdv}->{$ts});
			my $root = $tests_tree{$stdv};
			my $elem;
			while ($elem = shift @path_elems) {
				$elem =~ s/^<(.*)>$/$1/;
				$root->{$elem} = {} if (!exists($root->{$elem}));
				if (scalar(@path_elems) > 0) {
					$root->{$elem}->{'SUBCAT'} = {} if (!exists($root->{$elem}->{'SUBCAT'}));
					$root = $root->{$elem}->{'SUBCAT'};
				}
				else {
					$root->{$elem}->{'TESTS'} = [] if (!exists($root->{$elem}->{'TESTS'}));
					$root = $root->{$elem}->{'TESTS'};
				}
			}
			push @$root, $ts;
		}
	}
}

# Arrays for extended sorting:
# test suite categories
my @tests_tree_sort1 = ('Certification Tests', 'Unofficial Tests');
# test suite subcategories
my @tests_tree_sort2 = ('Command Tests', 'Static Interface Tests', 'Runtime Interface Tests', 'Moblin Specific Tests', 'Automated Application Battery Tests', 'Manual Tests');
# preferrable test suite status
my @preferred_status_list = ('cert', 'certold', 'noncert', 'beta', 'snapshot', 'local');

sub by_disp_name($$$) {
	my ($stdv, $a, $b) = @_;

	my $disp_name_a = $a;
	my ($descr_obj_a) = grep { version_filter($_->{'STANDARD'}, $stdv) } @{$manifest->{'DESCRIPTIONS'}->{$a}};
	if ($descr_obj_a) {
		$disp_name_a = $descr_obj_a->{'DISPLAYNAME'};
	}

	my $disp_name_b = $b;
	my ($descr_obj_b) = grep { version_filter($_->{'STANDARD'}, $stdv) } @{$manifest->{'DESCRIPTIONS'}->{$b}};
	if ($descr_obj_b) {
		$disp_name_b = $descr_obj_b->{'DISPLAYNAME'};
	}

	return (lc($disp_name_a) cmp lc($disp_name_b));
}

# PUBLIC: Generates HTML code with list of test suites, versions, etc.
sub generate_modlist() {
	@buildlist_errors = ();
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
		return '';
	}
	my $res = '';
	foreach my $arch (@supported_archs) {
		foreach my $stdv (@all_std_versions) {
			my $stdv_id = $stdv;
			$stdv_id =~ s/ /:/g;
			my @ts_list = ();
			$res .= "      <table id=\"arch-$arch-$stdv_id\" cellpadding=\"0\" cellspacing=\"0\" style=\"border: solid 4px white; width: 100%; display: none;\">\n";
			$res .= <<DATA;
        <tr class="list_head">
          <td style="width: 14.5em;"><input type="checkbox" name="check-all-$arch-$stdv_id" id="check-all-$arch-$stdv_id" onclick="javascript:un_check_all('all');" /></td>
          <td style="width: 7.5em; padding-left: 1em;"><span class="clickable" onclick="javascript:selectVersionAll(this); return false;" title="Select versions"><b>Version</b></span></td>
          <td style="width: 1em;"><img src="images/environment-spacer.gif" alt="" style="width: 1em;" /></td>
          <td style="width: 5em;"><span class="clickable" onclick="javascript:fileOptionsAll(this);" title="Select installation method"><b>Files</b></span></td>
          <td style="width: 1em;"><img src="images/environment-spacer.gif" alt="" style="width: 1em;" /></td>
          <td style="width: 5em;"><span class="clickable" onclick="javascript:customizeAll(this);" title="Select installation method"><b>Advanced</b></span></td>
          <td></td>
        </tr>
DATA

			foreach my $disp_name1 (sort { sort_cat($a, $b, \@tests_tree_sort1); } keys %{$tests_tree{$stdv}}) {
				my $int_name1 = lc($disp_name1);
				$int_name1 =~ s/ /_/g;
				$res .= <<DATA;
        <tr class="list_cat">
          <td colspan="7"><b><label><input type="checkbox" name="check-$int_name1-$arch-$stdv_id" id="check-$int_name1-$arch-$stdv_id" onclick="javascript:un_check_all('$int_name1');" />&nbsp;$disp_name1</label>&nbsp;&nbsp;&nbsp;<span class="active" onclick="javascript:show_hide_tooltip(this, '$int_name1', event);"><img src="images/question.png" width="16" height="17" alt="[?]" title="What's this?" /></span></b></td>
        </tr>
DATA
				foreach my $disp_name2 (sort { sort_cat($a, $b, \@tests_tree_sort2); } keys %{$tests_tree{$stdv}->{$disp_name1}->{'SUBCAT'}}) {
					my $int_name2 = lc($disp_name2);
					$int_name2 =~ s/ /_/g;
					my $row_color = 0;
					my $count3 = 0;
					my $res_tmp = '';
				  	foreach my $int_name3 (sort { by_disp_name($stdv, $a, $b) } @{$tests_tree{$stdv}->{$disp_name1}->{'SUBCAT'}->{$disp_name2}->{'TESTS'}}) {
				  		push @ts_list, $int_name3;
						$row_color = 1 - $row_color;
						my $mod_descr = $manifest->{'DESCRIPTIONS'}->{$int_name3};
						my $disp_name3 = $int_name3;
						my ($descr_obj) = grep { version_filter($_->{'STANDARD'}, $stdv) } @$mod_descr;
						if ($descr_obj) {
							$disp_name3 = $descr_obj->{'DISPLAYNAME'};
						}
						my $status_list = $manifest->{'COMPOSITION'}->{$stdv}->{$int_name3}->{$arch};
						my $enabled = 1;
						my $status = '';
						my $version = '';
						if ($status_list) {
							$status = (sort { sort_cat($a, $b, \@preferred_status_list); } keys %$status_list)[0];
							$version = (sort { vercmp($a, $b) } keys %{$status_list->{$status}})[0]
								or next;
							++$count3;
						}
						else {
							$enabled = 0;
							$status = 'unknown';
						}
						my $td_class = ($enabled ? 'dropdown' : 'dropdown_dis');
						my $color = ($enabled ? 'black' : 'gray');
						my $chkbox_state = ($enabled ? '' : '');
						my $chkbox_handlers = ($enabled ? ' onclick="javascript:update_elements_state();" ondblclick="javascript:setTimeout(\'update_elements_state()\', 500);"' : ' disabled="disabled"');
						my $statver_handlers = ($enabled ? ' onclick="javascript:selectVersion(this, \''.$int_name3.'\');"' : '');
						my $files_handlers = ($enabled ? ' onclick="javascript:selectInstall(this, \''.$int_name3.'\');"' : '');
						my $config_handlers = ($enabled ? ' onclick="javascript:customizeTS(this, \''.$int_name3.'\');"' : '');
						my $files_img = ($enabled ? 'files_normal.png' : 'files_disabled.png');
						my $config_img = ($enabled ? 'configure_answers.png' : 'configure_answers_dis.png');
						my $id_suffix = ($enabled ? '' : '-dis');
						$res_tmp .= <<DATA;
        <tr class="list_item$row_color">
          <td><label><input type="checkbox" name="mod-$arch-$stdv_id-$int_name3$id_suffix" id="mod-$arch-$stdv_id-$int_name3$id_suffix"$chkbox_handlers />&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<span style="color: $color;">$disp_name3</span></label></td>
          <td class="$td_class"$statver_handlers><img src="images/status_$status.png" alt="" width="16" height="16" id="stat-$arch-$stdv_id-$int_name3$id_suffix" />&nbsp;<span id="visver-$arch-$stdv_id-$int_name3$id_suffix">$version</span><img src="images/environment-spacer.gif" width="18" height="1" alt="" /><input type="hidden" name="ver-$arch-$stdv_id-$int_name3$id_suffix" id="ver-$arch-$stdv_id-$int_name3$id_suffix" value="$status|$version" /></td>
          <td></td>
          <td class="$td_class"$files_handlers><img src="images/$files_img" alt="" width="16" height="16" id="visinst-$arch-$stdv_id-$int_name3$id_suffix" /></td>
          <td></td>
          <td class="$td_class"$config_handlers><img src="images/$config_img" alt="" width="16" height="16" id="viscust-$arch-$stdv_id-$int_name3$id_suffix"/></td>
          <td></td>
        </tr>
DATA
						my $opt_list = $manifest->{'OPTIONS'}->{$int_name3};
						if ($opt_list) {
							my @glob_options = grep { $_->{'GLOBAL'} and version_filter($_->{'STANDARD'}, $stdv) } @$opt_list;
							foreach my $option (@glob_options) {
								foreach my $var (@{$option->{'VARIANTS'}}) {
									my $check = (($var eq $option->{'DEFAULT'}) ? ' checked="checked"' : '');
									my $id = $var;
									$id =~ s/[^A-Za-z0-9\-_:\.]/-/g;
									$res_tmp .= <<DATA;
        <tr class="list_item$row_color">
          <td colspan="7">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<label><input type="radio" name="modopt-$arch-$stdv_id-$int_name3-$option->{'ID'}$id_suffix" id="modopt-$arch-$stdv_id-$int_name3-$option->{'ID'}-$id$id_suffix" value="$var"$check disabled="disabled" onclick="javascript:update_elements_state();" ondblclick="javascript:setTimeout('update_elements_state()', 500);" />&nbsp;$option->{'DISPLAY'}->{$var}</label></td>
        </tr>
DATA
								}
							}
						}
					}
					my $id_suffix = (($count3 > 0) ? '' : '-dis');
					my $chkbox_handlers = (($count3 > 0) ? ' onclick="javascript:un_check_all(\''.$int_name1.'-'.$int_name2.'\');"' : ' disabled="disabled"');
					$res .= <<DATA;
        <tr class="list_subcat">
          <td colspan="7"><b><label><input type="checkbox" name="check-$int_name1-$int_name2-$arch-$stdv_id$id_suffix" id="check-$int_name1-$int_name2-$arch-$stdv_id$id_suffix"$chkbox_handlers />&nbsp;&nbsp;&nbsp;$disp_name2</label>&nbsp;&nbsp;&nbsp;<span class="active" onclick="javascript:show_hide_tooltip(this, '$int_name1-$int_name2', event); return false;"><img src="images/question.png" width="16" height="17" alt="[?]" title="What's this?" /></span></b></td>
        </tr>
DATA
					$res .= $res_tmp;
				}
			}
			$res .= "      </table>\n";
			$res .= "      <input type=\"hidden\" name=\"ts-sort-$arch-$stdv_id\" value=\"".join(';', @ts_list)."\" />\n";
		}
	}
	return $res;
}

# PUBLIC: Generates JS code for filling in the stdp_list array (list of profiles present in each standard version).
sub generate_std_profiles() {
	@buildlist_errors = ();
	my $res = '';
	$res .= "\tstdp_list = new Array();\n";
	$res .= "\tstdp_def_list = new Array();\n";
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
		return $res;
	}
	foreach my $stdv (sort keys %{$manifest->{'STD_PROFILES'}}) {
		my $stdv_id = $stdv;
		$stdv_id =~ s/ /:/g;
		my $std_profile_list = $manifest->{'STD_PROFILES'}->{$stdv}->{'DISPLAY_NAME'};
		$res .= "\tstdp_def_list['$stdv_id'] = '$manifest->{'STD_PROFILES'}->{$stdv}->{'DEFAULT'}';\n";
		$res .= "\tstdp_list['$stdv_id'] = new Array();\n";
		foreach my $id (sort keys %$std_profile_list) {
			my $value = $std_profile_list->{$id};
			$value =~ s/&lt;/</g;
			$value =~ s/&gt;/>/g;
			$value =~ s/&amp;/&/g;
			$res .= "\tstdp_list['$stdv_id']['$id'] = '$value';\n";
		}
	}
	return $res;
}

# List of all the files used by all test suite versions.
my @used_files = ();
my %used_packages = ();
my $local_ver_sel = {};

# Generates JS code for filling in the ts_files array (info about test suite versions, files they use, etc.).
sub generate_ts_files() {
	@buildlist_errors = ();
	my $res = "\tts_files = new Array();\n";
	$res .= "\tts_local_ver = new Array();\n";
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
		return $res;
	}
	foreach my $arch (@supported_archs) {
		foreach my $stdv (@all_std_versions) {
			my $stdv_id = $stdv;
			$stdv_id =~ s/ /:/g;
			$res .= "\tts_files['$arch-$stdv_id'] = new Array();\n";
			my $test_list = $manifest->{'COMPOSITION'}->{$stdv};
			foreach my $ts (keys %$test_list) {
				$res .= "\tts_files['$arch-$stdv_id']['$ts'] = new Array();\n";
				my $status_list = $test_list->{$ts}->{$arch};
				foreach my $status (keys %$status_list) {
					$res .= "\tts_files['$arch-$stdv_id']['$ts']['$status'] = new Array();\n";
					my $ver_list = $status_list->{$status};
					foreach my $ver (sort {$b cmp $a} keys %$ver_list) {
						my @fpaths = ();
						my $last_ver;
						foreach my $req (@{$ver_list->{$ver}}) {
							my $file;
							my $fname;
							if ($req->{LOCATION}) { 
								my $files = Manifest::select_file($manifest, $arch, $Packages::package_manager, $req);
								if (!is_ok($files) || !@$files) {
									push @buildlist_errors, $Error::Last->tostring() if defined $Error::Last;
									push @buildlist_errors, "Internal error: file not found for $Packages::package_manager/$arch/$stdv/$ts/$status/$ver (".$req->{'PACKAGE'}.($req->{'PACKAGE_VER'} ? '-'.$req->{'PACKAGE_VER'} : '').($req->{'LOCATION'} ? '@'.$req->{'LOCATION'} : '').")!";
									next;
								}
								if (scalar(@$files) > 1) {
									push @buildlist_errors, "Internal error: too many files for $Packages::package_manager/$arch/$stdv/$ts/$status/$ver (".$req->{'PACKAGE'}.($req->{'PACKAGE_VER'} ? '-'.$req->{'PACKAGE_VER'} : '').($req->{'LOCATION'} ? '@'.$req->{'LOCATION'} : '').")!";
								}
								$file = $files->[0];
								$fname = $file->{'SRC'}.'/'.$file->{'FILENAME'};
							}
							if (!$file || ($file->{'KIND'} ne 'file')) {
								my $pkg_ver = undef;
								my $pkg_versions = glob_hash($manifest, 'PACKAGES', $req->{PACKAGE}, 'VERSIONS', $arch);
								if ($pkg_versions) {
									($pkg_ver) = reverse sort grep(version_filter($req->{PACKAGE_VER}, $_), keys %$pkg_versions);
								}
								$last_ver = $pkg_ver;
								if (!$req->{LOCATION}) {
									my $pkg_disp = $req->{PACKAGE};
									if ($pkg_ver) {
										$pkg_disp .= ' '.$pkg_ver;
									}
									elsif ($req->{PACKAGE_VER}) {
										$pkg_disp .= ' '.$req->{PACKAGE_VER};
									}
									$fname = $pkg_disp;
								}
								$used_packages{$fname} = $pkg_ver;
							}
							push @fpaths, $fname;
							push @used_files, $fname;
						}
						$res .= "\tts_files['$arch-$stdv_id']['$ts']['$status']['$ver'] = new Array(".join(', ', map "'$_'", @fpaths).");\n";
						if ($ver eq '*') {
							$last_ver =~ s/\.lsb.*//;
							$local_ver_sel->{$arch}{$stdv}{$ts}{$status} = $last_ver;
							$res .= "\tts_local_ver['$arch-$stdv_id|$ts|$status']='$last_ver';\n";
						}
					}
				}
			}
		}
	}
	$res .= "\n\tbuild_file_ts();\n";
	return $res;
}

# Generates JS code for filling in the mod_time array (info about test suite estimated execution time).
sub generate_ts_times() {
	return "\tmod_time = new Array();\n";
}

# Generates JS code for filling in the file_size_full and file_size_disk arrays (sizes of files: remote and in cache).
# Prerequisited:
#   1. @used_files should be filled by generate_ts_files() call.
#   2. Package versions in Manifest->{PACKAGES} should be filled by init_pkg_info() call.
sub generate_file_sizes() {
	@buildlist_errors = ();
	my $res = "\tfile_size = new Array();\n\tfile_downloaded = new Array();\n\tfile_installed = new Array();\n\tfile_size[''] = 0;\n";
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
		return $res;
	}
	if (scalar(@used_files) == 0) {
		push @buildlist_errors, 'List of files is empty (generate_ts_files() not called)!';
		return $res;
	}
	foreach my $file (@used_files) {
		if ($manifest->{'FILES'}->{$file}) {
			$res .= "\tfile_size['$file'] = ".$manifest->{'FILES'}->{$file}->{'SIZE'}.";\n";
			$res .= "\tfile_downloaded['$file'] = ".$manifest->{'FILES'}->{$file}->{'DOWNLOADED'}.";\n";
			if ($manifest->{'FILES'}->{$file}->{'KIND'} eq 'file') {
				$res .= "\tfile_installed['$file'] = '';\n";
			}
		}
		if ($used_packages{$file}) {
			$res .= "\tfile_installed['$file'] = '".$used_packages{$file}."';\n";
		}
	}
	return $res;
}

sub generate_ts_options() {
	@buildlist_errors = ();
	my $res = "\tts_options = new Array();\n";
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
		return $res;
	}
	foreach my $arch (@supported_archs) {
		foreach my $stdv (@all_std_versions) {
			my $stdv_id = $stdv;
			$stdv_id =~ s/ /:/g;
			$res .= "\tts_options['$arch-$stdv_id'] = new Array();\n";
			foreach my $ts (keys %{$manifest->{'TREE'}->{$stdv}}) {
				my $id_list = '';
				my $opt_list = $manifest->{'OPTIONS'}->{$ts};
				if ($opt_list) {
					my @options = grep { $_->{'GLOBAL'} and version_filter($_->{'STANDARD'}, $stdv) } @$opt_list;
					$id_list = join(', ', map { "'$_->{'ID'}'" } @options);
				}
				$res .= "\tts_options['$arch-$stdv_id']['$ts'] = new Array($id_list);\n";
			}
		}
	}
	return $res;
}

# Generates JS code for filling in the ts_expect_info array (autoreplies to test suite questions).
sub generate_ts_expect_info() {
	@buildlist_errors = ();
	my $res = "\tts_expect_info = new Array();\n";
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
		return $res;
	}
	foreach my $arch (@supported_archs) {
		foreach my $stdv (@all_std_versions) {
			my $stdv_id = $stdv;
			$stdv_id =~ s/ /:/g;
			$res .= "\tts_expect_info['$arch-$stdv_id'] = new Array();\n";
			my $test_list = $manifest->{'COMPOSITION'}->{$stdv};
			foreach my $ts (keys %$test_list) {
				$res .= "\tts_expect_info['$arch-$stdv_id']['$ts'] = new Array();\n";
				my $status_list = $test_list->{$ts}->{$arch};
				foreach my $status (keys %$status_list) {
					$res .= "\tts_expect_info['$arch-$stdv_id']['$ts']['$status'] = new Array();\n";
					my $ver_list = $status_list->{$status};
					foreach my $ver (keys %$ver_list) {
						$res .= "\tts_expect_info['$arch-$stdv_id']['$ts']['$status']['$ver'] = new Array();\n";
						my $opt_list = $manifest->{'OPTIONS'}->{$ts};
						if ($opt_list) {
							my $ver_sel = $ver;
							if ($ver_sel eq '*') {
								$ver_sel = glob_hash($local_ver_sel, $arch, $stdv, $ts, $status)
							}
							next if !defined $ver_sel;
							my @options = grep { !$_->{'GLOBAL'} and version_filter($_->{'STANDARD'}, $stdv) and version_filter($_->{'TEST_VER'}, $ver_sel) } @$opt_list;
							foreach my $option (@options) {
								my $def_answer = (defined($option->{'DEFAULT'}) ? $option->{'DEFAULT'} : '[default]');
								$res .= "\tts_expect_info['$arch-$stdv_id']['$ts']['$status']['$ver']['$option->{'ID'}'] = {question: '$option->{'Q'}', def_answer: '$def_answer', answer: '$def_answer'};\n";
							}
						}
					}
				}
			}
		}
	}
	return $res;
}

# Generates JS code for filling in the mod_cat_list array (test suites to be selected on checking a category)
# and in the check_list array (subcategories to be selected on checking a category).
sub generate_mod_cat_list_check_list() {
	@buildlist_errors = ();
	my $res_m = "\tmod_cat_list = new Array();\n";
	my $res_c = "\tcheck_list = new Array();\n";
	if (!$manifest) {
		push @buildlist_errors, 'Manifest not defined!';
	return "$res_m\n$res_c";
	}
	foreach my $arch (@supported_archs) {
		foreach my $stdv (@all_std_versions) {
			my $stdv_id = $stdv;
			$stdv_id =~ s/ /:/g;
			$res_m .= "\tmod_cat_list['$arch-$stdv_id'] = new Array();\n";
			$res_c .= "\tcheck_list['$arch-$stdv_id'] = new Array();\n";
			my %mod_list = ('all' => {});
			my %cat_list = ('all' => {});
			foreach my $cat (keys %{$tests_tree{$stdv}}) {
				my $cat_id = lc($cat);
				$cat_id =~ s/ /_/g;
				$cat_list{'all'}->{$cat_id} = 1;
				foreach my $subcat (keys %{$tests_tree{$stdv}->{$cat}->{'SUBCAT'}}) {
					my $subcat_id = lc($subcat);
					$subcat_id =~ s/ /_/g;
					$subcat_id = "$cat_id-$subcat_id";
					$cat_list{'all'}->{$subcat_id} = 1;
					$cat_list{$cat_id}->{$subcat_id} = 1;
					foreach my $ts (@{$tests_tree{$stdv}->{$cat}->{'SUBCAT'}->{$subcat}->{'TESTS'}}) {
						$mod_list{'all'}->{$ts} = 1;
						$mod_list{$cat_id}->{$ts} = 1;
						$mod_list{$subcat_id}->{$ts} = 1;
					}
				}
			}
			foreach my $cat (sort keys %mod_list) {
				$res_m .= "\tmod_cat_list['$arch-$stdv_id']['$cat'] = new Array(".join(', ', map("'$_'", keys %{$mod_list{$cat}})).");\n";
			}
			foreach my $cat (sort keys %cat_list) {
				$res_c .= "\tcheck_list['$arch-$stdv_id']['$cat'] = new Array(".join(', ', map("'$_'", keys %{$cat_list{$cat}})).");\n";
			}
		}
	}
	return "$res_m\n$res_c";
}

# PUBLIC: Generates JS code for initializing all the necessary structures at once.
sub generate_js_init_data() {
	# Initialize the information about local storage
	init_pkg_info();
	return "\tarch_list = new Array(".join(', ', map("'$_'", @supported_archs)).");\n" .
		   "\tstdv_list = new Array(".join(', ', map { my $tmp = $_; $tmp =~ s/ /:/g; "'$tmp'" } @all_std_versions).");\n" .
		   "\tmod_list  = new Array(".join(', ', map("'$_'", @{$manifest->{'ALL'}->{'TESTSUITES'}})).");\n" .
		   "\n" .
		   generate_ts_files() . "\n" .
		   generate_ts_times() . "\n" .
		   generate_file_sizes() . "\n" .
		   generate_ts_options() . "\n" .
		   generate_ts_expect_info() . "\n" .
		   generate_mod_cat_list_check_list();
}

# PUBLIC: Generates JS code for setting all options according to profile data.
# $page specifies the caller page ('cert' for 'Get Certified' or 'conf' for 'Custom Tests').
sub generate_profile_init($$) {
	my ($profile, $page) = @_;
	return '' if (!$profile);
	my %js_names = (
		'NAME' => 'tester_name',
		'ORGANIZATION' => 'organization',
		'EMAIL' => 'tester_email',
		'SEND_EMAIL' => 'send_email_report',
		'VERBOSE_LEVEL' => 'verbose_level',
		'ARCHITECTURE' => 'architecture',
		'USE_INTERNET' => 'use_internet',
		'STD_VERSION' => 'std_version',
		'STD_PROFILE' => 'std_profile',
		'COMMENTS' => 'comments'
	);

	my $current_std = $profile->{'GENERAL'}->{'STD_VERSION'};
	my $current_std_id = $current_std;
	$current_std_id =~ s/ /:/g;
	my $current_arch = $profile->{'GENERAL'}->{'ARCHITECTURE'};

	my $res = <<DATA;
	var elem;
	// Reset all values changed by user
	main_form.reset();

	elem = document.getElementById('architecture');
	if (elem) {
		elem.value = '$current_arch';
	}
	elem = document.getElementById('std_version');
	if (elem) {
		elem.value = '$current_std';
	}

DATA

	if ($page ne 'cert') {
		$res .= <<DATA;
	ts_file_inst = new Array();
	ts_file_dl = new Array();
	for (var arch_std in ts_expect_info) {
		for (var ts in ts_expect_info[arch_std]) {
			for (var status in ts_expect_info[arch_std][ts]) {
				for (var ver in ts_expect_info[arch_std][ts][status]) {
					for (var id in ts_expect_info[arch_std][ts][status][ver]) {
						var option = ts_expect_info[arch_std][ts][status][ver][id];
						option.answer = option.def_answer;
					}
				}
			}
		}
	}
	show_modules();
DATA
	}

	$res .= "\tvar to_set_value = new Array();\n";
	foreach my $name (qw/NAME ORGANIZATION EMAIL VERBOSE_LEVEL STD_PROFILE/) {
		my $val = $profile->{'GENERAL'}->{$name};
		if (defined($val)) {
			$val =~ s/'/\\'/g;
			$res .= "\tto_set_value['".$js_names{$name}."'] = '$val';\n";
		}
	}
	my $comments = $profile->{'GENERAL'}->{'COMMENTS'};
	if (defined($comments)) {
		$comments =~ s/\r?\n/\\n/g;
		$comments =~ s/'/\\'/g;
		$res .= "\tto_set_value['".$js_names{'COMMENTS'}."'] = '".$comments."';\n";
	}

	$res .= "\tvar to_set_check = new Array();\n";
	foreach my $name (qw/SEND_EMAIL USE_INTERNET/) {
		$res .= "\tto_set_check['".$js_names{$name}."'] = ".($profile->{'GENERAL'}->{$name} ? "1" : "0").";\n" if (defined($profile->{'GENERAL'}->{$name}));
	}

	$res .= <<DATA;

	for (var elem_id in to_set_check) {
		elem = document.getElementById(elem_id);
		if (elem) {
			elem.checked = to_set_check[elem_id];
		}
	}
	for (var elem_id in to_set_value) {
		elem = document.getElementById(elem_id);
		if (elem) {
			elem.value = to_set_value[elem_id];
		}
	}

DATA

	if ($page ne 'cert') {
		my $set_versions = '';
		my $expect_info = '';
		my $check_ts = '';
		if (defined($profile->{'TEST_SUITES'})) {
			foreach my $ts (keys %{$profile->{'TEST_SUITES'}}) {
				my $ts_info = $profile->{'TEST_SUITES'}->{$ts};
				if ($ts_info->{'RUN'}) {
					$check_ts .= "\tto_check_ts['$ts'] = 1;\n";
				}
				if ($ts_info->{'OPTIONS'}) {
					my $check_ts_tmp = '';
					foreach (keys %{$ts_info->{'OPTIONS'}}) {
						$check_ts_tmp .= "\tto_check_ts_opt['$ts']['$_'] = '".$ts_info->{'OPTIONS'}->{$_}."';\n";
					}
					if ($check_ts_tmp ne '') {
						$check_ts .= "\tto_check_ts_opt['$ts'] = new Array();\n".$check_ts_tmp;
					}
				}
				if ($ts_info->{'AUTO_REPLIES'}) {
					$expect_info .= "\tif (check_array_elem(ts_expect_info, '$current_arch-$current_std_id', '$ts')) {\n";
					foreach my $status (keys %{$ts_info->{'AUTO_REPLIES'}}) {
						$expect_info .= "\t\tif (ts_expect_info['$current_arch-$current_std_id']['$ts']['$status']) {\n";
						foreach my $ver (keys %{$ts_info->{'AUTO_REPLIES'}->{$status}}) {
							$expect_info .= "\t\t\telem = ts_expect_info['$current_arch-$current_std_id']['$ts']['$status']['$ver'];\n\t\t\tif (elem) {\n";
							my $auto_replies = $ts_info->{'AUTO_REPLIES'}->{$status}->{$ver};
							foreach my $id (keys %$auto_replies) {
								$expect_info .= "\t\t\t\tif (elem['$id']) {\n\t\t\t\t\telem['$id'].answer = '".(defined($auto_replies->{$id}) ? $auto_replies->{$id} : '[default]')."';\n\t\t\t\t}\n";
							}
							$expect_info .= "\t\t\t}\n";
						}
						$expect_info .= "\t\t}\n";
					}
					$expect_info .= "\t}\n";
				}
				if ($ts_info->{'STATUS'} and $ts_info->{'VERSION'}) {
					$set_versions .= "\tacceptVersion('$ts', '".$ts_info->{'STATUS'}."', '".$ts_info->{'VERSION'}."', false);\n";
				}
			}
		}

		$set_versions .= "\n" if ($set_versions ne '');
		$expect_info .= "\n" if ($expect_info ne '');

		if ($check_ts ne '') {
			$res .= <<DATA;
	var to_check_ts = new Array();
	var to_check_ts_opt = new Array();
$check_ts
	var cat_list = new Array();
	// Make a copy of the array mod_cat_list (for current arch-std only).
	for (var cat in mod_cat_list[current_arch_std]) {
		cat_list[cat] = mod_cat_list[current_arch_std][cat].slice(0);
	}
	for (var mod in to_check_ts) {
		// Remove the test suite from the array cat.
		for (var cat in cat_list) {
			for (var i=0; i<cat_list[cat].length; ++i) {
				if (cat_list[cat][i] == mod) {
					cat_list[cat].splice(i, 1);
					break;
				}
			}
		}
		elem = document.getElementById('mod-' + current_arch_std + '-' + mod);
		if (elem) {
			elem.checked = true;
		}
		if (to_check_ts_opt[mod]) {
			for (var opt_name in to_check_ts_opt[mod]) {
				var opt_value = to_check_ts_opt[mod][opt_name];
				var elems = document.getElementsByName('modopt-' + current_arch_std + '-' + mod + '-' + opt_name);
				if (elems) {
					for (var i=0; i<elems.length; ++i) {
						elems[i].checked = (elems[i].value == opt_value);
					}
				}
			}
		}
	}
	for (var cat in cat_list) {
		elem = document.getElementById('check-' + cat + '-' + current_arch_std);
		if (elem) {
			elem.checked = ((cat_list[cat].length > 0) ? false : true);
		}
	}

DATA
		}

		if (defined($profile->{'FILES'})) {
			foreach my $file (keys %{$profile->{'FILES'}}) {
				if ($profile->{'FILES'}->{$file}->{'DOWNLOAD'}) {
					$res .= "\tts_file_dl['$file'] = ".$profile->{'FILES'}->{$file}->{'DOWNLOAD'}.";\n";
				}
				if ($profile->{'FILES'}->{$file}->{'INSTALL'}) {
					$res .= "\tts_file_inst['$file'] = ".$profile->{'FILES'}->{$file}->{'INSTALL'}.";\n";
				}
			}
			$res .= "\n";
		}

		$res .= $expect_info;
		$res .= $set_versions;
		$res .= "\tupdate_elements_state();\n";
	}
	else {
		$res .= "\tcontrol_profile_list();\n";
	}
	$res .= "\telem = document.getElementById('std_profile');\n\tif (elem) {\n\t\telem.value = to_set_value['std_profile'];\n\t}\n";

	return $res;
}

################################################################################
# Initialization code

sub init() {
	$Packages::package_manager = guess_package_manager(detect_OS());
	
	# Load the Manifest
	$manifest = load_manifest();
	if (!is_ok($manifest)) {
		push @buildlist_errors, error("Failed to load manifest", $Error::Last)->tostring;
		return;
	}
	@all_std_versions = @{$manifest->{'ALL'}->{'STANDARDS'}};

	# Detect the architecture
	my $arch = detect_architecture();
	if (!$arch) {
		push @buildlist_errors, "Can't determine the machine architecture";
		return;
	}
	@supported_archs = backward_compatible_platforms($arch);

	# Build the tests tree
	build_tests_tree();
}

# Initializes information about local storage
sub init_pkg_info() {
	# See what packages are installed
	
	my $packages = {};
	# $manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}[]
	for my $standard (keys %{$manifest->{COMPOSITION}}) {
	for my $testsuite (keys %{$manifest->{COMPOSITION}{$standard}}) {
	for my $arch (keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}}) {
	for my $status (keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}}) {
	for my $test_ver (keys %{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}}) {
		REQS: for my $req (@{$manifest->{COMPOSITION}{$standard}{$testsuite}{$arch}{$status}{$test_ver}}) {
			my $package_name = $req->{PACKAGE};

			# Check if this not a package but a plain file
			if ($manifest->{'PACKAGES'}->{$package_name}) {
				for my $file_key (@{$manifest->{'PACKAGES'}->{$package_name}->{'FILES'}}) {
					if ($manifest->{'FILES'}->{$file_key}->{'KIND'} eq 'file') {
						next REQS;
					}
				}
			}
			
			$packages->{$package_name} = 1;
		}
	}}}}}
	$packages = packages_info($packages);
	if (!is_ok($packages)) {
		push @buildlist_errors, $Error::Last->tostring;
	}
	else {
		$manifest->{'PACKAGES'}->{$_}->{'VERSIONS'} = $packages->{$_} for (keys %$packages);
	}

	check_files(values %{$manifest->{'FILES'}});
}

init();

1;
